@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(amqp10_client_connection).
@@ -30,7 +30,9 @@
          socket_ready/2,
          protocol_header_received/5,
          begin_session/1,
-         heartbeat/1]).
+         heartbeat/1,
+         encrypt_sasl/1,
+         decrypt_sasl/1]).
 
 %% gen_fsm callbacks.
 -export([init/1,
@@ -48,12 +50,17 @@
          opened/3,
          close_sent/3]).
 
+-export([format_status/1]).
+
 -type amqp10_socket() :: {tcp, gen_tcp:socket()} | {ssl, ssl:sslsocket()}.
 
 -type milliseconds() :: non_neg_integer().
 
 -type address() :: inet:socket_address() | inet:hostname().
 
+-type encrypted_sasl() :: {plaintext, binary()} | {encrypted, binary()}.
+-type decrypted_sasl() :: none | anon | {plain, User :: binary(), Pwd :: binary()}.
+
 -type connection_config() ::
     #{container_id => binary(), % AMQP container id
       hostname => binary(), % the dns name of the target host
@@ -70,7 +77,9 @@
       % set to a negative value to allow a sender to "overshoot" the flow
       % control by this margin
       transfer_limit_margin => 0 | neg_integer(),
-      sasl => none | anon | {plain, User :: binary(), Pwd :: binary()},
+      %% These credentials_obfuscation-wrapped values have the type of
+      %% decrypted_sasl/0
+      sasl => encrypted_sasl() | decrypted_sasl(),
       notify => pid(),
       notify_when_opened => pid() | none,
       notify_when_closed => pid() | none
@@ -90,7 +99,9 @@
         }).
 
 -export_type([connection_config/0,
-              amqp10_socket/0]).
+              amqp10_socket/0,
+              encrypted_sasl/0,
+              decrypted_sasl/0]).
 
 %% -------------------------------------------------------------------
 %% Public API.
@@ -123,6 +134,18 @@ open(Config) ->
 close(Pid, Reason) ->
     gen_statem:cast(Pid, {close, Reason}).
 
+-spec encrypt_sasl(decrypted_sasl()) -> encrypted_sasl().
+encrypt_sasl(none) ->
+    credentials_obfuscation:encrypt(none);
+encrypt_sasl(DecryptedSasl) ->
+    credentials_obfuscation:encrypt(term_to_binary(DecryptedSasl)).
+
+-spec decrypt_sasl(encrypted_sasl()) -> decrypted_sasl().
+decrypt_sasl(none) ->
+    credentials_obfuscation:decrypt(none);
+decrypt_sasl(EncryptedSasl) ->
+    binary_to_term(credentials_obfuscation:decrypt(EncryptedSasl)).
+
 %% -------------------------------------------------------------------
 %% Private API.
 %% -------------------------------------------------------------------
@@ -164,8 +187,9 @@ init([Sup, Config0]) ->
 expecting_socket(_EvtType, {socket_ready, Socket},
                  State = #state{config = Cfg}) ->
     State1 = State#state{socket = Socket},
-    case Cfg of
-        #{sasl := none} ->
+    Sasl = credentials_obfuscation:decrypt(maps:get(sasl, Cfg)),
+    case Sasl of
+        none ->
             ok = socket_send(Socket, ?AMQP_PROTOCOL_HEADER),
             {next_state, hdr_sent, State1};
         _ ->
@@ -191,16 +215,17 @@ sasl_hdr_sent({call, From}, begin_session,
 
 sasl_hdr_rcvds(_EvtType, #'v1_0.sasl_mechanisms'{
                   sasl_server_mechanisms = {array, symbol, Mechs}},
-               State = #state{config = #{sasl := Sasl}}) ->
-    SaslBin = {symbol, sasl_to_bin(Sasl)},
+               State = #state{config = #{sasl := EncryptedSasl}}) ->
+    DecryptedSasl = decrypt_sasl(EncryptedSasl),
+    SaslBin = {symbol, decrypted_sasl_to_bin(DecryptedSasl)},
     case lists:any(fun(S) when S =:= SaslBin -> true;
                       (_) -> false
                    end, Mechs) of
         true ->
-            ok = send_sasl_init(State, Sasl),
+            ok = send_sasl_init(State, DecryptedSasl),
             {next_state, sasl_init_sent, State};
         false ->
-            {stop, {sasl_not_supported, Sasl}, State}
+            {stop, {sasl_not_supported, DecryptedSasl}, State}
     end;
 sasl_hdr_rcvds({call, From}, begin_session,
                #state{pending_session_reqs = PendingSessionReqs} = State) ->
@@ -328,10 +353,45 @@ terminate(Reason, _StateName, #state{connection_sup = Sup,
 code_change(_OldVsn, StateName, State, _Extra) ->
     {ok, StateName, State}.
 
+format_status(Context = #{data := ProcState}) ->
+    %% Note: Context.state here refers to the gen_statem state name,
+    %%       so we need to use Context.data to get #state{}
+    Obfuscated = obfuscate_state(ProcState),
+    Context#{data => Obfuscated}.
+
+
 %% -------------------------------------------------------------------
 %% Internal functions.
 %% -------------------------------------------------------------------
 
+obfuscate_state(State = #state{config = Cfg0}) ->
+    Cfg1 = obfuscate_state_config_sasl(Cfg0),
+    Cfg2 = obfuscate_state_config_tls_opts(Cfg1),
+    State#state{config = Cfg2}.
+
+-spec obfuscate_state_config_sasl(connection_config()) -> connection_config().
+obfuscate_state_config_sasl(Cfg) ->
+    Sasl0 = maps:get(sasl, Cfg, none),
+    Sasl = case Sasl0 of
+               {plain, Username, _Password} ->
+                   {plain, Username, <<"[redacted]">>};
+               Other ->
+                   Other
+           end,
+    Cfg#{sasl => Sasl}.
+
+-spec obfuscate_state_config_tls_opts(connection_config()) -> connection_config().
+obfuscate_state_config_tls_opts(Cfg) ->
+    TlsOpts0 = maps:get(tls_opts, Cfg, undefined),
+    TlsOpts = case TlsOpts0 of
+        {secure_port, PropL0} ->
+            Obfuscated = proplists:delete(password, PropL0),
+            {secure_port, Obfuscated};
+        _ ->
+            TlsOpts0
+    end,
+    Cfg#{tls_opts => TlsOpts}.
+
 handle_begin_session({FromPid, _Ref},
                      #state{sessions_sup = Sup, reader = Reader,
                             next_channel = Channel,
@@ -485,8 +545,9 @@ translate_err(#'v1_0.error'{condition = Cond, description = Desc}) ->
 amqp10_event(Evt) ->
     {amqp10_event, {connection, self(), Evt}}.
 
-sasl_to_bin({plain, _, _}) -> <<"PLAIN">>;
-sasl_to_bin(anon) -> <<"ANONYMOUS">>.
+decrypted_sasl_to_bin({plain, _, _}) -> <<"PLAIN">>;
+decrypted_sasl_to_bin(anon) -> <<"ANONYMOUS">>;
+decrypted_sasl_to_bin(none) -> <<"ANONYMOUS">>.
 
 config_defaults() ->
     #{sasl => none,
