@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2017-2022 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2017-2023 Broadcom. All Rights Reserved. The term Broadcom refers to Broadcom Inc. and/or its subsidiaries.
 %%
 %% @doc The primary module for interacting with ra servers and clusters.
 
@@ -23,6 +23,8 @@
          %% queries
          members/1,
          members/2,
+         members_info/1,
+         members_info/2,
          initial_members/1,
          initial_members/2,
          local_query/2,
@@ -78,7 +80,8 @@
          cast_aux_command/2,
          register_external_log_reader/1,
          member_overview/1,
-         member_overview/2
+         member_overview/2,
+         key_metrics/1
         ]).
 
 %% xref should pick these up
@@ -283,7 +286,7 @@ start_or_restart_cluster(System, ClusterName, Machine, ServerIds) ->
 %% If there was no existing cluster and a new cluster could not be formed
 %% any servers that did manage to start are
 %% forcefully deleted.
-%% @see start_or_restart_cluster/3
+%% @see start_or_restart_cluster/4
 %% @end
 -spec start_or_restart_cluster(atom(), ra_cluster_name(), ra_server:machine_conf(),
                                [ra_server_id()], non_neg_integer()) ->
@@ -452,36 +455,37 @@ start_cluster(System, [#{cluster_name := ClusterName} | _] = ServerConfigs,
             end
     end.
 
-%% @doc Starts a new distributed ra cluster.
-%% @param ClusterName the name of the cluster.
-%% @param ServerId the ra_server_id() of the server
-%% @param Machine The {@link ra_machine:machine/0} configuration.
-%% @param ServerIds a list of initial (seed) server configurations
-%% @returns
-%% `{ok, Started, NotStarted}'  if a cluster could be successfully
-%% started. A cluster can be successfully started if more than half of the
-%% servers provided could be started. Servers that could not be started need to
-%% be retried periodically using {@link start_server/1}
-%%
-%% `{error, cluster_not_formed}' if a cluster could not be started.
-%%
-%% If a cluster could not be formed any servers that did manage to start are
-%% forcefully deleted.
-%% @see start_server/1
+%% @doc Starts an individual ra server of a cluster.
+%% @param System the system name.
+%% @param ClusterName the name of the cluster the server belongs to.
+%% @param ServerIdOrConf the `ra_server_id()' of the server, or a map with server id and settings.
+%% @param Machine The {@link ra_server:machine_conf()} configuration.
+%% @param ServerIds a list of initial (seed) server configurations for the cluster.
+%% @returns `ok'  if the server could be successfully started or `{error, Reason}' otherwise.
+%% @see start_server/2
 %% @end
--spec start_server(atom(), ra_cluster_name(), ra_server_id(),
-                   ra_server:machine_conf(), [ra_server_id()]) ->
-    ok | {error, term()}.
-start_server(System, ClusterName, {_, _} = ServerId, Machine, ServerIds)
+-spec start_server(System, ClusterName, ServerIdOrConf, Machine, ServerIds) ->
+    Ret when
+      System :: atom(),
+      ClusterName :: ra_cluster_name(),
+      ServerIdOrConf :: ServerId | ServerConf,
+      ServerId :: ra:server_id(),
+      ServerConf :: ra_new_server(),
+      Machine :: ra_server:machine_conf(),
+      ServerIds :: [ra:server_id()],
+      Ret :: ok | {error, Reason :: term()}.
+start_server(System, ClusterName, {_, _} = ServerId, Machine, ServerIds) ->
+    start_server(System, ClusterName, #{id => ServerId}, Machine, ServerIds);
+start_server(System, ClusterName, #{id := {_, _}} = Conf0, Machine, ServerIds)
   when is_atom(System) ->
-    UId = new_uid(ra_lib:to_binary(ClusterName)),
+    UId = maps:get(uid, Conf0,
+                   new_uid(ra_lib:to_binary(ClusterName))),
     Conf = #{cluster_name => ClusterName,
-             id => ServerId,
              uid => UId,
              initial_members => ServerIds,
              log_init_args => #{uid => UId},
              machine => Machine},
-    start_server(System, Conf).
+    start_server(System, maps:merge(Conf0, Conf)).
 
 %% @doc Starts a ra server in the default system
 %% @param Conf a ra_server_config() configuration map.
@@ -557,9 +561,10 @@ delete_cluster(ServerIds, Timeout) ->
 %% affect said cluster's availability characteristics (by increasing quorum node count).
 %%
 %% @param ServerLoc the ra server or servers to try to send the command to
-%% @param ServerId the ra server id of the new server.
+%% @param ServerId the ra server id of the new server, or a map with server id and settings.
 %% @end
--spec add_member(ra_server_id() | [ra_server_id()], ra_server_id()) ->
+-spec add_member(ra_server_id() | [ra_server_id()],
+                 ra_server_id() | ra_new_server()) ->
     ra_cmd_ret() |
     {error, already_member} |
     {error, cluster_change_not_permitted}.
@@ -570,7 +575,8 @@ add_member(ServerLoc, ServerId) ->
 %% @see add_member/2
 %% @end
 -spec add_member(ra_server_id() | [ra_server_id()],
-                 ra_server_id(), timeout()) ->
+                 ra_server_id() | ra_new_server(),
+                 timeout()) ->
     ra_cmd_ret() |
     {error, already_member} |
     {error, cluster_change_not_permitted}.
@@ -579,7 +585,6 @@ add_member(ServerLoc, ServerId, Timeout) ->
                            {'$ra_join', ServerId, after_log_append},
                            Timeout).
 
-
 %% @doc Removes a server from the cluster's membership configuration.
 %% This function returns after appending a cluster membership change
 %% command to the log.
@@ -589,8 +594,8 @@ add_member(ServerLoc, ServerId, Timeout) ->
 %%
 %% @param ServerRef the ra server to send the command to
 %% @param ServerId the ra server id of the server to remove
-%% @see leave_and_terminate/2
-%% @see leave_and_delete_server/2
+%% @see leave_and_terminate/4
+%% @see leave_and_delete_server/4
 %% @see add_member/2
 %% @see remove_member/3
 %% @end
@@ -638,7 +643,7 @@ trigger_election(ServerId, Timeout) ->
 %% or is permanently lost.
 %% @param ServerRef the ra server to send the command to and to remove
 %% @param ServerId the ra server to remove
-%% @see leave_and_terminate/3
+%% @see leave_and_terminate/4
 %% @end
 -spec leave_and_terminate(atom(),
                           ra_server_id() | [ra_server_id()], ra_server_id()) ->
@@ -646,11 +651,11 @@ trigger_election(ServerId, Timeout) ->
 leave_and_terminate(System, ServerRef, ServerId) ->
     leave_and_terminate(System, ServerRef, ServerId, ?DEFAULT_TIMEOUT).
 
-%% @doc Same as `leave_and_terminate/2' but also accepts a timeout.
+%% @doc Same as `leave_and_terminate/3' but also accepts a timeout.
 %% @param ServerRef the ra server to send the command to and to remove
 %% @param ServerId the ra server to remove
 %% @param Timeout timeout to use
-%% @see leave_and_terminate/2
+%% @see leave_and_terminate/3
 %% @end
 -spec leave_and_terminate(atom(),
                           ra_server_id() | [ra_server_id()],
@@ -678,7 +683,7 @@ leave_and_terminate(System, ServerRef, ServerId, Timeout) ->
 %% @param System the system identifier
 %% @param ServerRef the ra server to send the command to and to remove
 %% @param ServerId the ra server to force remove
-%% @see leave_and_delete_server/3
+%% @see leave_and_delete_server/4
 %% @end
 -spec leave_and_delete_server(atom(), ra_server_id() | [ra_server_id()],
                               ra_server_id()) ->
@@ -686,11 +691,11 @@ leave_and_terminate(System, ServerRef, ServerId, Timeout) ->
 leave_and_delete_server(System, ServerRef, ServerId) ->
     leave_and_delete_server(System, ServerRef, ServerId, ?DEFAULT_TIMEOUT).
 
-%% @doc Same as `leave_and_delete_server/2' but also accepts a timeout.
+%% @doc Same as `leave_and_delete_server/3' but also accepts a timeout.
 %% @param ServerRef the ra server to send the command to and to remove
 %% @param ServerId the ra server to force remove
 %% @param Timeout timeout to use
-%% @see leave_and_delete_server/2
+%% @see leave_and_delete_server/3
 %% @end
 -spec leave_and_delete_server(atom(), ra_server_id() | [ra_server_id()],
                               ra_server_id(), timeout()) ->
@@ -715,7 +720,6 @@ new_uid(Source) when is_binary(Source) ->
     Prefix = ra_lib:derive_safe_string(Source, 6),
     ra_lib:make_uid(string:uppercase(Prefix)).
 
-
 %% @doc Returns a map of overview data of the default Ra system on the current Erlang
 %% node.
 %% DEPRECATED: user overview/1
@@ -1035,6 +1039,45 @@ members({local, ServerId}, Timeout) ->
 members(ServerId, Timeout) ->
     ra_server_proc:state_query(ServerId, members, Timeout).
 
+%% @doc Returns a list of cluster members and their Raft metrics
+%%
+%% Except if `{local, ServerId}' is passed, the query is sent to the specified
+%% server which may redirect it to the leader if it is a follower. It may
+%% timeout if there is currently no leader (i.e. an election is in progress).
+%%
+%% With `{local, ServerId}', the query is always handled by the specified
+%% server. It means the returned list might be out-of-date compared to what the
+%% leader would have returned.
+%%
+%% @param ServerId the Ra server(s) to send the query to
+%% @end
+-spec members_info(ra_server_id() | [ra_server_id()] | {local, ra_server_id()}) ->
+    ra_server_proc:ra_leader_call_ret(ra_cluster()).
+members_info(ServerId) ->
+    members_info(ServerId, ?DEFAULT_TIMEOUT).
+
+%% @doc Returns a list of cluster members and their Raft metrics
+%%
+%% Except if `{local, ServerId}' is passed, the query is sent to the specified
+%% server which may redirect it to the leader if it is a follower. It may
+%% timeout if there is currently no leader (i.e. an election is in progress).
+%%
+%% With `{local, ServerId}', the query is always handled by the specified
+%% server. It means the returned list might be out-of-date compared to what the
+%% leader would have returned.
+%%
+%% @param ServerId the Ra server(s) to send the query to
+%% @param Timeout the timeout to use
+%% @end
+-spec members_info(ra_server_id() | [ra_server_id()] | {local, ra_server_id()},
+              timeout()) ->
+    ra_server_proc:ra_leader_call_ret(ra_cluster()).
+members_info({local, ServerId}, Timeout) ->
+    ra_server_proc:local_state_query(ServerId, members_info, Timeout);
+members_info(ServerId, Timeout) ->
+    ra_server_proc:state_query(ServerId, members_info, Timeout).
+
+
 %% @doc Returns a list of initial (seed) cluster members.
 %%
 %% This allows Ra-based systems with dynamic cluster membership
@@ -1107,6 +1150,46 @@ member_overview(ServerId) ->
 member_overview(ServerId, Timeout) ->
     ra_server_proc:local_state_query(ServerId, overview, Timeout).
 
+%% @doc Returns a map of key metrics about a Ra member
+%%
+%% The keys and values may vary depending on what state
+%% the member is in. This function will never call into the
+%% Ra process itself so is likely to return swiftly even
+%% when the Ra process is busy (such as when it is recovering)
+%%
+%% @param ServerId the Ra server to obtain key metrics for
+%% @end
+key_metrics({Name, N} = ServerId) when N == node() ->
+    Fields = [last_applied,
+              commit_index,
+              snapshot_index,
+              last_written_index,
+              last_index,
+              commit_latency,
+              term],
+    Counters = case ra_counters:counters(ServerId, Fields) of
+                   undefined ->
+                       #{};
+                   C -> C
+               end,
+    case whereis(Name) of
+        undefined ->
+            Counters#{state => noproc,
+                      membership => unknown};
+        _ ->
+            case ets:lookup(ra_state, Name) of
+                [] ->
+                    Counters#{state => unknown,
+                              membership => unknown};
+                [{_, State, Membership}] ->
+                    Counters#{state => State,
+                              membership => Membership}
+            end
+    end;
+key_metrics({_, N} = ServerId) ->
+    erpc:call(N, ?MODULE, ?FUNCTION_NAME, [ServerId]).
+
+
 %% internal
 
 -spec usr(UserCommand, ReplyMode) -> Command when
