@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2017-2022 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2017-2023 Broadcom. All Rights Reserved. The term Broadcom refers to Broadcom Inc. and/or its subsidiaries.
 %%
 %% @hidden
 -module(ra_log).
@@ -100,7 +100,7 @@
          % if this is set a snapshot write is in progress for the
          % index specified
          cache = ra_log_cache:init() :: ra_log_cache:state(),
-         last_resend_time :: 'maybe'(integer()),
+         last_resend_time :: option(integer()),
          reader :: ra_log_reader:state(),
          readers = [] :: [pid()]
         }).
@@ -200,8 +200,10 @@ init(#{uid := UId,
                         reader = Reader,
                         snapshot_state = SnapshotState
                        },
-
+    put_counter(Cfg, ?C_RA_SVR_METRIC_SNAPSHOT_INDEX, SnapIdx),
     LastIdx = State000#?MODULE.last_index,
+    put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_INDEX, LastIdx),
+    put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_WRITTEN_INDEX, LastIdx),
     % recover the last term
     {LastTerm0, State00} = case LastIdx of
                                SnapIdx ->
@@ -374,7 +376,8 @@ last_written(#?MODULE{last_written_index_term = LWTI}) ->
 %% forces the last index and last written index back to a prior index
 -spec set_last_index(ra_index(), state()) ->
     {ok, state()} | {not_found, state()}.
-set_last_index(Idx, #?MODULE{cache = Cache0,
+set_last_index(Idx, #?MODULE{cfg = Cfg,
+                             cache = Cache0,
                              last_written_index_term = {LWIdx0, _}} = State0) ->
     case fetch_term(Idx, State0) of
         {undefined, State} ->
@@ -385,6 +388,8 @@ set_last_index(Idx, #?MODULE{cache = Cache0,
             %% this should always be found but still assert just in case
             true = LWTerm =/= undefined,
             Cache = ra_log_cache:set_last(Idx, Cache0),
+            put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_INDEX, Idx),
+            put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_WRITTEN_INDEX, LWIdx),
             {ok, State2#?MODULE{last_index = Idx,
                                 last_term = Term,
                                 cache = Cache,
@@ -401,7 +406,8 @@ handle_event({written, {FromIdx, _ToIdx, _Term}},
     %% Just drop the event in this case as it is stale
     {State, []};
 handle_event({written, {FromIdx, ToIdx0, Term}},
-             #?MODULE{last_written_index_term = {LastWrittenIdx0,
+             #?MODULE{cfg = Cfg,
+                      last_written_index_term = {LastWrittenIdx0,
                                                  LastWrittenTerm0},
                       last_index = LastIdx,
                       snapshot_state = SnapState} = State0)
@@ -416,6 +422,7 @@ handle_event({written, {FromIdx, ToIdx0, Term}},
     ToIdx = min(ToIdx0, LastIdx),
     case fetch_term(ToIdx, State0) of
         {Term, State} when is_integer(Term) ->
+            ok = put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_WRITTEN_INDEX, ToIdx),
             {State#?MODULE{last_written_index_term = {ToIdx, Term}},
              %% delaying truncate_cache until the next event allows any entries
              %% that became committed to be read from cache rather than ETS
@@ -426,8 +433,10 @@ handle_event({written, {FromIdx, ToIdx0, Term}},
             % followers returning appending the entry and the leader committing
             % and processing a snapshot before the written event comes in.
             % ensure last_written_index_term does not go backwards
-            LastWrittenIdxTerm = {max(LastWrittenIdx0, ToIdx),
+            LastWrittenIdx = max(LastWrittenIdx0, ToIdx),
+            LastWrittenIdxTerm = {LastWrittenIdx,
                                   max(LastWrittenTerm0, Term)},
+            ok = put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_WRITTEN_INDEX, LastWrittenIdx),
             {State#?MODULE{last_written_index_term = LastWrittenIdxTerm},
              [{next_event, {ra_log_event, {truncate_cache, FromIdx, ToIdx}}}]};
         {OtherTerm, State} ->
@@ -491,13 +500,15 @@ handle_event({segments, Tid, NewSegs},
             {State, log_update_effects(Readers, Pid, State)}
     end;
 handle_event({snapshot_written, {SnapIdx, _} = Snap},
-             #?MODULE{first_index = FstIdx,
+             #?MODULE{cfg = Cfg,
+                      first_index = FstIdx,
                       snapshot_state = SnapState0} = State0)
 %% only update snapshot if it is newer than the last snapshot
   when SnapIdx >= FstIdx ->
     % delete any segments outside of first_index
     {State, Effects0} = delete_segments(SnapIdx, State0),
     SnapState = ra_snapshot:complete_snapshot(Snap, SnapState0),
+    put_counter(Cfg, ?C_RA_SVR_METRIC_SNAPSHOT_INDEX, SnapIdx),
     %% delete old snapshot files
     %% This is done as an effect
     %% so that if an old snapshot is still being replicated
@@ -535,7 +546,7 @@ next_index(#?MODULE{last_index = LastIdx}) ->
     LastIdx + 1.
 
 -spec fetch(ra_index(), state()) ->
-    {maybe(log_entry()), state()}.
+    {option(log_entry()), state()}.
 fetch(Idx, State0) ->
     case fold(Idx, Idx, fun(E, Acc) -> [E | Acc] end, [], State0) of
         {[], State} ->
@@ -545,7 +556,7 @@ fetch(Idx, State0) ->
     end.
 
 -spec fetch_term(ra_index(), state()) ->
-    {'maybe'(ra_term()), state()}.
+    {option(ra_term()), state()}.
 fetch_term(Idx, #?MODULE{last_index = LastIdx,
                          first_index = FirstIdx} = State0)
   when Idx < FirstIdx orelse Idx > LastIdx ->
@@ -573,6 +584,9 @@ install_snapshot({SnapIdx, _} = IdxTerm, SnapState,
                  #?MODULE{cfg = Cfg,
                          cache = Cache} = State0) ->
     ok = incr_counter(Cfg, ?C_RA_LOG_SNAPSHOTS_INSTALLED, 1),
+    ok = put_counter(Cfg, ?C_RA_SVR_METRIC_SNAPSHOT_INDEX, SnapIdx),
+    put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_INDEX, SnapIdx),
+    put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_WRITTEN_INDEX, SnapIdx),
     {State, Effs} = delete_segments(SnapIdx, State0),
     {State#?MODULE{snapshot_state = SnapState,
                    first_index = SnapIdx + 1,
@@ -584,7 +598,7 @@ install_snapshot({SnapIdx, _} = IdxTerm, SnapState,
      Effs}.
 
 -spec recover_snapshot(State :: state()) ->
-    'maybe'({ra_snapshot:meta(), term()}).
+    option({ra_snapshot:meta(), term()}).
 recover_snapshot(#?MODULE{snapshot_state = SnapState}) ->
     case ra_snapshot:recover(SnapState) of
         {ok, Meta, MacState} ->
@@ -593,7 +607,7 @@ recover_snapshot(#?MODULE{snapshot_state = SnapState}) ->
             undefined
     end.
 
--spec snapshot_index_term(State :: state()) -> 'maybe'(ra_idxterm()).
+-spec snapshot_index_term(State :: state()) -> option(ra_idxterm()).
 snapshot_index_term(#?MODULE{snapshot_state = SS}) ->
     ra_snapshot:current(SS).
 
@@ -623,7 +637,9 @@ update_release_cursor0(Idx, Cluster, MacVersion, MacState,
                        #?MODULE{cfg = #cfg{snapshot_interval = SnapInter},
                                 reader = Reader,
                                 snapshot_state = SnapState} = State0) ->
-    ClusterServerIds = maps:keys(Cluster),
+    ClusterServerIds = maps:map(fun (_, V) ->
+                                        maps:with([voter_status], V)
+                                end, Cluster),
     SnapLimit = case ra_snapshot:current(SnapState) of
                     undefined -> SnapInter;
                     {I, _} -> I + SnapInter
@@ -837,6 +853,7 @@ wal_truncate_write(#?MODULE{cfg = #cfg{uid = UId,
     % and that prior entries should be considered stale
     ok = ra_log_wal:truncate_write({UId, self()}, Wal, Idx, Term, Cmd),
     ok = incr_counter(Cfg, ?C_RA_LOG_WRITE_OPS, 1),
+    ok = put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_INDEX, Idx),
     State#?MODULE{last_index = Idx, last_term = Term,
                   cache = ra_log_cache:add(Entry, Cache)}.
 
@@ -847,6 +864,7 @@ wal_write(#?MODULE{cfg = #cfg{uid = UId,
     case ra_log_wal:write({UId, self()}, Wal, Idx, Term, Cmd) of
         ok ->
             ok = incr_counter(Cfg, ?C_RA_LOG_WRITE_OPS, 1),
+            put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_INDEX, Idx),
             State#?MODULE{last_index = Idx, last_term = Term,
                           cache = ra_log_cache:add(Entry, Cache)};
         {error, wal_down} ->
@@ -869,6 +887,7 @@ wal_write_batch(#?MODULE{cfg = #cfg{uid = UId,
     case ra_log_wal:write_batch(Wal, lists:reverse(WalCommands)) of
         ok ->
             ok = incr_counter(Cfg, ?C_RA_LOG_WRITE_OPS, Num),
+            put_counter(Cfg, ?C_RA_SVR_METRIC_LAST_INDEX, LastIdx),
             State#?MODULE{last_index = LastIdx,
                           last_term = LastTerm,
                           cache = Cache};
@@ -1030,6 +1049,11 @@ incr_counter(#cfg{counter = Cnt}, Ix, N) when Cnt =/= undefined ->
 incr_counter(#cfg{counter = undefined}, _Ix, _N) ->
     ok.
 
+put_counter(#cfg{counter = Cnt}, Ix, N) when Cnt =/= undefined ->
+    counters:put(Cnt, Ix, N);
+put_counter(#cfg{counter = undefined}, _Ix, _N) ->
+    ok.
+
 server_data_dir(Dir, UId) ->
     Me = ra_lib:to_list(UId),
     filename:join(Dir, Me).
