@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2017-2022 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2017-2023 Broadcom. All Rights Reserved. The term Broadcom refers to Broadcom Inc. and/or its subsidiaries.
 %%
 %% @hidden
 -module(ra_log_segment).
@@ -46,7 +46,7 @@
               max_count = ?SEGMENT_MAX_ENTRIES :: non_neg_integer(),
               max_pending = ?SEGMENT_MAX_PENDING :: non_neg_integer(),
               filename :: file:filename_all(),
-              fd :: 'maybe'(file:io_device()),
+              fd :: option(file:io_device()),
               index_size :: pos_integer(),
               access_pattern :: sequential | random,
               mode = append :: read | append,
@@ -59,8 +59,8 @@
          data_start :: pos_integer(),
          data_offset :: pos_integer(),
          data_write_offset :: pos_integer(),
-         index = undefined :: maybe(ra_segment_index()),
-         range :: 'maybe'({ra_index(), ra_index()}),
+         index = undefined :: option(ra_segment_index()),
+         range :: option({ra_index(), ra_index()}),
          pending_data = [] :: iodata(),
          pending_index = [] :: iodata(),
          pending_count = 0 :: non_neg_integer(),
@@ -331,7 +331,7 @@ map_get_(Key, Map) when is_map_key(Key, Map) ->
 map_get_(Key, _Map) ->
     exit({missing_key, Key}).
 
--spec term_query(state(), Idx :: ra_index()) -> 'maybe'(ra_term()).
+-spec term_query(state(), Idx :: ra_index()) -> option(ra_term()).
 term_query(#state{index = Index}, Idx) ->
     case Index of
         #{Idx := {Term, _, _, _}} ->
@@ -367,7 +367,7 @@ fold0(Cfg, Cache0, Idx, FinalIdx, Index, Fun, AccFun, Acc0) ->
                   Cfg#cfg.filename})
     end.
 
--spec range(state()) -> maybe({ra_index(), ra_index()}).
+-spec range(state()) -> option({ra_index(), ra_index()}).
 range(#state{range = Range}) ->
     Range.
 
@@ -379,7 +379,7 @@ max_count(#state{cfg = #cfg{max_count = Max}}) ->
 filename(#state{cfg = #cfg{filename = Fn}}) ->
     Fn.
 
--spec segref(state()) -> 'maybe'(ra_log:segment_ref()).
+-spec segref(state()) -> option(ra_log:segment_ref()).
 segref(#state{range = undefined}) ->
     undefined;
 segref(#state{range = {Start, End},
@@ -452,7 +452,7 @@ dump_index(File) ->
 dump(File) ->
     {ok, S0} = open(File, #{mode => read}),
     {Idx, Last} = range(S0),
-    L = fold(S0, Idx, Last - Idx + 1, fun erlang:binary_to_term/1,
+    L = fold(S0, Idx, Last, fun erlang:binary_to_term/1,
              fun (E, A) -> [E | A] end, []),
     close(S0),
     lists:reverse(L).
