@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2017-2022 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2017-2023 Broadcom. All Rights Reserved. The term Broadcom refers to Broadcom Inc. and/or its subsidiaries.
 %%
 %% @hidden
 -module(ra_server_proc).
@@ -85,7 +85,7 @@
 -type gen_statem_start_ret() :: {ok, pid()} | ignore | {error, term()}.
 
 %% ra_event types
--type ra_event_reject_detail() :: {not_leader, Leader :: 'maybe'(ra_server_id()),
+-type ra_event_reject_detail() :: {not_leader, Leader :: option(ra_server_id()),
                                    ra_server:command_correlation()}.
 
 -type ra_event_body() ::
@@ -183,7 +183,9 @@ log_fold(ServerId, Fun, InitialState, Timeout) ->
 -spec state_query(server_loc(),
                   all |
                   overview |
+                  voters |
                   members |
+                  members_info |
                   initial_members |
                   machine, timeout()) ->
     ra_leader_call_ret(term()).
@@ -193,7 +195,9 @@ state_query(ServerLoc, Spec, Timeout) ->
 -spec local_state_query(server_loc(),
                         all |
                         overview |
+                        voters |
                         members |
+                        members_info |
                         initial_members |
                         machine, timeout()) ->
     ra_local_call_ret(term()).
@@ -270,6 +274,8 @@ init(Config) ->
 
 do_init(#{id := Id,
           cluster_name := ClusterName} = Config0) ->
+    Key = ra_lib:ra_server_id_to_local_name(Id),
+    true = ets:insert(ra_state, {Key, init, unknown}),
     process_flag(trap_exit, true),
     Config = #{counter := Counter,
                system_config := SysConf} = maps:merge(config_defaults(Id),
@@ -283,7 +289,6 @@ do_init(#{id := Id,
     UId = ra_server:uid(ServerState),
     % ensure ra_directory has the new pid
     #{names := Names} = SysConf,
-    Key = ra_lib:ra_server_id_to_local_name(Id),
     ok = ra_directory:register_name(Names, UId, self(),
                                     maps:get(parent, Config, undefined), Key,
                                     ClusterName),
@@ -365,6 +370,7 @@ recover(_, _, State) ->
 %% effects post recovery
 recovered(enter, OldState, State0) ->
     {State, Actions} = handle_enter(?FUNCTION_NAME, OldState, State0),
+    ok = record_cluster_change(State),
     {keep_state, State, Actions};
 recovered(internal, next, #state{server_state = ServerState} = State) ->
     true = erlang:garbage_collect(),
@@ -373,7 +379,7 @@ recovered(internal, next, #state{server_state = ServerState} = State) ->
 
 leader(enter, OldState, State0) ->
     {State, Actions} = handle_enter(?FUNCTION_NAME, OldState, State0),
-    ok = record_leader_change(id(State0), State0),
+    ok = record_cluster_change(State),
     %% TODO: reset refs?
     {keep_state, State#state{leader_last_seen = undefined,
                              pending_notifys = #{},
@@ -387,17 +393,13 @@ leader(EventType, {local_call, Msg}, State) ->
 leader(EventType, {leader_cast, Msg}, State) ->
     leader(EventType, Msg, State);
 leader(EventType, {command, normal, {CmdType, Data, ReplyMode}},
-       #state{conf = Conf,
-              server_state = ServerState0} = State0) ->
+       #state{conf = Conf} = State0) ->
     case validate_reply_mode(ReplyMode) of
         ok ->
             %% normal priority commands are written immediately
             Cmd = make_command(CmdType, EventType, Data, ReplyMode),
-            {leader, ServerState, Effects} =
-                ra_server:handle_leader({command, Cmd}, ServerState0),
-            {State, Actions} =
-                ?HANDLE_EFFECTS(Effects, EventType,
-                                State0#state{server_state = ServerState}),
+            {leader, State1, Effects} = handle_leader({command, Cmd}, State0),
+            {State, Actions} = ?HANDLE_EFFECTS(Effects, EventType, State1),
             {keep_state, State, Actions};
         Error ->
             ok = incr_counter(Conf, ?C_RA_SRV_INVALID_REPLY_MODE_COMMANDS, 1),
@@ -448,17 +450,13 @@ leader(EventType, {aux_command, Cmd}, State0) ->
     {keep_state, State#state{server_state = ServerState}, Actions};
 leader(EventType, flush_commands,
        #state{conf = #conf{flush_commands_size = Size},
-              server_state = ServerState0,
               low_priority_commands = Delayed0} = State0) ->
 
     {Commands, Delayed} = ra_ets_queue:take(Size, Delayed0),
     %% write a batch of delayed commands
-    {leader, ServerState, Effects} =
-        ra_server:handle_leader({commands, Commands}, ServerState0),
+    {leader, State1, Effects} = handle_leader({commands, Commands}, State0),
 
-    {State, Actions} =
-        ?HANDLE_EFFECTS(Effects, EventType,
-                        State0#state{server_state = ServerState}),
+    {State, Actions} = ?HANDLE_EFFECTS(Effects, EventType, State1),
     case ra_ets_queue:len(Delayed) of
         0 ->
             ok;
@@ -509,16 +507,11 @@ leader(_, tick_timeout, State0) ->
     State = send_applied_notifications(State2, #{}),
     {keep_state, handle_tick_metrics(State),
      set_tick_timer(State, Actions)};
-leader({timeout, Name}, machine_timeout,
-       #state{server_state = ServerState0,
-              conf = #conf{}} = State0) ->
+leader({timeout, Name}, machine_timeout, State0) ->
     % the machine timer timed out, add a timeout message
     Cmd = make_command('$usr', cast, {timeout, Name}, noreply),
-    {leader, ServerState, Effects} = ra_server:handle_leader({command, Cmd},
-                                                             ServerState0),
-    {State, Actions} = ?HANDLE_EFFECTS(Effects, cast,
-                                       State0#state{server_state =
-                                                        ServerState}),
+    {leader, State1, Effects} = handle_leader({command, Cmd}, State0),
+    {State, Actions} = ?HANDLE_EFFECTS(Effects, cast, State1),
     {keep_state, State, Actions};
 leader({call, From}, trigger_election, State) ->
     {keep_state, State, [{reply, From, ok}]};
@@ -783,11 +776,19 @@ follower(_, tick_timeout, State0) ->
      set_tick_timer(State, Actions)};
 follower({call, From}, {log_fold, Fun, Term}, State) ->
     fold_log(From, Fun, Term, State);
-follower(EventType, Msg, State0) ->
+follower(EventType, Msg, #state{conf = #conf{name = Name},
+                                server_state = SS0} = State0) ->
     case handle_follower(Msg, State0) of
         {follower, State1, Effects} ->
             {State2, Actions} = ?HANDLE_EFFECTS(Effects, EventType, State1),
-            State = follower_leader_change(State0, State2),
+            State = #state{server_state = SS} = follower_leader_change(State0, State2),
+            Membership0 = ra_server:get_membership(SS0),
+            case ra_server:get_membership(SS) of
+                Membership0 ->
+                    ok;
+                Membership ->
+                    true = ets:update_element(ra_state, Name, {3, Membership})
+            end,
             {keep_state, State, Actions};
         {pre_vote, State1, Effects} ->
             {State, Actions} = ?HANDLE_EFFECTS(Effects, EventType, State1),
@@ -969,10 +970,8 @@ terminate(Reason, StateName,
     Parent = ra_directory:where_is_parent(Names, UId),
     case Reason of
         {shutdown, delete} ->
-            catch ra_leaderboard:clear(ClusterName),
             catch ra_directory:unregister_name(Names, UId),
             catch ra_log_meta:delete_sync(MetaName, UId),
-            catch ets:delete(ra_state, UId),
             catch ra_counters:delete(Id),
             Self = self(),
             %% we have to terminate the child spec from the supervisor as it
@@ -993,8 +992,22 @@ terminate(Reason, StateName,
 
         _ -> ok
     end,
+    catch ra_leaderboard:clear(ClusterName),
     _ = ets:delete(ra_metrics, MetricsKey),
     _ = ets:delete(ra_state, Key),
+    ok;
+%% This occurs if there is a crash in the init callback of the ra_machine,
+%% before a state has been built
+terminate(Reason, StateName, #{id := Id} = Config) ->
+    LogId = maps:get(friendly_name, Config,
+                     lists:flatten(io_lib:format("~w", [Id]))),
+    ?DEBUG("~ts: terminating with ~w in state ~w",
+           [LogId, Reason, StateName]),
+    ok;
+%% Unknown reason for termination
+terminate(Reason, StateName, State) ->
+    ?DEBUG("Terminating with ~w in state ~w with state ~w",
+           [Reason, StateName, State]),
     ok.
 
 code_change(_OldVsn, StateName, State, _Extra) ->
@@ -1028,7 +1041,8 @@ format_status(Opt, [_PDict, StateName,
 handle_enter(RaftState, OldRaftState,
              #state{conf = #conf{name = Name},
                     server_state = ServerState0} = State) ->
-    true = ets:insert(ra_state, {Name, RaftState}),
+    Membership = ra_server:get_membership(ServerState0),
+    true = ets:insert(ra_state, {Name, RaftState, Membership}),
     {ServerState, Effects} = ra_server:handle_state_enter(RaftState,
                                                           ServerState0),
     case RaftState == leader orelse OldRaftState == leader of
@@ -1051,6 +1065,7 @@ handle_leader(Msg, #state{server_state = ServerState0} = State0) ->
         {NextState, ServerState, Effects}  ->
             State = State0#state{server_state =
                                  ra_server:persist_last_applied(ServerState)},
+            maybe_record_cluster_change(State0, State),
             {NextState, State, Effects};
         OtherErr ->
             ?ERR("handle_leader err ~p", [OtherErr]),
@@ -1077,8 +1092,11 @@ handle_candidate(Msg, State) ->
 handle_pre_vote(Msg, State) ->
     handle_raft_state(?FUNCTION_NAME, Msg, State).
 
-handle_follower(Msg, State) ->
-    handle_raft_state(?FUNCTION_NAME, Msg, State).
+handle_follower(Msg, State0) ->
+    Ret = handle_raft_state(?FUNCTION_NAME, Msg, State0),
+    {_NextState, State, _Effects} = Ret,
+    maybe_record_cluster_change(State0, State),
+    Ret.
 
 handle_receive_snapshot(Msg, State) ->
     handle_raft_state(?FUNCTION_NAME, Msg, State).
@@ -1461,8 +1479,7 @@ follower_leader_change(Old, #state{pending_commands = Pending,
             LeaderNode = ra_lib:ra_server_id_node(NewLeader),
             ok = aten_register(LeaderNode),
             OldLeaderNode = ra_lib:ra_server_id_node(OldLeader),
-            ok = aten:unregister(OldLeaderNode),
-            ok = record_leader_change(NewLeader, New),
+            _ = aten:unregister(OldLeaderNode),
             % leader has either changed or just been set
             ?INFO("~ts: detected a new leader ~w in term ~b",
                   [log_id(New), NewLeader, current_term(New)]),
@@ -1477,7 +1494,12 @@ aten_register(Node) ->
     case node() of
         Node -> ok;
         _ ->
-            aten:register(Node)
+            case aten:register(Node) of
+                ignore ->
+                    ok;
+                Res ->
+                    Res
+            end
     end.
 
 swap_monitor(MRef, L) ->
@@ -1499,7 +1521,9 @@ gen_statem_safe_call(ServerId, Msg, Timeout) ->
          exit:{noproc, _} ->
             {error, noproc};
          exit:{{nodedown, _}, _} ->
-            {error, nodedown}
+            {error, nodedown};
+         exit:{shutdown, _} ->
+            {error, shutdown}
     end.
 
 do_state_query(all, State) -> State;
@@ -1507,8 +1531,67 @@ do_state_query(overview, State) ->
     ra_server:overview(State);
 do_state_query(machine, #{machine_state := MacState}) ->
     MacState;
+do_state_query(voters, #{cluster := Cluster}) ->
+    Vs = maps:fold(fun(K, V, Acc) ->
+                           case maps:get(voter_status, V, undefined) of
+                               undefined -> [K|Acc];
+                               S -> case maps:get(membership, S, undefined) of
+                                        undefined -> [K|Acc];
+                                        voter -> [K|Acc];
+                                        _ -> Acc
+                                    end
+                           end
+              end, [], Cluster),
+    Vs;
+do_state_query(leader, State) ->
+    maps:get(leader_id, State, undefined);
 do_state_query(members, #{cluster := Cluster}) ->
     maps:keys(Cluster);
+do_state_query(members_info, #{cfg := #cfg{id = Self}, cluster := Cluster,
+               leader_id := Self, query_index := QI, commit_index := CI,
+               membership := Membership}) ->
+    maps:map(fun(Id, Peer) ->
+                 case {Id, Peer} of
+                     {Self, Peer = #{voter_status := VoterStatus}} ->
+                         %% For completeness sake, preserve `target`
+                         %% of once promoted leader.
+                         #{next_index => CI+1,
+                           match_index => CI,
+                           query_index => QI,
+                           status => normal,
+                           voter_status => VoterStatus#{membership => Membership}};
+                     {Self, _} ->
+                         #{next_index => CI+1,
+                           match_index => CI,
+                           query_index => QI,
+                           status => normal,
+                           voter_status => #{membership => Membership}};
+                     {_, Peer = #{voter_status := _}} ->
+                          Peer;
+                     {_, Peer} ->
+                          %% Initial cluster members have no voter_status.
+                          Peer#{voter_status => #{membership => voter}}
+                 end
+             end, Cluster);
+do_state_query(members_info, #{cfg := #cfg{id = Self}, cluster := Cluster,
+               query_index := QI, commit_index := CI,
+               membership := Membership}) ->
+    %% Followers do not have sufficient information,
+    %% bail out and send whatever we have.
+    maps:map(fun(Id, Peer) ->
+                case {Id, Peer} of
+                    {Self, #{voter_status := VS}} ->
+                        #{match_index => CI,
+                          query_index => QI,
+                          voter_status => VS#{membership => Membership}};
+                    {Self, _} ->
+                        #{match_index => CI,
+                          query_index => QI,
+                          voter_status => #{membership => Membership}};
+                    _ ->
+                        #{}
+                end
+             end, Cluster);
 do_state_query(initial_members, #{log := Log}) ->
     case ra_log:read_config(Log) of
         {ok, #{initial_members := InitialMembers}} ->
@@ -1525,7 +1608,7 @@ config_defaults(ServerId) ->
       install_snap_rpc_timeout => ?INSTALL_SNAP_RPC_TIMEOUT,
       await_condition_timeout => ?DEFAULT_AWAIT_CONDITION_TIMEOUT,
       initial_members => [],
-      counter => ra_counters:new(ServerId, ?RA_COUNTER_FIELDS),
+      counter => ra_counters:new(ServerId, {persistent_term, ?FIELDSPEC_KEY}),
       system_config => ra_system:default_config()
      }.
 
@@ -1714,14 +1797,16 @@ handle_tick_metrics(State) ->
     _ = ets:insert(ra_metrics, Metrics),
     State.
 
-can_execute_locally(RaftState, TargetNode, State) ->
+can_execute_locally(RaftState, TargetNode,
+                    #state{server_state = ServerState} = State) ->
+    Membership = ra_server:get_membership(ServerState),
     case RaftState of
-        follower ->
+        follower when Membership == voter ->
             TargetNode == node();
         leader when TargetNode =/= node() ->
             %% We need to evaluate whether to send the message.
             %% Only send if there isn't a local node for the target pid.
-            Members = do_state_query(members, State#state.server_state),
+            Members = do_state_query(voters, State#state.server_state),
             not lists:any(fun ({_, N}) -> N == TargetNode end, Members);
         leader ->
             true;
@@ -1777,11 +1862,25 @@ handle_process_down(Pid, Info, RaftState,
                                     monitors = Monitors}),
     {keep_state, State, Actions}.
 
-record_leader_change(Leader, #state{conf = #conf{cluster_name = ClusterName},
-                                    server_state = ServerState}) ->
+maybe_record_cluster_change(#state{conf = #conf{cluster_name = ClusterName},
+                                   server_state = ServerStateA},
+                            #state{server_state = ServerStateB}) ->
+    LeaderA = ra_server:leader_id(ServerStateA),
+    LeaderB = ra_server:leader_id(ServerStateB),
+    if (map_get(cluster_index_term, ServerStateA) =/=
+        map_get(cluster_index_term, ServerStateB) orelse
+        LeaderA =/= LeaderB) ->
+            MembersB = do_state_query(members, ServerStateB),
+            ok = ra_leaderboard:record(ClusterName, LeaderB, MembersB);
+        true ->
+            ok
+    end.
+
+record_cluster_change(#state{conf = #conf{cluster_name = ClusterName},
+                             server_state = ServerState}) ->
+    Leader = do_state_query(leader, ServerState),
     Members = do_state_query(members, ServerState),
-    ok = ra_leaderboard:record(ClusterName, Leader, Members),
-    ok.
+    ok = ra_leaderboard:record(ClusterName, Leader, Members).
 
 incr_counter(#conf{counter = Cnt}, Ix, N) when Cnt =/= undefined ->
     counters:add(Cnt, Ix, N);
