@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_amqqueue_process).
@@ -89,7 +89,7 @@
             mirroring_policy_version = 0,
             %% running | flow | idle
             status,
-            %% true | false
+            %% boolean()
             single_active_consumer_on
            }).
 
@@ -212,7 +212,7 @@ init_it2(Recover, From, State = #q{q                   = Q,
                                        backing_queue_state = BQS}),
                     notify_decorators(startup, State),
                     rabbit_event:notify(queue_created,
-                                        infos(?CREATION_EVENT_KEYS, State1)),
+                                        queue_created_infos(State1)),
                     rabbit_event:if_enabled(State1, #q.stats_timer,
                                             fun() -> emit_stats(State1) end),
                     noreply(State1);
@@ -543,7 +543,7 @@ ensure_ttl_timer(undefined, State) ->
     State;
 ensure_ttl_timer(Expiry, State = #q{ttl_timer_ref       = undefined,
                                     args_policy_version = Version}) ->
-    After = (case Expiry - os:system_time(micro_seconds) of
+    After = (case Expiry - os:system_time(microsecond) of
                  V when V > 0 -> V + 999; %% always fire later
                  _            -> 0
              end) div 1000,
@@ -611,20 +611,20 @@ send_or_record_confirm(#delivery{confirm    = false}, State) ->
 send_or_record_confirm(#delivery{confirm    = true,
                                  sender     = SenderPid,
                                  msg_seq_no = MsgSeqNo,
-                                 message    = #basic_message {
-                                   is_persistent = true,
-                                   id            = MsgId}},
+                                 message    = Message
+                                },
                        State = #q{q                 = Q,
-                                  msg_id_to_channel = MTC})
-  when ?amqqueue_is_durable(Q) ->
-    MTC1 = maps:put(MsgId, {SenderPid, MsgSeqNo}, MTC),
-    {eventually, State#q{msg_id_to_channel = MTC1}};
-send_or_record_confirm(#delivery{confirm    = true,
-                                 sender     = SenderPid,
-                                 msg_seq_no = MsgSeqNo},
-                       #q{q = Q} = State) ->
-    confirm_to_sender(SenderPid, amqqueue:get_name(Q), [MsgSeqNo]),
-    {immediately, State}.
+                                  msg_id_to_channel = MTC}) ->
+    Persistent = mc:is_persistent(Message),
+    MsgId = mc:get_annotation(id, Message),
+    case Persistent  of
+        true when ?amqqueue_is_durable(Q) ->
+            MTC1 = maps:put(MsgId, {SenderPid, MsgSeqNo}, MTC),
+            {eventually, State#q{msg_id_to_channel = MTC1}};
+        _ ->
+            confirm_to_sender(SenderPid, amqqueue:get_name(Q), [MsgSeqNo]),
+            {immediately, State}
+    end.
 
 %% This feature was used by `rabbit_amqqueue_process` and
 %% `rabbit_mirror_queue_slave` up-to and including RabbitMQ 3.7.x. It is
@@ -641,7 +641,8 @@ send_mandatory(#delivery{mandatory  = true,
 discard(#delivery{confirm = Confirm,
                   sender  = SenderPid,
                   flow    = Flow,
-                  message = #basic_message{id = MsgId}}, BQ, BQS, MTC, QName) ->
+                  message = Msg}, BQ, BQS, MTC, QName) ->
+    MsgId = mc:get_annotation(id, Msg),
     MTC1 = case Confirm of
                true  -> confirm_messages([MsgId], MTC, QName);
                false -> MTC
@@ -809,12 +810,13 @@ send_reject_publish(#delivery{confirm = true,
                               sender = SenderPid,
                               flow = Flow,
                               msg_seq_no = MsgSeqNo,
-                              message = #basic_message{id = MsgId}},
+                              message = Msg},
                       _Delivered,
                       State = #q{ q = Q,
                                   backing_queue = BQ,
                                   backing_queue_state = BQS,
                                   msg_id_to_channel   = MTC}) ->
+    MsgId = mc:get_annotation(id, Msg),
     ok = rabbit_classic_queue:send_rejection(SenderPid,
                                              amqqueue:get_name(Q), MsgSeqNo),
 
@@ -834,9 +836,8 @@ will_overflow(#delivery{message = Message},
                  backing_queue_state = BQS}) ->
     ExpectedQueueLength = BQ:len(BQS) + 1,
 
-    #basic_message{content = #content{payload_fragments_rev = PFR}} = Message,
-    MessageSize = iolist_size(PFR),
-    ExpectedQueueSizeBytes = BQ:info(message_bytes_ready, BQS) + MessageSize,
+    {_, PayloadSize} = mc:size(Message),
+    ExpectedQueueSizeBytes = BQ:info(message_bytes_ready, BQS) + PayloadSize,
 
     ExpectedQueueLength > MaxLen orelse ExpectedQueueSizeBytes > MaxBytes.
 
@@ -855,6 +856,9 @@ requeue_and_run(AckTags, State = #q{backing_queue       = BQ,
 
 fetch(AckRequired, State = #q{backing_queue       = BQ,
                               backing_queue_state = BQS}) ->
+    %% @todo We should first drop expired messages then fetch
+    %%       the message, not the other way around. Otherwise
+    %%       we will send expired messages at times.
     {Result, BQS1} = BQ:fetch(AckRequired, BQS),
     State1 = drop_expired_msgs(State#q{backing_queue_state = BQS1}),
     {Result, maybe_send_drained(Result =:= empty, State1)}.
@@ -909,7 +913,7 @@ handle_ch_down(DownPid, State = #q{consumers                 = Consumers,
             {ok, State1};
         {ChAckTags, ChCTags, Consumers1} ->
             QName = qname(State1),
-            [emit_consumer_deleted(DownPid, CTag, QName, ?INTERNAL_USER) || CTag <- ChCTags],
+            [rabbit_core_metrics:consumer_deleted(DownPid, CTag, QName) || CTag <- ChCTags],
             Holder1 = new_single_active_consumer_after_channel_down(DownPid, Holder, SingleActiveConsumerOn, Consumers1),
             State2 = State1#q{consumers          = Consumers1,
                               active_consumer    = Holder1},
@@ -974,21 +978,18 @@ subtract_acks(ChPid, AckTags, State = #q{consumers = Consumers}, Fun) ->
                                    run_message_queue(true, Fun(State1))
     end.
 
-message_properties(Message = #basic_message{content = Content},
-                   Confirm, #q{ttl = TTL}) ->
-    #content{payload_fragments_rev = PFR} = Content,
+message_properties(Message, Confirm, #q{ttl = TTL}) ->
+    {_, Size} = mc:size(Message),
     #message_properties{expiry           = calculate_msg_expiry(Message, TTL),
                         needs_confirming = Confirm == eventually,
-                        size             = iolist_size(PFR)}.
+                        size             = Size}.
 
-calculate_msg_expiry(#basic_message{content = Content}, TTL) ->
-    #content{properties = Props} =
-        rabbit_binary_parser:ensure_content_decoded(Content),
-    %% We assert that the expiration must be valid - we check in the channel.
-    {ok, MsgTTL} = rabbit_basic:parse_expiration(Props),
+calculate_msg_expiry(Msg, TTL) ->
+    MsgTTL = mc:ttl(Msg),
     case lists:min([TTL, MsgTTL]) of
         undefined -> undefined;
-        T         -> os:system_time(micro_seconds) + T * 1000
+        T ->
+            os:system_time(microsecond) + T * 1000
     end.
 
 %% Logically this function should invoke maybe_send_drained/2.
@@ -999,7 +1000,7 @@ calculate_msg_expiry(#basic_message{content = Content}, TTL) ->
 drop_expired_msgs(State) ->
     case is_empty(State) of
         true  -> State;
-        false -> drop_expired_msgs(os:system_time(micro_seconds),
+        false -> drop_expired_msgs(os:system_time(microsecond),
                                    State)
     end.
 
@@ -1267,7 +1268,7 @@ prioritise_cast(Msg, _Len, State) ->
 
 consumer_bias(#q{backing_queue = BQ, backing_queue_state = BQS}, Low, High) ->
     case BQ:msg_rates(BQS) of
-        {0.0,          _} -> Low;
+        {Ingress,      _} when Ingress =:= +0.0 orelse Ingress =:= -0.0 -> Low;
         {Ingress, Egress} when Egress / Ingress < ?CONSUMER_BIAS_RATIO -> High;
         {_,            _} -> Low
     end.
@@ -1658,7 +1659,7 @@ handle_cast({credit, ChPid, CTag, Credit, Drain},
 % the correct values once the management agent has started
 handle_cast({force_event_refresh, Ref},
             State = #q{consumers = Consumers}) ->
-    rabbit_event:notify(queue_created, infos(?CREATION_EVENT_KEYS, State), Ref),
+    rabbit_event:notify(queue_created, queue_created_infos(State), Ref),
     QName = qname(State),
     AllConsumers = rabbit_queue_consumers:all(Consumers),
     rabbit_log:debug("Queue ~ts forced to re-emit events, consumers: ~tp", [rabbit_misc:rs(QName), AllConsumers]),
@@ -1677,13 +1678,37 @@ handle_cast(policy_changed, State = #q{q = Q0}) ->
     Name = amqqueue:get_name(Q0),
     %% We depend on the #q.q field being up to date at least WRT
     %% policy (but not mirror pids) in various places, so when it
-    %% changes we go and read it from Mnesia again.
+    %% changes we go and read it from the database again.
     %%
     %% This also has the side effect of waking us up so we emit a
     %% stats event - so event consumers see the changed policy.
     {ok, Q} = rabbit_amqqueue:lookup(Name),
     noreply(process_args_policy(State#q{q = Q}));
 
+handle_cast({policy_changed, Q0}, State) ->
+    Name = amqqueue:get_name(Q0),
+    PolicyVersion0 = amqqueue:get_policy_version(Q0),
+    %% We depend on the #q.q field being up to date at least WRT
+    %% policy (but not mirror pids) in various places, so when it
+    %% changes we go and read it from the database again.
+    %%
+    %% This also has the side effect of waking us up so we emit a
+    %% stats event - so event consumers see the changed policy.
+    {ok, Q} = rabbit_amqqueue:lookup(Name),
+    PolicyVersion = amqqueue:get_policy_version(Q),
+    case PolicyVersion >= PolicyVersion0 of
+        true ->
+            noreply(process_args_policy(State#q{q = Q}));
+        false ->
+            %% Update just the policy, as pids and mirrors could have been
+            %% updated simultaneously. A testcase on the `confirm_rejects_SUITE`
+            %% fails consistently if the internal state is updated directly to `Q0`.
+            Q1 = amqqueue:set_policy(Q, amqqueue:get_policy(Q0)),
+            Q2 = amqqueue:set_operator_policy(Q1, amqqueue:get_operator_policy(Q0)),
+            Q3 = amqqueue:set_policy_version(Q2, PolicyVersion0),
+            noreply(process_args_policy(State#q{q = Q3}))
+    end;
+
 handle_cast({sync_start, _, _}, State = #q{q = Q}) ->
     Name = amqqueue:get_name(Q),
     %% Only a mirror should receive this, it means we are a duplicated master
@@ -1691,9 +1716,12 @@ handle_cast({sync_start, _, _}, State = #q{q = Q}) ->
       Name, "Stopping after receiving sync_start from another master", []),
     stop(State).
 
-handle_info({maybe_expire, Vsn}, State = #q{args_policy_version = Vsn}) ->
+handle_info({maybe_expire, Vsn}, State = #q{q = Q, expires = Expiry, args_policy_version = Vsn}) ->
     case is_unused(State) of
-        true  -> stop(State);
+        true  ->
+            QResource = rabbit_misc:rs(amqqueue:get_name(Q)),
+            rabbit_log_queue:debug("Deleting 'classic ~ts' on expiry after ~tp milliseconds", [QResource, Expiry]),
+            stop(State);
         false -> noreply(State#q{expiry_timer_ref = undefined})
     end;
 
@@ -1779,7 +1807,7 @@ handle_pre_hibernate(State = #q{backing_queue = BQ,
       State, #q.stats_timer,
       fun () -> emit_stats(State,
                            [{idle_since,
-                             os:system_time(milli_seconds)}])
+                             os:system_time(millisecond)}])
                 end),
     State1 = rabbit_event:stop_stats_timer(State#q{backing_queue_state = BQS3},
                                            #q.stats_timer),
@@ -1787,6 +1815,7 @@ handle_pre_hibernate(State = #q{backing_queue = BQ,
 
 format_message_queue(Opt, MQ) -> rabbit_misc:format_message_queue(Opt, MQ).
 
+%% TODO: this can be removed after 3.13
 format(Q) when ?is_amqqueue(Q) ->
     case rabbit_mirror_queue_misc:is_mirrored(Q) of
         false ->
@@ -1881,3 +1910,8 @@ update_state(State, Q) ->
                                    Q1 = amqqueue:set_state(Q0, State),
                                    amqqueue:set_decorators(Q1, Decorators)
                            end).
+
+queue_created_infos(State) ->
+    %% On the events API, we use long names for queue types
+    Keys = ?CREATION_EVENT_KEYS -- [type],
+    infos(Keys, State) ++ [{type, rabbit_classic_queue}].
