@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_auth_backend_internal).
@@ -22,7 +22,9 @@
          add_user_sans_validation/3, put_user/2, put_user/3,
          update_user/5,
          update_user_with_hash/5,
-         add_user_sans_validation/6]).
+         add_user_sans_validation/6,
+         add_user_with_pre_hashed_password_sans_validation/3
+]).
 
 -export([set_user_limits/3, clear_user_limits/3, is_over_connection_limit/1,
          is_over_channel_limit/1, get_user_limits/0, get_user_limits/1]).
@@ -30,6 +32,8 @@
 -export([user_info_keys/0, perms_info_keys/0,
          user_perms_info_keys/0, vhost_perms_info_keys/0,
          user_vhost_perms_info_keys/0, all_users/0,
+         user_topic_perms_info_keys/0, vhost_topic_perms_info_keys/0,
+         user_vhost_topic_perms_info_keys/0,
          list_users/0, list_users/2, list_permissions/0,
          list_user_permissions/1, list_user_permissions/3,
          list_topic_permissions/0,
@@ -37,11 +41,15 @@
          list_user_vhost_permissions/2,
          list_user_topic_permissions/1, list_vhost_topic_permissions/1, list_user_vhost_topic_permissions/2]).
 
--export([state_can_expire/0]).
+-export([expiry_timestamp/1]).
 
-%% for testing
 -export([hashing_module_for_user/1, expand_topic_permission/2]).
 
+-ifdef(TEST).
+-export([extract_user_permission_params/2,
+         extract_topic_permission_params/2]).
+-endif.
+
 -import(rabbit_data_coercion, [to_atom/1, to_list/1, to_binary/1]).
 
 %%----------------------------------------------------------------------------
@@ -96,12 +104,12 @@ user_login_authentication(Username, AuthProps) ->
               end);
         false ->
             case proplists:get_value(rabbit_auth_backend_internal, AuthProps, undefined) of
-                undefined -> exit({unknown_auth_props, Username, AuthProps});
+                undefined -> {refused, ?BLANK_PASSWORD_REJECTION_MESSAGE, [Username]};
                 _ -> internal_check_user_login(Username, fun(_) -> true end)
             end
     end.
 
-state_can_expire() -> false.
+expiry_timestamp(_) -> never.
 
 user_login_authorization(Username, _AuthProps) ->
     case user_login_authentication(Username, []) of
@@ -216,6 +224,10 @@ add_user(Username, Password, ActingUser, Limits, Tags) ->
     validate_and_alternate_credentials(Username, Password, ActingUser,
                                        add_user_sans_validation(Limits, Tags)).
 
+add_user_with_pre_hashed_password_sans_validation(Username, PasswordHash, ActingUser) ->
+    HashingAlgorithm = rabbit_password:hashing_mod(),
+    add_user_sans_validation(Username, PasswordHash, HashingAlgorithm, [], undefined, ActingUser).
+
 add_user_sans_validation(Username, Password, ActingUser) ->
     add_user_sans_validation(Username, Password, ActingUser, undefined, []).
 
@@ -240,14 +252,12 @@ add_user_sans_validation(Username, Password, ActingUser, Limits, Tags) ->
            end,
     add_user_sans_validation_in(Username, User, ConvertedTags, Limits, ActingUser).
 
-add_user_sans_validation(Username, PasswordHash, HashingAlgorithm, Tags, Limits, ActingUser) ->
+add_user_sans_validation(Username, PasswordHash, HashingMod, Tags, Limits, ActingUser) ->
     rabbit_log:debug("Asked to create a new user '~ts' with password hash", [Username]),
     ConvertedTags = [rabbit_data_coercion:to_atom(I) || I <- Tags],
-    HashingMod = rabbit_password:hashing_mod(),
     User0 = internal_user:create_user(Username, PasswordHash, HashingMod),
     User1 = internal_user:set_tags(
-              internal_user:set_password_hash(User0,
-                                              PasswordHash, HashingAlgorithm),
+              internal_user:set_password_hash(User0, PasswordHash, HashingMod),
               ConvertedTags),
     User = case Limits of
                undefined -> User1;
@@ -292,7 +302,10 @@ delete_user(Username, ActingUser) ->
                                      {user_who_performed_action, ActingUser}]),
                 ok;
             false ->
-                ok
+                ok;
+            Error0 ->
+                rabbit_log:info("Failed to delete user '~ts': ~tp", [Username, Error0]),
+                throw(Error0)
         end
     catch
         Class:Error:Stacktrace ->
@@ -618,9 +631,9 @@ clear_topic_permissions(Username, VirtualHost, Exchange, ActingUser) ->
               Username, VirtualHost, Exchange),
         rabbit_log:info("Successfully cleared topic permissions on exchange '~ts' for user '~ts' in virtual host '~ts'",
                         [Exchange, Username, VirtualHost]),
-        rabbit_event:notify(permission_deleted, [{user,  Username},
-                                                 {vhost, VirtualHost},
-                                                 {user_who_performed_action, ActingUser}]),
+        rabbit_event:notify(topic_permission_deleted, [{user,  Username},
+                                                       {vhost, VirtualHost},
+                                                       {user_who_performed_action, ActingUser}]),
         R
     catch
         Class:Error:Stacktrace ->
@@ -685,8 +698,8 @@ put_user(User, Version, ActingUser) ->
                     throw({error, both_password_and_password_hash_are_provided});
                 %% clear password, update tags if needed
                 _ ->
-                    rabbit_auth_backend_internal:set_tags(Username, Tags, ActingUser),
-                    rabbit_auth_backend_internal:clear_password(Username, ActingUser)
+                    set_tags(Username, Tags, ActingUser),
+                    clear_password(Username, ActingUser)
             end;
         false ->
             case {HasPassword, HasPasswordHash} of
@@ -719,13 +732,13 @@ update_user_password_hash(Username, PasswordHash, Tags, Limits, User, Version) -
 
     Hash = rabbit_misc:b64decode_or_throw(PasswordHash),
     ConvertedTags = [rabbit_data_coercion:to_atom(I) || I <- Tags],
-    rabbit_auth_backend_internal:update_user_with_hash(
+    update_user_with_hash(
       Username, Hash, HashingAlgorithm, ConvertedTags, Limits).
 
 create_user_with_password(_PassedCredentialValidation = true,  Username, Password, Tags, undefined, Limits, ActingUser) ->
-    ok = rabbit_auth_backend_internal:add_user(Username, Password, ActingUser, Limits, Tags);
+    ok = add_user(Username, Password, ActingUser, Limits, Tags);
 create_user_with_password(_PassedCredentialValidation = true,  Username, Password, Tags, PreconfiguredPermissions, Limits, ActingUser) ->
-    ok = rabbit_auth_backend_internal:add_user(Username, Password, ActingUser, Limits, Tags),
+    ok = add_user(Username, Password, ActingUser, Limits, Tags),
     preconfigure_permissions(Username, PreconfiguredPermissions, ActingUser);
 create_user_with_password(_PassedCredentialValidation = false, _Username, _Password, _Tags, _, _, _) ->
     %% we don't log here because
@@ -738,14 +751,14 @@ create_user_with_password_hash(Username, PasswordHash, Tags, User, Version, Prec
     HashingAlgorithm = hashing_algorithm(User, Version),
     Hash             = rabbit_misc:b64decode_or_throw(PasswordHash),
 
-    rabbit_auth_backend_internal:add_user_sans_validation(Username, Hash, HashingAlgorithm, Tags, Limits, ActingUser),
+    add_user_sans_validation(Username, Hash, HashingAlgorithm, Tags, Limits, ActingUser),
     preconfigure_permissions(Username, PreconfiguredPermissions, ActingUser).
 
 preconfigure_permissions(_Username, undefined, _ActingUser) ->
     ok;
 preconfigure_permissions(Username, Map, ActingUser) when is_map(Map) ->
     _ = maps:map(fun(VHost, M) ->
-                     rabbit_auth_backend_internal:set_permissions(Username, VHost,
+                     set_permissions(Username, VHost,
                                                   maps:get(<<"configure">>, M),
                                                   maps:get(<<"write">>,     M),
                                                   maps:get(<<"read">>,      M),
