@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 %% The classic queue store works as follow:
@@ -71,7 +71,7 @@
 
 -type buffer() :: #{
     %% SeqId => {Offset, Size, Msg}
-    rabbit_variable_queue:seq_id() => {non_neg_integer(), non_neg_integer(), #basic_message{}}
+                    rabbit_variable_queue:seq_id() => {non_neg_integer(), non_neg_integer(), mc:state()}
 }.
 
 -record(qs, {
@@ -142,7 +142,7 @@ maybe_close_fd(Fd) ->
 info(#qs{ write_buffer = WriteBuffer }) ->
     [{qs_buffer_size, map_size(WriteBuffer)}].
 
--spec write(rabbit_variable_queue:seq_id(), rabbit_types:basic_message(),
+-spec write(rabbit_variable_queue:seq_id(), mc:state(),
             rabbit_types:message_properties(), State)
         -> {msg_location(), State} when State::state().
 
@@ -283,7 +283,7 @@ build_data({_, Size, Msg}, CheckCRC32) ->
     ].
 
 -spec read(rabbit_variable_queue:seq_id(), msg_location(), State)
-        -> {rabbit_types:basic_message(), State} when State::state().
+        -> {mc:state(), State} when State::state().
 
 read(SeqId, DiskLocation, State = #qs{ write_buffer = WriteBuffer,
                                        cache = Cache }) ->
@@ -328,7 +328,7 @@ read_from_disk(SeqId, {?MODULE, Offset, Size}, State0) ->
     {Msg, State}.
 
 -spec read_many([{rabbit_variable_queue:seq_id(), msg_location()}], State)
-        -> {[rabbit_types:basic_message()], State} when State::state().
+        -> {[mc:state()], State} when State::state().
 
 read_many([], State) ->
     {[], State};
