@@ -2,11 +2,12 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_db_exchange).
 
+-include_lib("khepri/include/khepri.hrl").
 -include_lib("rabbit_common/include/rabbit.hrl").
 
 -export([
@@ -31,15 +32,28 @@
 
 %% Used by other rabbit_db_* modules
 -export([
+         maybe_auto_delete_in_khepri/2,
          maybe_auto_delete_in_mnesia/2,
          next_serial_in_mnesia_tx/1,
+         next_serial_in_khepri_tx/1,
+         delete_in_khepri/3,
          delete_in_mnesia/3,
-         update_in_mnesia_tx/2
+         get_in_khepri_tx/1,
+         update_in_mnesia_tx/2,
+         update_in_khepri_tx/2,
+         path/1
          ]).
 
 %% For testing
 -export([clear/0]).
 
+-export([
+         khepri_exchange_path/1,
+         khepri_exchange_serial_path/1,
+         khepri_exchanges_path/0,
+         khepri_exchange_serials_path/0
+        ]).
+
 -define(MNESIA_TABLE, rabbit_exchange).
 -define(MNESIA_DURABLE_TABLE, rabbit_durable_exchange).
 -define(MNESIA_SERIAL_TABLE, rabbit_exchange_serial).
@@ -57,12 +71,17 @@
 %% @private
 
 get_all() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_in_mnesia() end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_in_mnesia() end,
+        khepri => fun() -> get_all_in_khepri() end
+       }).
 
 get_all_in_mnesia() ->
     rabbit_db:list_in_mnesia(?MNESIA_TABLE, #exchange{_ = '_'}).
 
+get_all_in_khepri() ->
+    rabbit_db:list_in_khepri(khepri_exchanges_path() ++ [rabbit_khepri:if_has_data_wildcard()]).
+
 -spec get_all(VHostName) -> [Exchange] when
       VHostName :: vhost:name(),
       Exchange :: rabbit_types:exchange().
@@ -73,14 +92,18 @@ get_all_in_mnesia() ->
 %% @private
 
 get_all(VHost) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_in_mnesia(VHost) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_in_mnesia(VHost) end,
+        khepri => fun() -> get_all_in_khepri(VHost) end
        }).
 
 get_all_in_mnesia(VHost) ->
     Match = #exchange{name = rabbit_misc:r(VHost, exchange), _ = '_'},
     rabbit_db:list_in_mnesia(?MNESIA_TABLE, Match).
 
+get_all_in_khepri(VHost) ->
+    rabbit_db:list_in_khepri(khepri_exchanges_path() ++ [VHost, rabbit_khepri:if_has_data_wildcard()]).
+
 %% -------------------------------------------------------------------
 %% get_all_durable().
 %% -------------------------------------------------------------------
@@ -94,13 +117,17 @@ get_all_in_mnesia(VHost) ->
 %% @private
 
 get_all_durable() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_durable_in_mnesia() end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_durable_in_mnesia() end,
+        khepri => fun() -> get_all_durable_in_khepri() end
        }).
 
 get_all_durable_in_mnesia() ->
     rabbit_db:list_in_mnesia(rabbit_durable_exchange, #exchange{_ = '_'}).
 
+get_all_durable_in_khepri() ->
+    rabbit_db:list_in_khepri(khepri_exchanges_path() ++ [rabbit_khepri:if_has_data_wildcard()]).
+
 %% -------------------------------------------------------------------
 %% list().
 %% -------------------------------------------------------------------
@@ -114,13 +141,23 @@ get_all_durable_in_mnesia() ->
 %% @private
 
 list() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> list_in_mnesia() end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> list_in_mnesia() end,
+        khepri => fun() -> list_in_khepri() end
        }).
 
 list_in_mnesia() ->
     mnesia:dirty_all_keys(?MNESIA_TABLE).
 
+list_in_khepri() ->
+    case rabbit_khepri:match(khepri_exchanges_path() ++
+                                 [rabbit_khepri:if_has_data_wildcard()]) of
+        {ok, Map} ->
+            maps:fold(fun(_K, X, Acc) -> [X#exchange.name | Acc] end, [], Map);
+        _ ->
+            []
+    end.
+
 %% -------------------------------------------------------------------
 %% get().
 %% -------------------------------------------------------------------
@@ -136,13 +173,34 @@ list_in_mnesia() ->
 %% @private
 
 get(Name) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_in_mnesia(Name) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_in_mnesia(Name) end,
+        khepri => fun() -> get_in_khepri(Name) end
        }).
- 
+
 get_in_mnesia(Name) ->
     rabbit_mnesia:dirty_read({?MNESIA_TABLE, Name}).
 
+get_in_khepri(Name) ->
+    case ets:lookup(rabbit_khepri_exchange, Name) of
+        [X] -> {ok, X};
+        []  -> {error, not_found}
+    end.
+
+%% -------------------------------------------------------------------
+%% get_in_khepri_tx().
+%% -------------------------------------------------------------------
+
+-spec get_in_khepri_tx(ExchangeName) -> Ret when
+      ExchangeName :: rabbit_exchange:name(),
+      Ret :: [Exchange :: rabbit_types:exchange()].
+
+get_in_khepri_tx(Name) ->
+    case khepri_tx:get(khepri_exchange_path(Name)) of
+        {ok, X} -> [X];
+        _ -> []
+    end.
+
 %% -------------------------------------------------------------------
 %% get_many().
 %% -------------------------------------------------------------------
@@ -157,8 +215,9 @@ get_in_mnesia(Name) ->
 %% @private
 
 get_many(Names) when is_list(Names) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_many_in_mnesia(?MNESIA_TABLE, Names) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_many_in_mnesia(?MNESIA_TABLE, Names) end,
+        khepri => fun() -> get_many_in_khepri(Names) end
        }).
 
 get_many_in_mnesia(Table, [Name]) -> ets:lookup(Table, Name);
@@ -167,6 +226,9 @@ get_many_in_mnesia(Table, Names) when is_list(Names) ->
     %% expensive for reasons explained in rabbit_mnesia:dirty_read/1.
     lists:append([ets:lookup(Table, Name) || Name <- Names]).
 
+get_many_in_khepri(Names) when is_list(Names) ->
+    lists:append([ets:lookup(rabbit_khepri_exchange, Name) || Name <- Names]).
+
 %% -------------------------------------------------------------------
 %% count().
 %% -------------------------------------------------------------------
@@ -179,12 +241,17 @@ get_many_in_mnesia(Table, Names) when is_list(Names) ->
 %% @private
 
 count() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> count_in_mnesia() end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> count_in_mnesia() end,
+        khepri => fun() -> count_in_khepri() end
+       }).
 
 count_in_mnesia() ->
     mnesia:table_info(?MNESIA_TABLE, size).
 
+count_in_khepri() ->
+    rabbit_khepri:count_children(khepri_exchanges_path() ++ [?KHEPRI_WILDCARD_STAR]).
+
 %% -------------------------------------------------------------------
 %% update().
 %% -------------------------------------------------------------------
@@ -201,8 +268,9 @@ count_in_mnesia() ->
 %% @private
 
 update(XName, Fun) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> update_in_mnesia(XName, Fun) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> update_in_mnesia(XName, Fun) end,
+        khepri => fun() -> update_in_khepri(XName, Fun) end
        }).
 
 update_in_mnesia(XName, Fun) ->
@@ -238,6 +306,52 @@ set_ram_in_mnesia_tx(X) ->
     ok = mnesia:write(?MNESIA_TABLE, X1, write),
     X1.
 
+update_in_khepri(XName, Fun) ->
+    Path = khepri_exchange_path(XName),
+    Ret1 = rabbit_khepri:adv_get(Path),
+    case Ret1 of
+        {ok, #{data := X, payload_version := Vsn}} ->
+            X1 = Fun(X),
+            UpdatePath =
+                khepri_path:combine_with_conditions(
+                  Path, [#if_payload_version{version = Vsn}]),
+            Ret2 = rabbit_khepri:put(UpdatePath, X1),
+            case Ret2 of
+                ok ->
+                    ok;
+                {error, {khepri, mismatching_node, _}} ->
+                    update_in_khepri(XName, Fun);
+                {error, {khepri, node_not_found, _}} ->
+                    ok;
+                {error, _} = Error ->
+                    Error
+            end;
+        {error, {khepri, node_not_found, _}} ->
+            ok;
+        {error, _} = Error ->
+            Error
+    end.
+
+%% -------------------------------------------------------------------
+%% update_in_khepri_tx().
+%% -------------------------------------------------------------------
+
+-spec update_in_khepri_tx(ExchangeName, UpdateFun) -> Ret when
+      ExchangeName :: rabbit_exchange:name(),
+      Exchange :: rabbit_types:exchange(),
+      UpdateFun :: fun((Exchange) -> Exchange),
+      Ret :: not_found | Exchange.
+
+update_in_khepri_tx(Name, Fun) ->
+    Path = khepri_exchange_path(Name),
+    case khepri_tx:get(Path) of
+        {ok, X} ->
+            X1 = Fun(X),
+            ok = khepri_tx:put(Path, X1),
+            X1;
+        _ -> not_found
+    end.
+
 %% -------------------------------------------------------------------
 %% create_or_get().
 %% -------------------------------------------------------------------
@@ -254,8 +368,9 @@ set_ram_in_mnesia_tx(X) ->
 %% @private
 
 create_or_get(X) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> create_or_get_in_mnesia(X) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> create_or_get_in_mnesia(X) end,
+        khepri => fun() -> create_or_get_in_khepri(X) end
        }).
 
 create_or_get_in_mnesia(#exchange{name = XName} = X) ->
@@ -269,6 +384,15 @@ create_or_get_in_mnesia(#exchange{name = XName} = X) ->
               end
       end).
 
+create_or_get_in_khepri(#exchange{name = XName} = X) ->
+    Path = khepri_exchange_path(XName),
+    case rabbit_khepri:create(Path, X) of
+        ok ->
+            {new, X};
+        {error, {khepri, mismatching_node, #{node_props := #{data := ExistingX}}}} ->
+            {existing, ExistingX}
+    end.
+
 %% -------------------------------------------------------------------
 %% set().
 %% -------------------------------------------------------------------
@@ -282,8 +406,9 @@ create_or_get_in_mnesia(#exchange{name = XName} = X) ->
 %% @private
 
 set(Xs) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> set_in_mnesia(Xs) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> set_in_mnesia(Xs) end,
+        khepri => fun() -> set_in_khepri(Xs) end
        }).
 
 set_in_mnesia(Xs) when is_list(Xs) ->
@@ -293,6 +418,18 @@ set_in_mnesia(Xs) when is_list(Xs) ->
       end),
     ok.
 
+set_in_khepri(Xs) when is_list(Xs) ->
+    rabbit_khepri:transaction(
+      fun() ->
+              [set_in_khepri_tx(X) || X <- Xs]
+      end, rw),
+    ok.
+
+set_in_khepri_tx(X) ->
+    Path = khepri_exchange_path(X#exchange.name),
+    ok = khepri_tx:put(Path, X),
+    X.
+
 %% -------------------------------------------------------------------
 %% peek_serial().
 %% -------------------------------------------------------------------
@@ -307,8 +444,9 @@ set_in_mnesia(Xs) when is_list(Xs) ->
 %% @private
 
 peek_serial(XName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> peek_serial_in_mnesia(XName) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> peek_serial_in_mnesia(XName) end,
+        khepri => fun() -> peek_serial_in_khepri(XName) end
        }).
 
 peek_serial_in_mnesia(XName) ->
@@ -323,6 +461,15 @@ peek_serial_in_mnesia_tx(XName, LockType) ->
         _                                  -> 1
     end.
 
+peek_serial_in_khepri(XName) ->
+    Path = khepri_exchange_serial_path(XName),
+    case rabbit_khepri:get(Path) of
+        {ok, Serial} ->
+            Serial;
+        _ ->
+            1
+    end.
+
 %% -------------------------------------------------------------------
 %% next_serial().
 %% -------------------------------------------------------------------
@@ -337,8 +484,9 @@ peek_serial_in_mnesia_tx(XName, LockType) ->
 %% @private
 
 next_serial(XName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> next_serial_in_mnesia(XName) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> next_serial_in_mnesia(XName) end,
+        khepri => fun() -> next_serial_in_khepri(XName) end
        }).
 
 next_serial_in_mnesia(XName) ->
@@ -356,6 +504,43 @@ next_serial_in_mnesia_tx(XName) ->
                       #exchange_serial{name = XName, next = Serial + 1}, write),
     Serial.
 
+next_serial_in_khepri(XName) ->
+    %% Just storing the serial number is enough, no need to keep #exchange_serial{}
+    Path = khepri_exchange_serial_path(XName),
+    Ret1 = rabbit_khepri:adv_get(Path),
+    case Ret1 of
+        {ok, #{data := Serial,
+               payload_version := Vsn}} ->
+            UpdatePath =
+                khepri_path:combine_with_conditions(
+                  Path, [#if_payload_version{version = Vsn}]),
+            case rabbit_khepri:put(UpdatePath, Serial + 1) of
+                ok ->
+                    Serial;
+                {error, {khepri, mismatching_node, _}} ->
+                    next_serial_in_khepri(XName);
+                Err ->
+                    Err
+            end;
+        _ ->
+            Serial = 1,
+            ok = rabbit_khepri:put(Path, Serial + 1),
+            Serial
+    end.
+
+-spec next_serial_in_khepri_tx(Exchange) -> Serial when
+      Exchange :: rabbit_types:exchange(),
+      Serial :: integer().
+
+next_serial_in_khepri_tx(#exchange{name = XName}) ->
+    Path = khepri_exchange_serial_path(XName),
+    Serial = case khepri_tx:get(Path) of
+                 {ok, Serial0} -> Serial0;
+                 _ -> 1
+             end,
+    ok = khepri_tx:put(Path, Serial + 1),
+    Serial.
+
 %% -------------------------------------------------------------------
 %% delete().
 %% -------------------------------------------------------------------
@@ -380,8 +565,9 @@ next_serial_in_mnesia_tx(XName) ->
 %% @private
 
 delete(XName, IfUnused) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> delete_in_mnesia(XName, IfUnused) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> delete_in_mnesia(XName, IfUnused) end,
+        khepri => fun() -> delete_in_khepri(XName, IfUnused) end
        }).
 
 delete_in_mnesia(XName, IfUnused) ->
@@ -420,6 +606,32 @@ delete_in_mnesia(X = #exchange{name = XName}, OnlyDurable, RemoveBindingsForSour
     rabbit_db_binding:delete_all_for_exchange_in_mnesia(
       X, OnlyDurable, RemoveBindingsForSource).
 
+delete_in_khepri(XName, IfUnused) ->
+    DeletionFun = case IfUnused of
+                      true  -> fun conditional_delete_in_khepri/2;
+                      false -> fun unconditional_delete_in_khepri/2
+                  end,
+    rabbit_khepri:transaction(
+      fun() ->
+              case khepri_tx:get(khepri_exchange_path(XName)) of
+                  {ok, X} -> DeletionFun(X, false);
+                  _ -> {error, not_found}
+              end
+      end, rw).
+
+conditional_delete_in_khepri(X = #exchange{name = XName}, OnlyDurable) ->
+    case rabbit_db_binding:has_for_source_in_khepri(XName) of
+        false  -> delete_in_khepri(X, OnlyDurable, false);
+        true   -> {error, in_use}
+    end.
+
+unconditional_delete_in_khepri(X, OnlyDurable) ->
+    delete_in_khepri(X, OnlyDurable, true).
+
+delete_in_khepri(X = #exchange{name = XName}, OnlyDurable, RemoveBindingsForSource) ->
+    ok = khepri_tx:delete(khepri_exchange_path(XName)),
+    rabbit_db_binding:delete_all_for_exchange_in_khepri(X, OnlyDurable, RemoveBindingsForSource).
+
 %% -------------------------------------------------------------------
 %% delete_serial().
 %% -------------------------------------------------------------------
@@ -433,8 +645,9 @@ delete_in_mnesia(X = #exchange{name = XName}, OnlyDurable, RemoveBindingsForSour
 %% @private
 
 delete_serial(XName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> delete_serial_in_mnesia(XName) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> delete_serial_in_mnesia(XName) end,
+        khepri => fun() -> delete_serial_in_khepri(XName) end
        }).
 
 delete_serial_in_mnesia(XName) ->
@@ -443,6 +656,10 @@ delete_serial_in_mnesia(XName) ->
               mnesia:delete({?MNESIA_SERIAL_TABLE, XName})
       end).
 
+delete_serial_in_khepri(XName) ->
+    Path = khepri_exchange_serial_path(XName),
+    ok = rabbit_khepri:delete(Path).
+
 %% -------------------------------------------------------------------
 %% recover().
 %% -------------------------------------------------------------------
@@ -457,8 +674,9 @@ delete_serial_in_mnesia(XName) ->
 %% @private
 
 recover(VHost) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> recover_in_mnesia(VHost) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> recover_in_mnesia(VHost) end,
+        khepri => fun() -> recover_in_khepri(VHost) end
        }).
 
 recover_in_mnesia(VHost) ->
@@ -477,6 +695,28 @@ recover_in_mnesia(VHost) ->
       end,
       ?MNESIA_DURABLE_TABLE).
 
+recover_in_khepri(VHost) ->
+    %% Transient exchanges are deprecated in Khepri, all exchanges are recovered
+    %% Node boot and recovery should hang until the data is ready.
+    %% Recovery needs to wait until progress can be done, as it
+    %% cannot be skipped and stopping the node is not an option -
+    %% the next boot most likely would behave the same way.
+    %% Any other request stays with the default timeout, currently 30s.
+    Exchanges0 = rabbit_db:list_in_khepri(khepri_exchanges_path() ++ [VHost, rabbit_khepri:if_has_data_wildcard()],
+                                #{timeout => infinity}),
+    Exchanges = [rabbit_exchange_decorator:set(X) || X <- Exchanges0],
+
+    rabbit_khepri:transaction(
+      fun() ->
+              [_ = set_in_khepri_tx(X) || X <- Exchanges]
+      end, rw, #{timeout => infinity}),
+    %% TODO once mnesia is gone, this callback should go back to `rabbit_exchange`
+    [begin
+         Serial = rabbit_exchange:serial(X),
+         rabbit_exchange:callback(X, create, Serial, [X])
+     end || X <- Exchanges],
+    Exchanges.
+
 %% -------------------------------------------------------------------
 %% match().
 %% -------------------------------------------------------------------
@@ -492,11 +732,12 @@ recover_in_mnesia(VHost) ->
 %% @private
 
 match(Pattern) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> match_in_mnesia(Pattern) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> match_in_mnesia(Pattern) end,
+        khepri => fun() -> match_in_khepri(Pattern) end
        }).
 
-match_in_mnesia(Pattern) -> 
+match_in_mnesia(Pattern) ->
     case mnesia:transaction(
            fun() ->
                    mnesia:match_object(?MNESIA_TABLE, Pattern, read)
@@ -505,6 +746,10 @@ match_in_mnesia(Pattern) ->
         {aborted, Err} -> {error, Err}
     end.
 
+match_in_khepri(Pattern0) ->
+    Pattern = #if_data_matches{pattern = Pattern0},
+    rabbit_db:list_in_khepri(khepri_exchanges_path() ++ [?KHEPRI_WILDCARD_STAR, Pattern]).
+
 %% -------------------------------------------------------------------
 %% exists().
 %% -------------------------------------------------------------------
@@ -519,12 +764,17 @@ match_in_mnesia(Pattern) ->
 %% @private
 
 exists(Name) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> exists_in_mnesia(Name) end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> exists_in_mnesia(Name) end,
+        khepri => fun() -> exists_in_khepri(Name) end
+       }).
 
 exists_in_mnesia(Name) ->
     ets:member(?MNESIA_TABLE, Name).
 
+exists_in_khepri(Name) ->
+    rabbit_khepri:exists(khepri_exchange_path(Name)).
+
 %% -------------------------------------------------------------------
 %% clear().
 %% -------------------------------------------------------------------
@@ -535,8 +785,10 @@ exists_in_mnesia(Name) ->
 %% @private
 
 clear() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> clear_in_mnesia() end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> clear_in_mnesia() end,
+        khepri => fun() -> clear_in_khepri() end
+       }).
 
 clear_in_mnesia() ->
     {atomic, ok} = mnesia:clear_table(?MNESIA_TABLE),
@@ -544,6 +796,16 @@ clear_in_mnesia() ->
     {atomic, ok} = mnesia:clear_table(?MNESIA_SERIAL_TABLE),
     ok.
 
+clear_in_khepri() ->
+    khepri_delete(khepri_exchanges_path()),
+    khepri_delete(khepri_exchange_serials_path()).
+
+khepri_delete(Path) ->
+    case rabbit_khepri:delete(Path) of
+        ok -> ok;
+        Error -> throw(Error)
+    end.
+
 %% -------------------------------------------------------------------
 %% maybe_auto_delete_in_mnesia().
 %% -------------------------------------------------------------------
@@ -567,3 +829,54 @@ maybe_auto_delete_in_mnesia(XName, OnlyDurable) ->
                 {deleted, X, [], Deletions} -> {deleted, X, Deletions}
             end
     end.
+
+%% -------------------------------------------------------------------
+%% maybe_auto_delete_in_khepri().
+%% -------------------------------------------------------------------
+
+-spec maybe_auto_delete_in_khepri(ExchangeName, boolean()) -> Ret when
+      ExchangeName :: rabbit_exchange:name(),
+      Exchange :: rabbit_types:exchange(),
+      Deletions :: rabbit_binding:deletions(),
+      Ret ::  {'not_deleted', 'undefined' | Exchange} |
+              {'deleted', Exchange, Deletions}.
+
+maybe_auto_delete_in_khepri(XName, OnlyDurable) ->
+    case khepri_tx:get(khepri_exchange_path(XName)) of
+        {ok, #exchange{auto_delete = false} = X} ->
+            {not_deleted, X};
+        {ok, #exchange{auto_delete = true} = X} ->
+            case conditional_delete_in_khepri(X, OnlyDurable) of
+                {error, in_use}             -> {not_deleted, X};
+                {deleted, X, [], Deletions} -> {deleted, X, Deletions}
+            end;
+        {error, _} ->
+            {not_deleted, undefined}
+    end.
+
+%% -------------------------------------------------------------------
+%% Khepri paths
+%% -------------------------------------------------------------------
+
+khepri_exchanges_path() ->
+    [?MODULE, exchanges].
+
+khepri_exchange_path(#resource{virtual_host = VHost, name = Name}) ->
+    [?MODULE, exchanges, VHost, Name].
+
+khepri_exchange_serials_path() ->
+    [?MODULE, exchange_serials].
+
+khepri_exchange_serial_path(#resource{virtual_host = VHost, name = Name}) ->
+    [?MODULE, exchange_serials, VHost, Name].
+
+%% -------------------------------------------------------------------
+%% path().
+%% -------------------------------------------------------------------
+
+-spec path(ExchangeName) -> Path when
+      ExchangeName :: rabbit_exchange:name(),
+      Path :: khepri_path:path().
+
+path(Name) ->
+    khepri_exchange_path(Name).
