@@ -2,13 +2,16 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_db_queue).
 
--include_lib("rabbit_common/include/rabbit.hrl").
+-include_lib("kernel/include/logger.hrl").
 -include_lib("stdlib/include/qlc.hrl").
+
+-include_lib("khepri/include/khepri.hrl").
+-include_lib("rabbit_common/include/rabbit.hrl").
 -include("amqqueue.hrl").
 
 -export([
@@ -26,10 +29,12 @@
          set_many/1,
          delete/2,
          update/2,
-         update_decorators/1,
+         update_decorators/2,
          exists/1
         ]).
 
+%% Once mnesia is removed, all transient entities will be deleted. These can be replaced
+%% with the plain get_all* functions
 -export([
          get_all_durable/0,
          get_all_durable_by_type/1,
@@ -40,28 +45,46 @@
          consistent_exists/1
         ]).
 
-%% Used by on_node_up and on_node_down
+%% Used by on_node_up and on_node_down.
+%% Can be deleted once transient entities/mnesia are removed.
 -export([foreach_transient/1,
          delete_transient/1]).
 
-%% Used only by forget all durable
+%% Only used by rabbit_amqqueue:forget_node_for_queue, which is only called
+%% by `rabbit_mnesia:remove_node_if_mnesia_running`. Thus, once mnesia and/or
+%% HA queues are removed it can be deleted.
 -export([foreach_durable/2,
          internal_delete/3]).
 
+%% Storing it on Khepri is not needed, this function is just used in
+%% rabbit_quorum_queue to ensure the queue is present in the rabbit_queue
+%% table and not just in rabbit_durable_queue. Can be deleted with mnesia removal
 -export([set_dirty/1]).
 
 %% Used by other rabbit_db_* modules
 -export([
          update_in_mnesia_tx/2,
-         get_durable_in_mnesia_tx/1
+         update_in_khepri_tx/2,
+         get_durable_in_mnesia_tx/1,
+         get_in_khepri_tx/1
         ]).
 
 %% For testing
 -export([clear/0]).
 
+-export([
+         khepri_queue_path/1,
+         khepri_queues_path/0
+        ]).
+
+-dialyzer({nowarn_function, [foreach_transient/1,
+                             foreach_transient_in_khepri/1]}).
+
 -define(MNESIA_TABLE, rabbit_queue).
 -define(MNESIA_DURABLE_TABLE, rabbit_durable_queue).
 
+-define(KHEPRI_PROJECTION, rabbit_khepri_queue).
+
 %% -------------------------------------------------------------------
 %% get_all().
 %% -------------------------------------------------------------------
@@ -76,8 +99,9 @@
 %% @private
 
 get_all() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_in_mnesia() end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_in_mnesia() end,
+        khepri => fun() -> get_all_in_khepri() end
        }).
 
 get_all_in_mnesia() ->
@@ -86,6 +110,12 @@ get_all_in_mnesia() ->
               rabbit_db:list_in_mnesia(?MNESIA_TABLE, amqqueue:pattern_match_all())
       end).
 
+get_all_in_khepri() ->
+    list_with_possible_retry_in_khepri(
+      fun() ->
+              ets:tab2list(?KHEPRI_PROJECTION)
+      end).
+
 -spec get_all(VHostName) -> [Queue] when
       VHostName :: vhost:name(),
       Queue :: amqqueue:amqqueue().
@@ -97,8 +127,9 @@ get_all_in_mnesia() ->
 %% @private
 
 get_all(VHostName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_in_mnesia(VHostName) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_in_mnesia(VHostName) end,
+        khepri => fun() -> get_all_in_khepri(VHostName) end
        }).
 
 get_all_in_mnesia(VHostName) ->
@@ -108,6 +139,13 @@ get_all_in_mnesia(VHostName) ->
               rabbit_db:list_in_mnesia(?MNESIA_TABLE, Pattern)
       end).
 
+get_all_in_khepri(VHostName) ->
+    list_with_possible_retry_in_khepri(
+      fun() ->
+              Pattern = amqqueue:pattern_match_on_name(rabbit_misc:r(VHostName, queue)),
+              ets:match_object(?KHEPRI_PROJECTION, Pattern)
+      end).
+
 %% -------------------------------------------------------------------
 %% get_all_durable().
 %% -------------------------------------------------------------------
@@ -122,8 +160,9 @@ get_all_in_mnesia(VHostName) ->
 %% @private
 
 get_all_durable() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_durable_in_mnesia() end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_durable_in_mnesia() end,
+        khepri => fun() -> get_all_durable_in_khepri() end
        }).
 
 get_all_durable_in_mnesia() ->
@@ -132,6 +171,13 @@ get_all_durable_in_mnesia() ->
               rabbit_db:list_in_mnesia(?MNESIA_DURABLE_TABLE, amqqueue:pattern_match_all())
       end).
 
+get_all_durable_in_khepri() ->
+    list_with_possible_retry_in_khepri(
+      fun() ->
+              Pattern = amqqueue:pattern_match_on_durable(true),
+              ets:match_object(?KHEPRI_PROJECTION, Pattern)
+      end).
+
 -spec get_all_durable_by_type(Type) -> [Queue] when
       Type :: atom(),
       Queue :: amqqueue:amqqueue().
@@ -143,14 +189,19 @@ get_all_durable_in_mnesia() ->
 %% @private
 
 get_all_durable_by_type(Type) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_durable_by_type_in_mnesia(Type) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_durable_by_type_in_mnesia(Type) end,
+        khepri => fun() -> get_all_durable_by_type_in_khepri(Type) end
        }).
 
 get_all_durable_by_type_in_mnesia(Type) ->
     Pattern = amqqueue:pattern_match_on_type(Type),
     rabbit_db:list_in_mnesia(?MNESIA_DURABLE_TABLE, Pattern).
 
+get_all_durable_by_type_in_khepri(Type) ->
+    Pattern = amqqueue:pattern_match_on_type_and_durable(Type, true),
+    ets:match_object(?KHEPRI_PROJECTION, Pattern).
+
 %% -------------------------------------------------------------------
 %% filter_all_durable().
 %% -------------------------------------------------------------------
@@ -166,8 +217,9 @@ get_all_durable_by_type_in_mnesia(Type) ->
 %% @private
 
 filter_all_durable(FilterFun) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> filter_all_durable_in_mnesia(FilterFun) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> filter_all_durable_in_mnesia(FilterFun) end,
+        khepri => fun() -> filter_all_durable_in_khepri(FilterFun) end
        }).
 
 filter_all_durable_in_mnesia(FilterFun) ->
@@ -178,6 +230,16 @@ filter_all_durable_in_mnesia(FilterFun) ->
                           ]))
       end).
 
+filter_all_durable_in_khepri(FilterFun) ->
+    ets:foldl(
+      fun(Q, Acc0) ->
+              case amqqueue:is_durable(Q) andalso FilterFun(Q) of
+                  true -> [Q | Acc0];
+                  false -> Acc0
+              end
+      end,
+      [], ?KHEPRI_PROJECTION).
+
 %% -------------------------------------------------------------------
 %% list().
 %% -------------------------------------------------------------------
@@ -192,13 +254,18 @@ filter_all_durable_in_mnesia(FilterFun) ->
 %% @private
 
 list() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> list_in_mnesia() end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> list_in_mnesia() end,
+        khepri => fun() -> list_in_khepri() end
        }).
 
 list_in_mnesia() ->
     mnesia:dirty_all_keys(?MNESIA_TABLE).
 
+list_in_khepri() ->
+    Pattern = amqqueue:pattern_match_on_name('$1'),
+    ets:select(?KHEPRI_PROJECTION, [{Pattern, [], ['$1']}]).
+
 %% -------------------------------------------------------------------
 %% count().
 %% -------------------------------------------------------------------
@@ -213,13 +280,17 @@ list_in_mnesia() ->
 %% @private
 
 count() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> count_in_mnesia() end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> count_in_mnesia() end,
+        khepri => fun() -> count_in_khepri() end
        }).
 
 count_in_mnesia() ->
     mnesia:table_info(?MNESIA_TABLE, size).
 
+count_in_khepri() ->
+    ets:info(?KHEPRI_PROJECTION, size).
+
 -spec count(VHostName) -> Count when
       VHostName :: vhost:name(),
       Count :: integer().
@@ -240,8 +311,9 @@ count(VHostName) ->
     end.
 
 list_for_count(VHostName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> list_for_count_in_mnesia(VHostName) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> list_for_count_in_mnesia(VHostName) end,
+        khepri => fun() -> list_for_count_in_khepri(VHostName) end
        }).
 
 list_for_count_in_mnesia(VHostName) ->
@@ -257,6 +329,10 @@ list_for_count_in_mnesia(VHostName) ->
                                              amqqueue:field_vhost()))
       end).
 
+list_for_count_in_khepri(VHostName) ->
+    Pattern = amqqueue:pattern_match_on_name(rabbit_misc:r(VHostName, queue)),
+    ets:select_count(?KHEPRI_PROJECTION, [{Pattern, [], [true]}]).
+
 %% -------------------------------------------------------------------
 %% delete().
 %% -------------------------------------------------------------------
@@ -267,8 +343,9 @@ list_for_count_in_mnesia(VHostName) ->
       Ret :: ok | Deletions :: rabbit_binding:deletions().
 
 delete(QueueName, Reason) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> delete_in_mnesia(QueueName, Reason) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> delete_in_mnesia(QueueName, Reason) end,
+        khepri => fun() -> delete_in_khepri(QueueName) end
        }).
 
 delete_in_mnesia(QueueName, Reason) ->
@@ -287,6 +364,23 @@ delete_in_mnesia(QueueName, Reason) ->
               end
       end).
 
+delete_in_khepri(QueueName) ->
+    delete_in_khepri(QueueName, false).
+
+delete_in_khepri(QueueName, OnlyDurable) ->
+    rabbit_khepri:transaction(
+      fun () ->
+              Path = khepri_queue_path(QueueName),
+              case khepri_tx_adv:delete(Path) of
+                  {ok, #{data := _}} ->
+                      %% we want to execute some things, as decided by rabbit_exchange,
+                      %% after the transaction.
+                      rabbit_db_binding:delete_for_destination_in_khepri(QueueName, OnlyDurable);
+                  {ok, _} ->
+                      ok
+              end
+      end, rw).
+
 %% -------------------------------------------------------------------
 %% internal_delete().
 %% -------------------------------------------------------------------
@@ -301,8 +395,9 @@ internal_delete(QueueName, OnlyDurable, Reason) ->
     %% Only used by rabbit_amqqueue:forget_node_for_queue, which is only called
     %% by `rabbit_mnesia:remove_node_if_mnesia_running'. Thus, once mnesia and/or
     %% HA queues are removed it can be removed.
-    rabbit_db:run(
-      #{mnesia => fun() -> internal_delete_in_mnesia(QueueName, OnlyDurable, Reason) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> internal_delete_in_mnesia(QueueName, OnlyDurable, Reason) end,
+        khepri => fun() -> delete_in_khepri(QueueName, OnlyDurable) end
        }).
 
 internal_delete_in_mnesia(QueueName, OnlyDurable, Reason) ->
@@ -326,20 +421,35 @@ internal_delete_in_mnesia(QueueName, OnlyDurable, Reason) ->
 %% get_many().
 %% -------------------------------------------------------------------
 
--spec get_many([QName]) -> Ret when
-      QName :: rabbit_amqqueue:name(),
-      Ret :: [Queue :: amqqueue:amqqueue()].
+-spec get_many(rabbit_exchange:route_return()) ->
+    [amqqueue:amqqueue() | {amqqueue:amqqueue(), rabbit_exchange:route_infos()}].
 get_many(Names) when is_list(Names) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_many_in_mnesia(?MNESIA_TABLE, Names) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_many_in_ets(?MNESIA_TABLE, Names) end,
+        khepri => fun() -> get_many_in_ets(?KHEPRI_PROJECTION, Names) end
        }).
 
-get_many_in_mnesia(Table, [Name]) ->
+get_many_in_ets(Table, [{Name, RouteInfos}])
+  when is_map(RouteInfos) ->
+    case ets:lookup(Table, Name) of
+        [] -> [];
+        [Q] -> [{Q, RouteInfos}]
+    end;
+get_many_in_ets(Table, [Name]) ->
     ets:lookup(Table, Name);
-get_many_in_mnesia(Table, Names) when is_list(Names) ->
-    %% Normally we'd call mnesia:dirty_read/1 here, but that is quite
-    %% expensive for reasons explained in rabbit_mnesia:dirty_read/1.
-    lists:append([ets:lookup(Table, Name) || Name <- Names]).
+get_many_in_ets(Table, Names) when is_list(Names) ->
+    lists:filtermap(fun({Name, RouteInfos})
+                          when is_map(RouteInfos) ->
+                            case ets:lookup(Table, Name) of
+                                [] -> false;
+                                [Q] -> {true, {Q, RouteInfos}}
+                            end;
+                       (Name) ->
+                            case ets:lookup(Table, Name) of
+                                [] -> false;
+                                [Q] -> {true, Q}
+                            end
+                    end, Names).
 
 %% -------------------------------------------------------------------
 %% get().
@@ -349,13 +459,20 @@ get_many_in_mnesia(Table, Names) when is_list(Names) ->
       QName :: rabbit_amqqueue:name(),
       Ret :: {ok, Queue :: amqqueue:amqqueue()} | {error, not_found}.
 get(Name) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_in_mnesia(Name) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_in_mnesia(Name) end,
+        khepri => fun() -> get_in_khepri(Name) end
        }).
 
 get_in_mnesia(Name) ->
     rabbit_mnesia:dirty_read({?MNESIA_TABLE, Name}).
 
+get_in_khepri(Name) ->
+    case ets:lookup(?KHEPRI_PROJECTION, Name) of
+        [Q] -> {ok, Q};
+        []  -> {error, not_found}
+    end.
+
 %% -------------------------------------------------------------------
 %% get_durable().
 %% -------------------------------------------------------------------
@@ -365,13 +482,25 @@ get_in_mnesia(Name) ->
       Ret :: {ok, Queue :: amqqueue:amqqueue()} | {error, not_found}.
 
 get_durable(Name) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_durable_in_mnesia(Name) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_durable_in_mnesia(Name) end,
+        khepri => fun() -> get_durable_in_khepri(Name) end
        }).
 
 get_durable_in_mnesia(Name) ->
     rabbit_mnesia:dirty_read({?MNESIA_DURABLE_TABLE, Name}).
 
+get_durable_in_khepri(Name) ->
+    case get_in_khepri(Name) of
+        {ok, Queue} = Ret ->
+            case amqqueue:is_durable(Queue) of
+                true  -> Ret;
+                false -> {error, not_found}
+            end;
+        Error ->
+            Error
+    end.
+
 %% -------------------------------------------------------------------
 %% get_many_durable().
 %% -------------------------------------------------------------------
@@ -381,10 +510,18 @@ get_durable_in_mnesia(Name) ->
       Ret :: [Queue :: amqqueue:amqqueue()].
 
 get_many_durable(Names) when is_list(Names) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_many_in_mnesia(?MNESIA_DURABLE_TABLE, Names) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_many_durable_in_mnesia(Names) end,
+        khepri => fun() -> get_many_durable_in_khepri(Names) end
        }).
 
+get_many_durable_in_mnesia(Names) ->
+    get_many_in_ets(?MNESIA_DURABLE_TABLE, Names).
+
+get_many_durable_in_khepri(Names) ->
+    Queues = get_many_in_ets(?KHEPRI_PROJECTION, Names),
+    [Q || Q <- Queues, amqqueue:is_durable(Q)].
+
 %% -------------------------------------------------------------------
 %% update().
 %% -------------------------------------------------------------------
@@ -393,14 +530,15 @@ get_many_durable(Names) when is_list(Names) ->
       QName :: rabbit_amqqueue:name(),
       Queue :: amqqueue:amqqueue(),
       UpdateFun :: fun((Queue) -> Queue),
-      Ret :: Queue | not_found. 
+      Ret :: Queue | not_found.
 %% @doc Updates an existing queue record using `UpdateFun'.
 %%
 %% @private
 
 update(QName, Fun) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> update_in_mnesia(QName, Fun) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> update_in_mnesia(QName, Fun) end,
+        khepri => fun() -> update_in_khepri(QName, Fun) end
        }).
 
 update_in_mnesia(QName, Fun) ->
@@ -409,31 +547,77 @@ update_in_mnesia(QName, Fun) ->
               update_in_mnesia_tx(QName, Fun)
       end).
 
+update_in_khepri(QName, Fun) ->
+    Path = khepri_queue_path(QName),
+    Ret1 = rabbit_khepri:adv_get(Path),
+    case Ret1 of
+        {ok, #{data := Q, payload_version := Vsn}} ->
+            UpdatePath = khepri_path:combine_with_conditions(
+                           Path, [#if_payload_version{version = Vsn}]),
+            Q1 = Fun(Q),
+            Ret2 = rabbit_khepri:put(UpdatePath, Q1),
+            case Ret2 of
+                ok -> Q1;
+                {error, {khepri, mismatching_node, _}} ->
+                    update_in_khepri(QName, Fun);
+                Err -> Err
+            end;
+        _  ->
+            not_found
+    end.
+
 %% -------------------------------------------------------------------
 %% update_decorators().
 %% -------------------------------------------------------------------
 
--spec update_decorators(QName) -> ok when
-      QName :: rabbit_amqqueue:name().
+-spec update_decorators(QName, [Decorator]) -> ok when
+      QName :: rabbit_amqqueue:name(),
+      Decorator :: atom().
 %% @doc Updates an existing queue record adding the active queue decorators.
 %%
 %% @private
 
-update_decorators(QName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> update_decorators_in_mnesia(QName) end
+update_decorators(QName, Decorators) ->
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> update_decorators_in_mnesia(QName, Decorators) end,
+        khepri => fun() -> update_decorators_in_khepri(QName, Decorators) end
        }).
 
-update_decorators_in_mnesia(Name) ->
+update_decorators_in_mnesia(Name, Decorators) ->
     rabbit_mnesia:execute_mnesia_transaction(
       fun() ->
               case mnesia:wread({?MNESIA_TABLE, Name}) of
-                  [Q] -> ok = mnesia:write(?MNESIA_TABLE, rabbit_queue_decorator:set(Q),
+                  [Q] -> ok = mnesia:write(?MNESIA_TABLE, amqqueue:set_decorators(Q, Decorators),
                                            write);
                   []  -> ok
               end
       end).
 
+update_decorators_in_khepri(QName, Decorators) ->
+    %% Decorators are stored on an ETS table, so we need to query them before the transaction.
+    %% Also, to verify which ones are active could lead to any kind of side-effects.
+    %% Thus it needs to be done outside of the transaction.
+    %% Decorators have just been calculated on `rabbit_queue_decorator:maybe_recover/1`, thus
+    %% we can update them here directly.
+    Path = khepri_queue_path(QName),
+    Ret1 = rabbit_khepri:adv_get(Path),
+    case Ret1 of
+        {ok, #{data := Q0, payload_version := Vsn}} ->
+            Q1 = amqqueue:reset_mirroring_and_decorators(Q0),
+            Q2 = amqqueue:set_decorators(Q1, Decorators),
+            UpdatePath = khepri_path:combine_with_conditions(
+                           Path, [#if_payload_version{version = Vsn}]),
+            Ret2 = rabbit_khepri:put(UpdatePath, Q2),
+            case Ret2 of
+                ok -> ok;
+                {error, {khepri, mismatching_node, _}} ->
+                    update_decorators_in_khepri(QName, Decorators);
+                {error, _} = Error -> Error
+            end;
+        _  ->
+            ok
+    end.
+
 %% -------------------------------------------------------------------
 %% update_durable().
 %% -------------------------------------------------------------------
@@ -447,9 +631,11 @@ update_decorators_in_mnesia(Name) ->
 %% @private
 
 update_durable(UpdateFun, FilterFun) ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
-            fun() -> update_durable_in_mnesia(UpdateFun, FilterFun) end
+            fun() -> update_durable_in_mnesia(UpdateFun, FilterFun) end,
+        khepri =>
+            fun() -> update_durable_in_khepri(UpdateFun, FilterFun) end
        }).
 
 update_durable_in_mnesia(UpdateFun, FilterFun) ->
@@ -464,6 +650,63 @@ update_durable_in_mnesia(UpdateFun, FilterFun) ->
           end),
     ok.
 
+update_durable_in_khepri(UpdateFun, FilterFun) ->
+    PathPattern = khepri_queues_path() ++
+        [?KHEPRI_WILDCARD_STAR,
+         #if_data_matches{
+            pattern = amqqueue:pattern_match_on_durable(true)}],
+    %% The `FilterFun' or `UpdateFun' might attempt to do something
+    %% incompatible with Khepri transactions (such as dynamic apply, sending
+    %% a message, etc.), so this function cannot be written as a regular
+    %% transaction. Instead we can get all queues and track their versions,
+    %% update them, then apply the updates in a transaction, failing if any
+    %% queue has changed since reading the queue record.
+    case rabbit_khepri:adv_get_many(PathPattern) of
+        {ok, Props} ->
+            Updates = maps:fold(
+                        fun(Path0, #{data := Q0, payload_version := Vsn}, Acc)
+                            when ?is_amqqueue(Q0) ->
+                                case FilterFun(Q0) of
+                                    true ->
+                                        Path = khepri_path:combine_with_conditions(
+                                                 Path0,
+                                                 [#if_payload_version{version = Vsn}]),
+                                        Q = UpdateFun(Q0),
+                                        [{Path, Q} | Acc];
+                                    false ->
+                                        Acc
+                                end
+                        end, [], Props),
+            Res = rabbit_khepri:transaction(
+                    fun() ->
+                            for_each_while_ok(
+                              fun({Path, Q}) -> khepri_tx:put(Path, Q) end,
+                              Updates)
+                    end),
+            case Res of
+                ok ->
+                    ok;
+                {error, {khepri, mismatching_node, _}} ->
+                    %% One of the queues changed while attempting to update
+                    %% all queues. Retry the operation.
+                    update_durable_in_khepri(UpdateFun, FilterFun);
+                {error, _} = Error ->
+                    Error
+            end;
+        {error, _} = Error ->
+            Error
+    end.
+
+for_each_while_ok(Fun, [Elem | Rest]) ->
+    case Fun(Elem) of
+        ok ->
+            for_each_while_ok(Fun, Rest);
+        {error, _} = Error ->
+            Error
+    end;
+for_each_while_ok(_, []) ->
+    ok.
+
 %% -------------------------------------------------------------------
 %% exists().
 %% -------------------------------------------------------------------
@@ -478,13 +721,17 @@ update_durable_in_mnesia(UpdateFun, FilterFun) ->
 %% @private
 
 exists(QName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> exists_in_mnesia(QName) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> exists_in_mnesia(QName) end,
+        khepri => fun() -> exists_in_khepri(QName) end
        }).
 
 exists_in_mnesia(QName) ->
     ets:member(?MNESIA_TABLE, QName).
 
+exists_in_khepri(QName) ->
+    ets:member(?KHEPRI_PROJECTION, QName).
+
 %% -------------------------------------------------------------------
 %% exists().
 %% -------------------------------------------------------------------
@@ -501,8 +748,9 @@ exists_in_mnesia(QName) ->
 %% @private
 
 consistent_exists(QName) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> consistent_exists_in_mnesia(QName) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> consistent_exists_in_mnesia(QName) end,
+        khepri => fun() -> exists_in_khepri(QName) end
        }).
 
 consistent_exists_in_mnesia(QName) ->
@@ -527,13 +775,17 @@ consistent_exists_in_mnesia(QName) ->
 
 get_all_by_type(Type) ->
     Pattern = amqqueue:pattern_match_on_type(Type),
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_by_pattern_in_mnesia(Pattern) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_by_pattern_in_mnesia(Pattern) end,
+        khepri => fun() -> get_all_by_pattern_in_khepri(Pattern) end
        }).
 
 get_all_by_pattern_in_mnesia(Pattern) ->
     rabbit_db:list_in_mnesia(?MNESIA_TABLE, Pattern).
 
+get_all_by_pattern_in_khepri(Pattern) ->
+    rabbit_db:list_in_khepri(khepri_queues_path() ++ [rabbit_khepri:if_has_data([?KHEPRI_WILDCARD_STAR_STAR, #if_data_matches{pattern = Pattern}])]).
+
 %% -------------------------------------------------------------------
 %% get_all_by_type_and_node().
 %% -------------------------------------------------------------------
@@ -551,8 +803,9 @@ get_all_by_pattern_in_mnesia(Pattern) ->
 %% @private
 
 get_all_by_type_and_node(VHostName, Type, Node) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_by_type_and_node_in_mnesia(VHostName, Type, Node) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_by_type_and_node_in_mnesia(VHostName, Type, Node) end,
+        khepri => fun() -> get_all_by_type_and_node_in_khepri(VHostName, Type, Node) end
        }).
 
 get_all_by_type_and_node_in_mnesia(VHostName, Type, Node) ->
@@ -564,6 +817,11 @@ get_all_by_type_and_node_in_mnesia(VHostName, Type, Node) ->
                                 amqqueue:qnode(Q) == Node]))
       end).
 
+get_all_by_type_and_node_in_khepri(VHostName, Type, Node) ->
+    Pattern = amqqueue:pattern_match_on_type(Type),
+    Qs = rabbit_db:list_in_khepri(khepri_queues_path() ++ [VHostName, rabbit_khepri:if_has_data([?KHEPRI_WILDCARD_STAR_STAR, #if_data_matches{pattern = Pattern}])]),
+    [Q || Q <- Qs, amqqueue:qnode(Q) == Node].
+
 %% -------------------------------------------------------------------
 %% create_or_get().
 %% -------------------------------------------------------------------
@@ -574,13 +832,14 @@ get_all_by_type_and_node_in_mnesia(VHostName, Type, Node) ->
 %% @doc Writes a queue record if it doesn't exist already or returns the existing one
 %%
 %% @returns the existing record if there is one in the database already, or the newly
-%% created record. 
+%% created record.
 %%
 %% @private
 
 create_or_get(Q) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> create_or_get_in_mnesia(Q) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> create_or_get_in_mnesia(Q) end,
+        khepri => fun() -> create_or_get_in_khepri(Q) end
        }).
 
 create_or_get_in_mnesia(Q) ->
@@ -602,6 +861,18 @@ create_or_get_in_mnesia(Q) ->
               end
       end).
 
+create_or_get_in_khepri(Q) ->
+    QueueName = amqqueue:get_name(Q),
+    Path = khepri_queue_path(QueueName),
+    case rabbit_khepri:adv_create(Path, Q) of
+        {error, {khepri, mismatching_node, #{node_props := #{data := ExistingQ}}}} ->
+            {existing, ExistingQ};
+        {ok, _} ->
+            {created, Q};
+        Error ->
+            Error
+    end.
+
 %% -------------------------------------------------------------------
 %% set().
 %% -------------------------------------------------------------------
@@ -615,8 +886,9 @@ create_or_get_in_mnesia(Q) ->
 %% @private
 
 set(Q) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> set_in_mnesia(Q) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> set_in_mnesia(Q) end,
+        khepri => fun() -> set_in_khepri(Q) end
        }).
 
 set_in_mnesia(Q) ->
@@ -627,7 +899,7 @@ set_in_mnesia(Q) ->
       end).
 
 set_in_mnesia_tx(DurableQ, Q) ->
-    case ?amqqueue_is_durable(Q) of
+    case amqqueue:is_durable(Q) of
         true ->
             ok = mnesia:write(?MNESIA_DURABLE_TABLE, DurableQ, write);
         false ->
@@ -635,6 +907,10 @@ set_in_mnesia_tx(DurableQ, Q) ->
     end,
     ok = mnesia:write(?MNESIA_TABLE, Q, write).
 
+set_in_khepri(Q) ->
+    Path = khepri_queue_path(amqqueue:get_name(Q)),
+    rabbit_khepri:put(Path, Q).
+
 %% -------------------------------------------------------------------
 %% set_many().
 %% -------------------------------------------------------------------
@@ -642,14 +918,16 @@ set_in_mnesia_tx(DurableQ, Q) ->
 -spec set_many([Queue]) -> ok when
       Queue :: amqqueue:amqqueue().
 %% @doc Writes a list of durable queue records.
-%% It is responsibility of the calling function to ensure all records are durable.
-%% Once transient entities are deprecated, this is a non-issue.
+%%
+%% It is responsibility of the calling function to ensure all records are
+%% durable.
 %%
 %% @private
 
 set_many(Qs) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> set_many_in_mnesia(Qs) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> set_many_in_mnesia(Qs) end,
+        khepri => fun() -> set_many_in_khepri(Qs) end
        }).
 
 set_many_in_mnesia(Qs) ->
@@ -657,11 +935,28 @@ set_many_in_mnesia(Qs) ->
         %% Just to be nested in forget_node_for_queue
         mnesia:transaction(
           fun() ->
-                  [ok = mnesia:write(?MNESIA_DURABLE_TABLE, Q, write) || Q <- Qs],
+                  [begin
+                       true = amqqueue:is_durable(Q),
+                       ok = mnesia:write(?MNESIA_DURABLE_TABLE, Q, write)
+                   end || Q <- Qs],
                   ok
           end),
     ok.
 
+set_many_in_khepri(Qs) ->
+    rabbit_khepri:transaction(
+      fun() ->
+              [begin
+                   true = amqqueue:is_durable(Q),
+                   Path = khepri_queue_path(amqqueue:get_name(Q)),
+                   case khepri_tx:put(Path, Q) of
+                       ok      -> ok;
+                       Error   -> khepri_tx:abort(Error)
+                   end
+               end || Q <- Qs]
+      end),
+    ok.
+
 %% -------------------------------------------------------------------
 %% delete_transient().
 %% -------------------------------------------------------------------
@@ -676,8 +971,9 @@ set_many_in_mnesia(Qs) ->
 %% @private
 
 delete_transient(FilterFun) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> delete_transient_in_mnesia(FilterFun) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> delete_transient_in_mnesia(FilterFun) end,
+        khepri => fun() -> delete_transient_in_khepri(FilterFun) end
        }).
 
 delete_transient_in_mnesia(FilterFun) ->
@@ -719,6 +1015,41 @@ partition_queues([Q0,Q1,Q2,Q3,Q4,Q5,Q6,Q7,Q8,Q9 | T]) ->
 partition_queues(T) ->
     [T].
 
+delete_transient_in_khepri(FilterFun) ->
+    PathPattern = khepri_queues_path() ++
+        [?KHEPRI_WILDCARD_STAR,
+         #if_data_matches{
+            pattern = amqqueue:pattern_match_on_durable(false)}],
+    %% The `FilterFun' might try to determine if the queue's process is alive.
+    %% This can cause a `calling_self' exception if we use the `FilterFun'
+    %% within the function passed to `khepri:fold/5' since the Khepri server
+    %% process might call itself. Instead we can fetch all of the transient
+    %% queues with `get_many' and then filter and fold the results outside of
+    %% Khepri's Ra server process.
+    case rabbit_khepri:get_many(PathPattern) of
+        {ok, Qs} ->
+            Items = maps:fold(
+                     fun(Path, Queue, Acc) when ?is_amqqueue(Queue) ->
+                             case FilterFun(Queue) of
+                                 true ->
+                                     QueueName = khepri_queue_path_to_name(
+                                                   Path),
+                                     case delete_in_khepri(QueueName, false) of
+                                         ok ->
+                                             Acc;
+                                         Deletions ->
+                                             [{QueueName, Deletions} | Acc]
+                                     end;
+                                 false ->
+                                     Acc
+                             end
+                     end, [], Qs),
+            {QueueNames, Deletions} = lists:unzip(Items),
+            {QueueNames, lists:flatten(Deletions)};
+        {error, _} = Error ->
+            Error
+    end.
+
 %% -------------------------------------------------------------------
 %% foreach_transient().
 %% -------------------------------------------------------------------
@@ -731,8 +1062,9 @@ partition_queues(T) ->
 %% @private
 
 foreach_transient(UpdateFun) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> foreach_transient_in_mnesia(UpdateFun) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> foreach_transient_in_mnesia(UpdateFun) end,
+        khepri => fun() -> foreach_transient_in_khepri(UpdateFun) end
        }).
 
 foreach_transient_in_mnesia(UpdateFun) ->
@@ -744,6 +1076,27 @@ foreach_transient_in_mnesia(UpdateFun) ->
               ok
       end).
 
+foreach_transient_in_khepri(UpdateFun) ->
+    PathPattern = khepri_queues_path() ++
+        [?KHEPRI_WILDCARD_STAR,
+         #if_data_matches{
+            pattern = amqqueue:pattern_match_on_durable(false)}],
+    %% The `UpdateFun' might try to determine if the queue's process is alive.
+    %% This can cause a `calling_self' exception if we use the `UpdateFun'
+    %% within the function passed to `khepri:fold/5' since the Khepri server
+    %% process might call itself. Instead we can fetch all of the transient
+    %% queues with `get_many' and then filter and fold the results outside of
+    %% Khepri's Ra server process.
+    case rabbit_khepri:get_many(PathPattern) of
+        {ok, Qs} ->
+            maps:foreach(
+              fun(_Path, Queue) when ?is_amqqueue(Queue) ->
+                      UpdateFun(Queue)
+              end, Qs);
+        {error, _} = Error ->
+            Error
+    end.
+
 %% -------------------------------------------------------------------
 %% foreach_durable().
 %% -------------------------------------------------------------------
@@ -756,9 +1109,9 @@ foreach_transient_in_mnesia(UpdateFun) ->
 %% @private
 
 foreach_durable(UpdateFun, FilterFun) ->
-    rabbit_db:run(
-      #{mnesia =>
-            fun() -> foreach_durable_in_mnesia(UpdateFun, FilterFun) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> foreach_durable_in_mnesia(UpdateFun, FilterFun) end,
+        khepri => fun() -> foreach_durable_in_khepri(UpdateFun, FilterFun) end
        }).
 
 foreach_durable_in_mnesia(UpdateFun, FilterFun) ->
@@ -772,7 +1125,22 @@ foreach_durable_in_mnesia(UpdateFun, FilterFun) ->
                   _ = [UpdateFun(Q) || Q <- Qs, FilterFun(Q)],
                   ok
           end),
-    ok.    
+    ok.
+
+foreach_durable_in_khepri(UpdateFun, FilterFun) ->
+    Path = khepri_queues_path() ++
+        [?KHEPRI_WILDCARD_STAR,
+         #if_data_matches{
+            pattern = amqqueue:pattern_match_on_durable(true)}],
+    case rabbit_khepri:filter(Path, fun(_, #{data := Q}) ->
+                                            FilterFun(Q)
+                                    end) of
+        {ok, Qs} ->
+            _ = [UpdateFun(Q) || Q <- maps:values(Qs)],
+            ok;
+        Error ->
+            Error
+    end.
 
 %% -------------------------------------------------------------------
 %% set_dirty().
@@ -785,8 +1153,9 @@ foreach_durable_in_mnesia(UpdateFun, FilterFun) ->
 %% @private
 
 set_dirty(Q) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> set_dirty_in_mnesia(Q) end
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> set_dirty_in_mnesia(Q) end,
+        khepri => ok
        }).
 
 set_dirty_in_mnesia(Q) ->
@@ -817,6 +1186,27 @@ update_in_mnesia_tx(Name, Fun) ->
             not_found
     end.
 
+%% -------------------------------------------------------------------
+%% update_in_khepri_tx().
+%% -------------------------------------------------------------------
+
+-spec update_in_khepri_tx(QName, UpdateFun) -> Ret when
+      QName :: rabbit_amqqueue:name(),
+      Queue :: amqqueue:amqqueue(),
+      UpdateFun :: fun((Queue) -> Queue),
+      Ret :: Queue | not_found.
+
+update_in_khepri_tx(Name, Fun) ->
+    Path = khepri_queue_path(Name),
+    case khepri_tx:get(Path) of
+        {ok, Q} ->
+            Q1 = Fun(Q),
+            ok = khepri_tx:put(Path, Q1),
+            Q1;
+        _  ->
+            not_found
+    end.
+
 %% -------------------------------------------------------------------
 %% get_durable_in_mnesia_tx().
 %% -------------------------------------------------------------------
@@ -831,6 +1221,13 @@ get_durable_in_mnesia_tx(Name) ->
         [Q] -> {ok, Q}
     end.
 
+%% TODO this should be internal, it's here because of mirrored queues
+get_in_khepri_tx(Name) ->
+    case khepri_tx:get(khepri_queue_path(Name)) of
+        {ok, X} -> [X];
+        _ -> []
+    end.
+
 %% -------------------------------------------------------------------
 %% clear().
 %% -------------------------------------------------------------------
@@ -841,14 +1238,23 @@ get_durable_in_mnesia_tx(Name) ->
 %% @private
 
 clear() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> clear_in_mnesia() end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> clear_in_mnesia() end,
+        khepri => fun() -> clear_in_khepri() end}).
 
 clear_in_mnesia() ->
     {atomic, ok} = mnesia:clear_table(?MNESIA_TABLE),
     {atomic, ok} = mnesia:clear_table(?MNESIA_DURABLE_TABLE),
     ok.
 
+clear_in_khepri() ->
+    Path = khepri_queues_path(),
+    case rabbit_khepri:delete(Path) of
+        ok -> ok;
+        Error -> throw(Error)
+    end.
+
+%% --------------------------------------------------------------
 %% Internal
 %% --------------------------------------------------------------
 
@@ -883,3 +1289,37 @@ list_with_possible_retry_in_mnesia(Fun) ->
         Ret ->
             Ret
     end.
+
+list_with_possible_retry_in_khepri(Fun) ->
+    %% See equivalent `list_with_possible_retry_in_mnesia` first.
+    %% Not sure how much of this is possible in Khepri, as there is no dirty read,
+    %% but the amqqueue record migration is still happening.
+    %% Let's retry just in case
+    AmqqueueRecordVersion = amqqueue:record_version_to_use(),
+    case Fun() of
+        [] ->
+            case khepri_tx:is_transaction() of
+                true ->
+                    [];
+                false ->
+                    case amqqueue:record_version_to_use() of
+                        AmqqueueRecordVersion -> [];
+                        _                     -> Fun()
+                    end
+            end;
+        Ret ->
+            Ret
+    end.
+
+%% --------------------------------------------------------------
+%% Khepri paths
+%% --------------------------------------------------------------
+
+khepri_queues_path() ->
+    [?MODULE, queues].
+
+khepri_queue_path(#resource{virtual_host = VHost, name = Name}) ->
+    [?MODULE, queues, VHost, Name].
+
+khepri_queue_path_to_name([?MODULE, queues, VHost, Name]) ->
+    rabbit_misc:r(VHost, queue, Name).
