@@ -2,13 +2,14 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_db_user).
 
 -include_lib("stdlib/include/assert.hrl").
 
+-include_lib("khepri/include/khepri.hrl").
 -include_lib("rabbit_common/include/rabbit.hrl").
 
 -export([create/1,
@@ -16,6 +17,7 @@
          get/1,
          get_all/0,
          with_fun_in_mnesia_tx/2,
+         with_fun_in_khepri_tx/2,
          get_user_permissions/2,
          check_and_match_user_permissions/2,
          set_user_permissions/1,
@@ -28,8 +30,44 @@
          clear_matching_topic_permissions/3,
          delete/1]).
 
+-export([khepri_users_path/0,
+         khepri_user_path/1,
+         khepri_user_permission_path/2,
+         khepri_topic_permission_path/3]).
+
+%% for testing
 -export([clear/0]).
 
+-ifdef(TEST).
+-export([get_in_mnesia/1,
+         get_in_khepri/1,
+         create_in_mnesia/2,
+         create_in_khepri/2,
+         get_all_in_mnesia/0,
+         get_all_in_khepri/0,
+         update_in_mnesia/2,
+         update_in_khepri/2,
+         delete_in_mnesia/1,
+         delete_in_khepri/1,
+         get_user_permissions_in_mnesia/2,
+         get_user_permissions_in_khepri/2,
+         set_user_permissions_in_mnesia/3,
+         set_user_permissions_in_khepri/3,
+         set_topic_permissions_in_mnesia/3,
+         set_topic_permissions_in_khepri/3,
+         match_user_permissions_in_mnesia/2,
+         match_user_permissions_in_khepri/2,
+         clear_user_permissions_in_mnesia/2,
+         clear_user_permissions_in_khepri/2,
+         get_topic_permissions_in_mnesia/3,
+         get_topic_permissions_in_khepri/3,
+         match_topic_permissions_in_mnesia/3,
+         match_topic_permissions_in_khepri/3,
+         clear_topic_permissions_in_mnesia/3,
+         clear_topic_permissions_in_khepri/3
+        ]).
+-endif.
+
 -define(MNESIA_TABLE, rabbit_user).
 -define(PERM_MNESIA_TABLE, rabbit_user_permission).
 -define(TOPIC_PERM_MNESIA_TABLE, rabbit_topic_permission).
@@ -49,8 +87,9 @@
 
 create(User) ->
     Username = internal_user:get_username(User),
-    rabbit_db:run(
-      #{mnesia => fun() -> create_in_mnesia(Username, User) end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> create_in_mnesia(Username, User) end,
+        khepri => fun() -> create_in_khepri(Username, User) end}).
 
 create_in_mnesia(Username, User) ->
     rabbit_mnesia:execute_mnesia_transaction(
@@ -62,6 +101,17 @@ create_in_mnesia_tx(Username, User) ->
         _  -> mnesia:abort({user_already_exists, Username})
     end.
 
+create_in_khepri(Username, User) ->
+    Path = khepri_user_path(Username),
+    case rabbit_khepri:create(Path, User) of
+        ok ->
+            ok;
+        {error, {khepri, mismatching_node, _}} ->
+            throw({error, {user_already_exists, Username}});
+        {error, _} = Error ->
+            throw(Error)
+    end.
+
 %% -------------------------------------------------------------------
 %% update().
 %% -------------------------------------------------------------------
@@ -76,8 +126,9 @@ create_in_mnesia_tx(Username, User) ->
 
 update(Username, UpdateFun)
   when is_binary(Username) andalso is_function(UpdateFun, 1) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> update_in_mnesia(Username, UpdateFun) end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> update_in_mnesia(Username, UpdateFun) end,
+        khepri => fun() -> update_in_khepri(Username, UpdateFun) end}).
 
 update_in_mnesia(Username, UpdateFun) ->
     rabbit_mnesia:execute_mnesia_transaction(
@@ -92,6 +143,21 @@ update_in_mnesia_tx(Username, UpdateFun) ->
             mnesia:abort({no_such_user, Username})
     end.
 
+update_in_khepri(Username, UpdateFun) ->
+    rabbit_khepri:transaction(
+      fun () ->
+              Path = khepri_user_path(Username),
+              case khepri_tx:get(Path) of
+                  {ok, User} ->
+                      case khepri_tx:put(Path, UpdateFun(User)) of
+                          ok -> ok;
+                          Error -> khepri_tx:abort(Error)
+                      end;
+                  _ ->
+                      khepri_tx:abort({no_such_user, Username})
+              end
+      end).
+
 %% -------------------------------------------------------------------
 %% get().
 %% -------------------------------------------------------------------
@@ -99,7 +165,6 @@ update_in_mnesia_tx(Username, UpdateFun) ->
 -spec get(Username) -> User | undefined when
       Username :: internal_user:username(),
       User :: internal_user:internal_user().
-
 %% @doc Returns the record of the internal user named `Username'.
 %%
 %% @returns the internal user record or `undefined' if no internal user is named
@@ -108,8 +173,9 @@ update_in_mnesia_tx(Username, UpdateFun) ->
 %% @private
 
 get(Username) when is_binary(Username) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_in_mnesia(Username) end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_in_mnesia(Username) end,
+        khepri => fun() -> get_in_khepri(Username) end}).
 
 get_in_mnesia(Username) ->
     case ets:lookup(?MNESIA_TABLE, Username) of
@@ -117,6 +183,12 @@ get_in_mnesia(Username) ->
         []     -> undefined
     end.
 
+get_in_khepri(Username) ->
+    case ets:lookup(rabbit_khepri_users, Username) of
+        [User] -> User;
+        _      -> undefined
+    end.
+
 %% -------------------------------------------------------------------
 %% get_all().
 %% -------------------------------------------------------------------
@@ -130,14 +202,22 @@ get_in_mnesia(Username) ->
 %% @private
 
 get_all() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> get_all_in_mnesia() end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> get_all_in_mnesia() end,
+        khepri => fun() -> get_all_in_khepri() end}).
 
 get_all_in_mnesia() ->
     mnesia:dirty_match_object(
       ?MNESIA_TABLE,
       internal_user:pattern_match_all()).
 
+get_all_in_khepri() ->
+    Path = khepri_users_path(),
+    case rabbit_khepri:list(Path) of
+        {ok, Users} -> maps:values(Users);
+        _           -> []
+    end.
+
 %% -------------------------------------------------------------------
 %% with_fun_in_*().
 %% -------------------------------------------------------------------
@@ -166,6 +246,16 @@ with_fun_in_mnesia_tx(Username, TxFun)
             end
     end.
 
+with_fun_in_khepri_tx(Username, TxFun)
+  when is_binary(Username) andalso is_function(TxFun, 0) ->
+    fun() ->
+            Path = khepri_user_path(Username),
+            case khepri_tx:exists(Path) of
+                true  -> TxFun();
+                false -> khepri_tx:abort({no_such_user, Username})
+            end
+    end.
+
 %% -------------------------------------------------------------------
 %% get_user_permissions().
 %% -------------------------------------------------------------------
@@ -184,9 +274,11 @@ with_fun_in_mnesia_tx(Username, TxFun)
 
 get_user_permissions(Username, VHostName)
   when is_binary(Username) andalso is_binary(VHostName) ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
-        fun() -> get_user_permissions_in_mnesia(Username, VHostName) end}).
+        fun() -> get_user_permissions_in_mnesia(Username, VHostName) end,
+        khepri =>
+        fun() -> get_user_permissions_in_khepri(Username, VHostName) end}).
 
 get_user_permissions_in_mnesia(Username, VHostName) ->
     Key = #user_vhost{username     = Username,
@@ -196,6 +288,14 @@ get_user_permissions_in_mnesia(Username, VHostName) ->
         []               -> undefined
     end.
 
+get_user_permissions_in_khepri(Username, VHostName) ->
+    UserVHost = #user_vhost{username     = Username,
+                            virtual_host = VHostName},
+    case ets:lookup(rabbit_khepri_user_permissions, UserVHost) of
+        [UserPermission] -> UserPermission;
+        _      -> undefined
+    end.
+
 %% -------------------------------------------------------------------
 %% check_and_match_user_permissions().
 %% -------------------------------------------------------------------
@@ -215,9 +315,11 @@ get_user_permissions_in_mnesia(Username, VHostName) ->
 check_and_match_user_permissions(Username, VHostName)
   when (is_binary(Username) orelse Username =:= '_') andalso
        (is_binary(VHostName) orelse VHostName =:= '_') ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
-        fun() -> match_user_permissions_in_mnesia(Username, VHostName) end}).
+        fun() -> match_user_permissions_in_mnesia(Username, VHostName) end,
+        khepri =>
+        fun() -> match_user_permissions_in_khepri(Username, VHostName) end}).
 
 match_user_permissions_in_mnesia('_' = Username, '_' = VHostName) ->
     rabbit_mnesia:execute_mnesia_transaction(
@@ -257,6 +359,45 @@ match_user_permissions_in_mnesia_tx(Username, VHostName) ->
                        permission = '_'},
       read).
 
+match_user_permissions_in_khepri('_' = _Username, '_' = _VHostName) ->
+    Path = khepri_user_permission_path(?KHEPRI_WILDCARD_STAR, ?KHEPRI_WILDCARD_STAR),
+    case rabbit_khepri:match(Path) of
+        {ok, Map} ->
+            maps:values(Map);
+        _ ->
+            []
+    end;
+match_user_permissions_in_khepri('_' = _Username, VHostName) ->
+    rabbit_khepri:transaction(
+        rabbit_db_vhost:with_fun_in_khepri_tx(
+          VHostName,
+          fun() ->
+                  match_user_permissions_in_khepri_tx(?KHEPRI_WILDCARD_STAR, VHostName)
+          end));
+match_user_permissions_in_khepri(Username, '_' = _VHostName) ->
+    rabbit_khepri:transaction(
+        with_fun_in_khepri_tx(
+          Username,
+          fun() ->
+                  match_user_permissions_in_khepri_tx(Username, ?KHEPRI_WILDCARD_STAR)
+          end));
+match_user_permissions_in_khepri(Username, VHostName) ->
+    rabbit_khepri:transaction(
+        with_fun_in_khepri_tx(
+          Username,
+          rabbit_db_vhost:with_fun_in_khepri_tx(
+            VHostName,
+            fun() ->
+                    match_user_permissions_in_khepri_tx(Username, VHostName)
+            end))).
+
+match_user_permissions_in_khepri_tx(Username, VHostName) ->
+    Path = khepri_user_permission_path(Username, VHostName),
+    case khepri_tx:get_many(Path) of
+        {ok, Map} -> maps:values(Map);
+        _         -> []
+    end.
+
 %% -------------------------------------------------------------------
 %% set_user_permissions().
 %% -------------------------------------------------------------------
@@ -272,11 +413,16 @@ set_user_permissions(
   #user_permission{user_vhost = #user_vhost{username = Username,
                                             virtual_host = VHostName}}
   = UserPermission) ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
         fun() ->
                 set_user_permissions_in_mnesia(
                   Username, VHostName, UserPermission)
+        end,
+        khepri =>
+        fun() ->
+                set_user_permissions_in_khepri(
+                  Username, VHostName, UserPermission)
         end}).
 
 set_user_permissions_in_mnesia(Username, VHostName, UserPermission) ->
@@ -290,6 +436,32 @@ set_user_permissions_in_mnesia(Username, VHostName, UserPermission) ->
 set_user_permissions_in_mnesia_tx(UserPermission) ->
     mnesia:write(?PERM_MNESIA_TABLE, UserPermission, write).
 
+set_user_permissions_in_khepri(Username, VHostName, UserPermission) ->
+    rabbit_khepri:transaction(
+      with_fun_in_khepri_tx(
+        Username,
+        rabbit_db_vhost:with_fun_in_khepri_tx(
+          VHostName,
+          fun() ->
+                  set_user_permissions_in_khepri_tx(Username, VHostName, UserPermission)
+          end)), rw).
+
+set_user_permissions_in_khepri_tx(Username, VHostName, UserPermission) ->
+    Path = khepri_user_permission_path(
+             #if_all{conditions =
+                         [Username,
+                          #if_node_exists{exists = true}]},
+             VHostName),
+    Extra = #{keep_while =>
+                  #{rabbit_db_vhost:khepri_vhost_path(VHostName) =>
+                        #if_node_exists{exists = true}}},
+    Ret = khepri_tx:put(
+            Path, UserPermission, Extra),
+    case Ret of
+        ok      -> ok;
+        Error   -> khepri_tx:abort(Error)
+    end.
+
 %% -------------------------------------------------------------------
 %% clear_user_permissions().
 %% -------------------------------------------------------------------
@@ -304,9 +476,12 @@ set_user_permissions_in_mnesia_tx(UserPermission) ->
 
 clear_user_permissions(Username, VHostName)
   when is_binary(Username) andalso is_binary(VHostName) ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
-        fun() -> clear_user_permissions_in_mnesia(Username, VHostName) end}).
+        fun() -> clear_user_permissions_in_mnesia(Username, VHostName) end,
+        khepri =>
+        fun() -> clear_user_permissions_in_khepri(Username, VHostName) end
+       }).
 
 clear_user_permissions_in_mnesia(Username, VHostName) ->
     rabbit_mnesia:execute_mnesia_transaction(
@@ -317,6 +492,13 @@ clear_user_permissions_in_mnesia_tx(Username, VHostName) ->
                    #user_vhost{username     = Username,
                                virtual_host = VHostName}}).
 
+clear_user_permissions_in_khepri(Username, VHostName) ->
+    Path = khepri_user_permission_path(Username, VHostName),
+    case rabbit_khepri:delete(Path) of
+        ok      -> ok;
+        Error   -> khepri_tx:abort(Error)
+    end.
+
 %% -------------------------------------------------------------------
 %% clear_matching_user_permissions().
 %% -------------------------------------------------------------------
@@ -324,27 +506,31 @@ clear_user_permissions_in_mnesia_tx(Username, VHostName) ->
 -spec clear_matching_user_permissions(Username, VHostName) -> Ret when
       Username :: internal_user:username() | '_',
       VHostName :: vhost:name() | '_',
-      Ret :: [#user_permission{}].
+      Ret :: ok.
 %% @doc Clears all user permissions matching arguments.
 %%
-%% @returns a list of matching user permissions.
-%%
 %% @private
 
 clear_matching_user_permissions(Username, VHostName)
   when (is_binary(Username) orelse Username =:= '_') andalso
        (is_binary(VHostName) orelse VHostName =:= '_') ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
         fun() ->
                 clear_matching_user_permissions_in_mnesia(Username, VHostName)
+        end,
+        khepri =>
+        fun() ->
+                clear_matching_user_permissions_in_khepri(Username, VHostName)
         end
        }).
 
 clear_matching_user_permissions_in_mnesia(Username, VHostName) ->
     rabbit_mnesia:execute_mnesia_transaction(
       fun() ->
-              clear_matching_user_permissions_in_mnesia_tx( Username, VHostName)
+              _ = clear_matching_user_permissions_in_mnesia_tx(
+                    Username, VHostName),
+              ok
       end).
 
 clear_matching_user_permissions_in_mnesia_tx(Username, VHostName) ->
@@ -355,6 +541,13 @@ clear_matching_user_permissions_in_mnesia_tx(Username, VHostName) ->
      || #user_permission{user_vhost = Key} = Record
         <- match_user_permissions_in_mnesia_tx(Username, VHostName)].
 
+clear_matching_user_permissions_in_khepri(Username, VHostName) ->
+    Path = khepri_user_permission_path(any(Username), any(VHostName)),
+    ok = rabbit_khepri:delete(Path).
+
+any('_') -> ?KHEPRI_WILDCARD_STAR;
+any(Value) -> Value.
+
 %% -------------------------------------------------------------------
 %% get_topic_permissions().
 %% -------------------------------------------------------------------
@@ -376,11 +569,16 @@ get_topic_permissions(Username, VHostName, ExchangeName)
   when is_binary(Username) andalso
        is_binary(VHostName) andalso
        is_binary(ExchangeName) ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
         fun() ->
                 get_topic_permissions_in_mnesia(
                   Username, VHostName, ExchangeName)
+        end,
+        khepri =>
+        fun() ->
+                get_topic_permissions_in_khepri(
+                  Username, VHostName, ExchangeName)
         end}).
 
 get_topic_permissions_in_mnesia(Username, VHostName, ExchangeName) ->
@@ -393,6 +591,13 @@ get_topic_permissions_in_mnesia(Username, VHostName, ExchangeName) ->
         []                -> undefined
     end.
 
+get_topic_permissions_in_khepri(Username, VHostName, ExchangeName) ->
+    Path = khepri_topic_permission_path(Username, VHostName, ExchangeName),
+    case rabbit_khepri:get(Path) of
+        {ok, TopicPermission} -> TopicPermission;
+        _                     -> undefined
+    end.
+
 %% -------------------------------------------------------------------
 %% check_and_match_topic_permissions().
 %% -------------------------------------------------------------------
@@ -415,12 +620,18 @@ check_and_match_topic_permissions(Username, VHostName, ExchangeName)
   when (is_binary(Username) orelse Username =:= '_') andalso
        (is_binary(VHostName) orelse VHostName =:= '_') andalso
        (is_binary(ExchangeName) orelse ExchangeName =:= '_') ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
         fun() ->
                 match_topic_permissions_in_mnesia(
                   Username, VHostName, ExchangeName)
-        end}).
+        end,
+        khepri =>
+        fun() ->
+                match_topic_permissions_in_khepri(
+                  Username, VHostName, ExchangeName)
+        end
+       }).
 
 match_topic_permissions_in_mnesia(
   '_' = Username, '_' = VHostName, ExchangeName) ->
@@ -471,6 +682,48 @@ match_topic_permissions_in_mnesia_tx(Username, VHostName, ExchangeName) ->
          permission = '_'},
       read).
 
+match_topic_permissions_in_khepri('_' = _Username, '_' = _VHostName, ExchangeName) ->
+    rabbit_khepri:transaction(
+      fun() ->
+              match_topic_permissions_in_khepri_tx(
+                ?KHEPRI_WILDCARD_STAR, ?KHEPRI_WILDCARD_STAR, any(ExchangeName))
+      end);
+match_topic_permissions_in_khepri('_' = _Username, VHostName, ExchangeName) ->
+    rabbit_khepri:transaction(
+        rabbit_db_vhost:with_fun_in_khepri_tx(
+          VHostName,
+          fun() ->
+                  match_topic_permissions_in_khepri_tx(
+                    ?KHEPRI_WILDCARD_STAR, VHostName, any(ExchangeName))
+          end));
+match_topic_permissions_in_khepri(
+  Username, '_' = _VHostName, ExchangeName) ->
+    rabbit_khepri:transaction(
+        with_fun_in_khepri_tx(
+          Username,
+          fun() ->
+                  match_topic_permissions_in_khepri_tx(
+                    Username, ?KHEPRI_WILDCARD_STAR, any(ExchangeName))
+          end));
+match_topic_permissions_in_khepri(
+  Username, VHostName, ExchangeName) ->
+    rabbit_khepri:transaction(
+        with_fun_in_khepri_tx(
+          Username,
+          rabbit_db_vhost:with_fun_in_khepri_tx(
+            VHostName,
+            fun() ->
+                    match_topic_permissions_in_khepri_tx(
+                      Username, VHostName, any(ExchangeName))
+            end))).
+
+match_topic_permissions_in_khepri_tx(Username, VHostName, ExchangeName) ->
+    Path = khepri_topic_permission_path(Username, VHostName, ExchangeName),
+    case khepri_tx:get_many(Path) of
+        {ok, Map} -> maps:values(Map);
+        _         -> []
+    end.
+
 %% -------------------------------------------------------------------
 %% set_topic_permissions().
 %% -------------------------------------------------------------------
@@ -489,12 +742,18 @@ set_topic_permissions(
         user_vhost = #user_vhost{username = Username,
                                  virtual_host = VHostName}}}
   = TopicPermission) ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
         fun() ->
                 set_topic_permissions_in_mnesia(
                   Username, VHostName, TopicPermission)
-        end}).
+        end,
+        khepri =>
+            fun() ->
+                    set_topic_permissions_in_khepri(
+                      Username, VHostName, TopicPermission)
+            end
+       }).
 
 set_topic_permissions_in_mnesia(Username, VHostName, TopicPermission) ->
     rabbit_mnesia:execute_mnesia_transaction(
@@ -509,6 +768,34 @@ set_topic_permissions_in_mnesia(Username, VHostName, TopicPermission) ->
 set_topic_permissions_in_mnesia_tx(TopicPermission) ->
     mnesia:write(?TOPIC_PERM_MNESIA_TABLE, TopicPermission, write).
 
+set_topic_permissions_in_khepri(Username, VHostName, TopicPermission) ->
+    rabbit_khepri:transaction(
+      with_fun_in_khepri_tx(
+        Username,
+        rabbit_db_vhost:with_fun_in_khepri_tx(
+          VHostName,
+          fun() ->
+                  set_topic_permissions_in_khepri_tx(Username, VHostName, TopicPermission)
+          end)), rw).
+
+set_topic_permissions_in_khepri_tx(Username, VHostName, TopicPermission) ->
+    #topic_permission{topic_permission_key =
+                          #topic_permission_key{exchange = ExchangeName}} = TopicPermission,
+    Path = khepri_topic_permission_path(
+             #if_all{conditions =
+                         [Username,
+                          #if_node_exists{exists = true}]},
+             VHostName,
+             ExchangeName),
+    Extra = #{keep_while =>
+                  #{rabbit_db_vhost:khepri_vhost_path(VHostName) =>
+                        #if_node_exists{exists = true}}},
+    Ret = khepri_tx:put(Path, TopicPermission, Extra),
+    case Ret of
+        ok -> ok;
+        Error   -> khepri_tx:abort(Error)
+    end.
+
 %% -------------------------------------------------------------------
 %% clear_topic_permissions().
 %% -------------------------------------------------------------------
@@ -525,12 +812,18 @@ set_topic_permissions_in_mnesia_tx(TopicPermission) ->
 clear_topic_permissions(Username, VHostName, ExchangeName)
   when is_binary(Username) andalso is_binary(VHostName) andalso
        (is_binary(ExchangeName) orelse ExchangeName =:= '_') ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
         fun() ->
                 clear_topic_permissions_in_mnesia(
                   Username, VHostName, ExchangeName)
-        end}).
+        end,
+        khepri =>
+            fun() ->
+                    clear_topic_permissions_in_khepri(
+                      Username, VHostName, ExchangeName)
+            end
+       }).
 
 clear_topic_permissions_in_mnesia(Username, VHostName, ExchangeName) ->
     rabbit_mnesia:execute_mnesia_transaction(
@@ -542,6 +835,10 @@ clear_topic_permissions_in_mnesia(Username, VHostName, ExchangeName) ->
 clear_topic_permissions_in_mnesia_tx(Username, VHostName, ExchangeName) ->
     delete_topic_permission_in_mnesia_tx(Username, VHostName, ExchangeName).
 
+clear_topic_permissions_in_khepri(Username, VHostName, ExchangeName) ->
+    Path = khepri_topic_permission_path(any(Username), any(VHostName), any(ExchangeName)),
+    rabbit_khepri:delete(Path).
+
 %% -------------------------------------------------------------------
 %% clear_matching_topic_permissions().
 %% -------------------------------------------------------------------
@@ -551,30 +848,35 @@ clear_topic_permissions_in_mnesia_tx(Username, VHostName, ExchangeName) ->
       Username :: rabbit_types:username() | '_',
       VHostName :: vhost:name() | '_',
       ExchangeName :: binary() | '_',
-      Ret :: [#topic_permission{}].
+      Ret :: ok.
 %% @doc Clears all topic permissions matching arguments.
 %%
-%% @returns a list of matching topic permissions.
-%%
 %% @private
 
 clear_matching_topic_permissions(Username, VHostName, ExchangeName)
   when (is_binary(Username) orelse Username =:= '_') andalso
        (is_binary(VHostName) orelse VHostName =:= '_') andalso
        (is_binary(ExchangeName) orelse ExchangeName =:= '_') ->
-    rabbit_db:run(
+    rabbit_khepri:handle_fallback(
       #{mnesia =>
         fun() ->
                 clear_matching_topic_permissions_in_mnesia(
                   Username, VHostName, ExchangeName)
-        end}).
+        end,
+        khepri =>
+        fun() ->
+                clear_matching_topic_permissions_in_khepri(
+                  Username, VHostName, ExchangeName)
+        end
+       }).
 
 clear_matching_topic_permissions_in_mnesia(
   Username, VHostName, ExchangeName) ->
     rabbit_mnesia:execute_mnesia_transaction(
       fun() ->
-              clear_matching_topic_permissions_in_mnesia_tx(
-                Username, VHostName, ExchangeName)
+              _ = clear_matching_topic_permissions_in_mnesia_tx(
+                    Username, VHostName, ExchangeName),
+              ok
       end).
 
 clear_matching_topic_permissions_in_mnesia_tx(
@@ -587,13 +889,19 @@ clear_matching_topic_permissions_in_mnesia_tx(
         <- match_topic_permissions_in_mnesia_tx(
              Username, VHostName, ExchangeName)].
 
+clear_matching_topic_permissions_in_khepri(
+  Username, VHostName, ExchangeName) ->
+    Path = khepri_topic_permission_path(
+             any(Username), any(VHostName), any(ExchangeName)),
+    ok = rabbit_khepri:delete(Path).
+
 %% -------------------------------------------------------------------
 %% delete().
 %% -------------------------------------------------------------------
 
 -spec delete(Username) -> Existed when
       Username :: internal_user:username(),
-      Existed :: boolean().
+      Existed :: boolean() | {error, any()}.
 %% @doc Deletes a user and its permissions from the database.
 %%
 %% @returns a boolean indicating if the user existed. It throws an exception
@@ -602,8 +910,10 @@ clear_matching_topic_permissions_in_mnesia_tx(
 %% @private
 
 delete(Username) when is_binary(Username) ->
-    rabbit_db:run(
-      #{mnesia => fun() -> delete_in_mnesia(Username) end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> delete_in_mnesia(Username) end,
+        khepri => fun() -> delete_in_khepri(Username) end
+       }).
 
 delete_in_mnesia(Username) ->
     rabbit_mnesia:execute_mnesia_transaction(
@@ -631,6 +941,14 @@ delete_topic_permission_in_mnesia_tx(Username, VHostName, ExchangeName) ->
          R <- mnesia:match_object(?TOPIC_PERM_MNESIA_TABLE, Pattern, write)],
     ok.
 
+delete_in_khepri(Username) ->
+    Path = khepri_user_path(Username),
+    case rabbit_khepri:delete_or_fail(Path) of
+        ok -> true;
+        {error, {node_not_found, _}} -> false;
+        Error -> Error
+    end.
+
 user_permission_pattern(Username, VHostName) ->
     #user_permission{user_vhost = #user_vhost{
                                      username = Username,
@@ -657,11 +975,32 @@ topic_permission_pattern(Username, VHostName, ExchangeName) ->
 %% @private
 
 clear() ->
-    rabbit_db:run(
-      #{mnesia => fun() -> clear_in_mnesia() end}).
+    rabbit_khepri:handle_fallback(
+      #{mnesia => fun() -> clear_in_mnesia() end,
+        khepri => fun() -> clear_in_khepri() end}).
 
 clear_in_mnesia() ->
     {atomic, ok} = mnesia:clear_table(?MNESIA_TABLE),
     {atomic, ok} = mnesia:clear_table(?PERM_MNESIA_TABLE),
     {atomic, ok} = mnesia:clear_table(?TOPIC_PERM_MNESIA_TABLE),
     ok.
+
+clear_in_khepri() ->
+    Path = khepri_users_path(),
+    case rabbit_khepri:delete(Path) of
+        ok    -> ok;
+        Error -> throw(Error)
+    end.
+
+%% --------------------------------------------------------------
+%% Paths
+%% --------------------------------------------------------------
+
+khepri_users_path()        -> [?MODULE, users].
+khepri_user_path(Username) -> [?MODULE, users, Username].
+
+khepri_user_permission_path(Username, VHostName) ->
+    [?MODULE, users, Username, user_permissions, VHostName].
+
+khepri_topic_permission_path(Username, VHostName, Exchange) ->
+    [?MODULE, users, Username, topic_permissions, VHostName, Exchange].
