@@ -2,11 +2,11 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2021-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 %% @author The RabbitMQ team
-%% @copyright 2023 VMware, Inc. or its affiliates.
+%% @copyright 2007-2024 Broadcom. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 %% @doc
 %% The feature flag controller is responsible for synchronization and managing
@@ -31,6 +31,8 @@
 
 -include_lib("rabbit_common/include/logging.hrl").
 
+-include("src/rabbit_feature_flags.hrl").
+
 -export([is_supported/1, is_supported/2,
          enable/1,
          enable_default/0,
@@ -46,7 +48,9 @@
          all_nodes/0,
          running_nodes/0,
          collect_inventory_on_nodes/1, collect_inventory_on_nodes/2,
-         mark_as_enabled_on_nodes/4]).
+         mark_as_enabled_on_nodes/4,
+         wait_for_task_and_stop/0,
+         is_running/0]).
 
 %% gen_statem callbacks.
 -export([callback_mode/0,
@@ -61,6 +65,8 @@
 -record(?MODULE, {from,
                   notify = #{}}).
 
+-type run_callback_error() :: {error, any()}.
+
 -define(LOCAL_NAME, ?MODULE).
 -define(GLOBAL_NAME, {?MODULE, global}).
 
@@ -73,6 +79,15 @@ start() ->
 start_link() ->
     gen_statem:start_link({local, ?LOCAL_NAME}, ?MODULE, none, []).
 
+wait_for_task_and_stop() ->
+    case erlang:whereis(rabbit_sup) of
+        undefined -> gen_statem:stop(?LOCAL_NAME);
+        _         -> rabbit_sup:stop_child(?LOCAL_NAME)
+    end.
+
+is_running() ->
+    is_pid(erlang:whereis(?LOCAL_NAME)).
+
 is_supported(FeatureNames) ->
     is_supported(FeatureNames, ?TIMEOUT).
 
@@ -168,6 +183,10 @@ callback_mode() ->
     state_functions.
 
 init(_Args) ->
+    ?LOG_DEBUG(
+       "Feature flags: controller standing by",
+       #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
+    process_flag(trap_exit, true),
     {ok, standing_by, none}.
 
 standing_by(
@@ -279,8 +298,32 @@ proceed_with_task(refresh_after_app_load) ->
     refresh_after_app_load_task().
 
 terminate(_Reason, _State, _Data) ->
+    %% We block until a possible in-flight operation finishes. This increases
+    %% the chance that an in-flight `enable' operation finishes and records the
+    %% new feature flag state on this node before the `rabbit' application
+    %% stops or this Erlang node exits.
+    wait_for_in_flight_operations(),
     ok.
 
+wait_for_in_flight_operations() ->
+    case register_globally() of
+        yes ->
+            %% We don't unregister so the controller holds the lock until it
+            %% exits.
+            ?LOG_DEBUG(
+               "Feature flags: controller exiting, no in-flight operations",
+               #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
+            ok;
+        no ->
+            ?LOG_DEBUG(
+               "Feature flags: can't stop local controller while there is "
+               "an in-flight operation; waiting for it to finish",
+               #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
+            RequestId = notify_me_when_done(),
+            _ = gen_statem:wait_response(RequestId, infinity),
+            wait_for_in_flight_operations()
+    end.
+
 code_change(_OldVsn, OldState, OldData, _Extra) ->
     {ok, OldState, OldData}.
 
@@ -536,11 +579,13 @@ enable_default_task() ->
             #{feature_flags := FeatureFlags} = Inventory,
             StableFeatureNames =
             maps:fold(
-              fun
-                  (FeatureName, #{stability := stable}, Acc) ->
-                      [FeatureName | Acc];
-                  (_FeatureName, _FeatureProps, Acc) ->
-                      Acc
+              fun(FeatureName, FeatureProps, Acc) ->
+                      Stability = rabbit_feature_flags:get_stability(
+                                    FeatureProps),
+                      case Stability of
+                          stable -> [FeatureName | Acc];
+                          _      -> Acc
+                      end
               end, [], FeatureFlags),
             enable_many(Inventory, StableFeatureNames);
         [] ->
@@ -673,9 +718,32 @@ sync_cluster_task(Nodes) ->
 
     case collect_inventory_on_nodes(Nodes) of
         {ok, Inventory} ->
-            FeatureNames = list_feature_flags_enabled_somewhere(
-                             Inventory, false),
-            enable_many(Inventory, FeatureNames);
+            CantEnable = list_deprecated_features_that_cant_be_denied(
+                           Inventory),
+            case CantEnable of
+                [] ->
+                    FeatureNames = list_feature_flags_enabled_somewhere(
+                                     Inventory, false),
+                    enable_many(Inventory, FeatureNames);
+                _ ->
+                    ?LOG_ERROR(
+                       "Feature flags: the following deprecated features "
+                       "can't be denied because their associated feature "
+                       "is being actively used: ~0tp",
+                       [CantEnable],
+                       #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
+                    lists:foreach(
+                      fun(FeatureName) ->
+                              Warning =
+                              rabbit_deprecated_features:get_warning(
+                                FeatureName),
+                              ?LOG_ERROR(
+                                 "~ts", [Warning],
+                                 #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS})
+                      end, CantEnable),
+                    {error,
+                     {failed_to_deny_deprecated_features, CantEnable}}
+            end;
         Error ->
             Error
     end.
@@ -698,8 +766,8 @@ refresh_after_app_load_task() ->
 
 enable_many(#{states_per_node := _} = Inventory, [FeatureName | Rest]) ->
     case enable_if_supported(Inventory, FeatureName) of
-        ok    -> enable_many(Inventory, Rest);
-        Error -> Error
+        {ok, Inventory1} -> enable_many(Inventory1, Rest);
+        Error            -> Error
     end;
 enable_many(_Inventory, []) ->
     ok.
@@ -707,7 +775,8 @@ enable_many(_Inventory, []) ->
 -spec enable_if_supported(Inventory, FeatureName) -> Ret when
       Inventory :: rabbit_feature_flags:cluster_inventory(),
       FeatureName :: rabbit_feature_flags:feature_name(),
-      Ret :: ok | {error, Reason},
+      Ret :: {ok, NewInventory} | {error, Reason},
+      NewInventory :: rabbit_feature_flags:cluster_inventory(),
       Reason :: term().
 
 enable_if_supported(#{states_per_node := _} = Inventory, FeatureName) ->
@@ -717,10 +786,7 @@ enable_if_supported(#{states_per_node := _} = Inventory, FeatureName) ->
                "Feature flags: `~ts`: supported; continuing",
                [FeatureName],
                #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
-            case lock_registry_and_enable(Inventory, FeatureName) of
-                {ok, _Inventory} -> ok;
-                Error            -> Error
-            end;
+            lock_registry_and_enable(Inventory, FeatureName);
         false ->
             ?LOG_DEBUG(
                "Feature flags: `~ts`: unsupported; aborting",
@@ -910,15 +976,22 @@ update_feature_state_and_enable(
                     case Ret2 of
                         {ok, Inventory2} ->
                             post_enable(
-                              Inventory2, FeatureName, NodesWhereDisabled),
+                              Inventory2, FeatureName, NodesWhereDisabled,
+                              true),
                             Ret2;
                         Error ->
+                            post_enable(
+                              Inventory1, FeatureName, NodesWhereDisabled,
+                              false),
                             restore_feature_flag_state(
-                              Nodes, NodesWhereDisabled, Inventory,
+                              Nodes, NodesWhereDisabled, Inventory1,
                               FeatureName),
                             Error
                     end;
                 Error ->
+                    post_enable(
+                      Inventory, FeatureName, NodesWhereDisabled,
+                      false),
                     restore_feature_flag_state(
                       Nodes, NodesWhereDisabled, Inventory, FeatureName),
                     Error
@@ -937,7 +1010,8 @@ restore_feature_flag_state(
        "it:~n"
        "  enabled on:  ~0p~n"
        "  disabled on: ~0p",
-       [FeatureName, NodesWhereEnabled, NodesWhereDisabled]),
+       [FeatureName, NodesWhereEnabled, NodesWhereDisabled],
+       #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
     _ = mark_as_enabled_on_nodes(
           NodesWhereEnabled, Inventory, FeatureName, true),
     _ = mark_as_enabled_on_nodes(
@@ -969,16 +1043,28 @@ do_enable(#{states_per_node := _} = Inventory, FeatureName, Nodes) ->
             Error
     end.
 
--spec post_enable(Inventory, FeatureName, Nodes) -> Ret when
+-spec post_enable(Inventory, FeatureName, Nodes, Enabled) -> Ret when
       Inventory :: rabbit_feature_flags:cluster_inventory(),
       FeatureName :: rabbit_feature_flags:feature_name(),
       Nodes :: [node()],
+      Enabled :: boolean(),
       Ret :: ok.
 
-post_enable(#{states_per_node := _}, FeatureName, Nodes) ->
+post_enable(#{states_per_node := _}, FeatureName, Nodes, Enabled) ->
     case rabbit_feature_flags:uses_callbacks(FeatureName) of
         true ->
-            _ = run_callback(Nodes, FeatureName, post_enable, #{}, infinity),
+            ?LOG_DEBUG(
+               "Feature flags: `~ts`: run `post_enable` callback (if any) "
+               "to ~ts",
+               [FeatureName,
+                case Enabled of
+                    true  -> "perform any cleanups";
+                    false -> "rollback any changes"
+                end],
+               #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
+            _ = run_callback(
+                  Nodes, FeatureName, post_enable, #{enabled => Enabled},
+                  infinity),
             ok;
         false ->
             ok
@@ -990,17 +1076,20 @@ post_enable(#{states_per_node := _}, FeatureName, Nodes) ->
 
 -ifndef(TEST).
 all_nodes() ->
-    lists:usort([node() | rabbit_nodes:list_members()]).
+    lists:sort(rabbit_nodes:list_members()).
 
 running_nodes() ->
     lists:usort([node() | rabbit_nodes:list_running()]).
 -else.
 all_nodes() ->
-    RemoteNodes = case rabbit_feature_flags:get_overriden_nodes() of
-                      undefined -> rabbit_nodes:list_members();
-                      Nodes     -> Nodes
-                  end,
-    lists:usort([node() | RemoteNodes]).
+    AllNodes = case rabbit_feature_flags:get_overriden_nodes() of
+                   undefined ->
+                       rabbit_nodes:list_members();
+                   Nodes ->
+                       ?assert(lists:member(node(), Nodes)),
+                       Nodes
+               end,
+    lists:sort(AllNodes).
 
 running_nodes() ->
     RemoteNodes = case rabbit_feature_flags:get_overriden_running_nodes() of
@@ -1028,7 +1117,7 @@ collect_inventory_on_nodes(Nodes, Timeout) ->
     Inventory0 = #{feature_flags => #{},
                    applications_per_node => #{},
                    states_per_node => #{}},
-    Rets = rpc_calls(Nodes, rabbit_ff_registry, inventory, [], Timeout),
+    Rets = inventory_rpcs(Nodes, Timeout),
     maps:fold(
       fun
           (Node,
@@ -1054,10 +1143,51 @@ collect_inventory_on_nodes(Nodes, Timeout) ->
               Error
       end, {ok, Inventory0}, Rets).
 
+inventory_rpcs(Nodes, Timeout) ->
+    %% We must use `rabbit_ff_registry_wrapper' if it is available to avoid
+    %% any deadlock with the Code server. If it is unavailable, we fall back
+    %% to `rabbit_ff_registry'.
+    %%
+    %% See commit aacfa1978e24bcacd8de7d06a7c3c5d9d8bd098e and pull request
+    %% #8155.
+    Rets0 = rpc_calls(
+              Nodes,
+              rabbit_ff_registry_wrapper, inventory, [], Timeout),
+    OlderNodes = maps:fold(
+                   fun
+                       (Node,
+                        {error,
+                         {exception, undef,
+                          [{rabbit_ff_registry_wrapper,_ , _, _} | _]}},
+                        Acc) ->
+                           [Node | Acc];
+                       (_Node, _Ret, Acc) ->
+                           Acc
+                   end, [], Rets0),
+    case OlderNodes of
+        [] ->
+            Rets0;
+        _ ->
+            ?LOG_INFO(
+               "Feature flags: the following nodes run an older version of "
+               "RabbitMQ causing the "
+               "\"rabbit_ff_registry_wrapper:inventory[] undefined\" error "
+               "above: ~2p~n"
+               "Feature flags: falling back to another method for those "
+               "nodes; this may trigger a bug causing them to hang",
+               [lists:sort(OlderNodes)],
+               #{domain => ?RMQLOG_DOMAIN_FEAT_FLAGS}),
+            Rets1 = rpc_calls(
+                      OlderNodes,
+                      rabbit_ff_registry, inventory, [], Timeout),
+            maps:merge(Rets0, Rets1)
+    end.
+
 merge_feature_flags(FeatureFlagsA, FeatureFlagsB) ->
     FeatureFlags = maps:merge(FeatureFlagsA, FeatureFlagsB),
     maps:map(
-      fun(FeatureName, FeatureProps) ->
+      fun
+          (FeatureName, FeatureProps) when ?IS_FEATURE_FLAG(FeatureProps) ->
               %% When we collect feature flag properties from all nodes, we
               %% start with an empty cluster inventory (a common Erlang
               %% recursion pattern). This means that all feature flags are
@@ -1096,7 +1226,28 @@ merge_feature_flags(FeatureFlagsA, FeatureFlagsB) ->
 
               FeatureProps1 = FeatureProps#{stability => Stability},
               FeatureProps2 = maps:remove(callbacks, FeatureProps1),
-              FeatureProps2
+              FeatureProps2;
+          (FeatureName, FeatureProps) when ?IS_DEPRECATION(FeatureProps) ->
+              UnknownProps = #{deprecation_phase => permitted_by_default},
+              FeaturePropsA = maps:get(
+                                FeatureName, FeatureFlagsA, UnknownProps),
+              FeaturePropsB = maps:get(
+                                FeatureName, FeatureFlagsB, UnknownProps),
+
+              PhaseA = rabbit_deprecated_features:get_phase(FeaturePropsA),
+              PhaseB = rabbit_deprecated_features:get_phase(FeaturePropsB),
+              Phase = case {PhaseA, PhaseB} of
+                          {removed, _}           -> removed;
+                          {_, removed}           -> removed;
+                          {disconnected, _}      -> disconnected;
+                          {_, disconnected}      -> disconnected;
+                          {denied_by_default, _} -> denied_by_default;
+                          {_, denied_by_default} -> denied_by_default;
+                          _                      -> permitted_by_default
+                      end,
+
+              FeatureProps1 = FeatureProps#{deprecation_phase => Phase},
+              FeatureProps1
       end, FeatureFlags).
 
 -spec list_feature_flags_enabled_somewhere(Inventory, HandleStateChanging) ->
@@ -1126,6 +1277,41 @@ list_feature_flags_enabled_somewhere(
                      end, #{}, StatesPerNode),
     lists:sort(maps:keys(MergedStates)).
 
+-spec list_deprecated_features_that_cant_be_denied(Inventory) ->
+    Ret when
+      Inventory :: rabbit_feature_flags:cluster_inventory(),
+      Ret :: [FeatureName],
+      FeatureName :: rabbit_feature_flags:feature_name().
+
+list_deprecated_features_that_cant_be_denied(
+  #{feature_flags := FeatureFlags,
+    states_per_node := StatesPerNode}) ->
+    ThisNode = node(),
+    States = maps:get(ThisNode, StatesPerNode),
+
+    maps:fold(
+      fun
+          (FeatureName, true, Acc) ->
+              FeatureProps = maps:get(FeatureName, FeatureFlags),
+              Stability = rabbit_feature_flags:get_stability(FeatureProps),
+              case Stability of
+                  required ->
+                      Acc;
+                  _ ->
+                      #{ThisNode := IsUsed} = run_callback(
+                                                [ThisNode], FeatureName,
+                                                is_feature_used, #{},
+                                                infinity),
+                      case IsUsed of
+                          true   -> [FeatureName | Acc];
+                          false  -> Acc;
+                          _Error -> [FeatureName | Acc]
+                      end
+              end;
+          (_FeatureName, false, Acc) ->
+              Acc
+      end, [], States).
+
 -spec list_nodes_who_know_the_feature_flag(Inventory, FeatureName) ->
     Ret when
       Inventory :: rabbit_feature_flags:cluster_inventory(),
@@ -1197,7 +1383,7 @@ rpc_call(Node, Module, Function, Args, Timeout) ->
         %% time, for instance, if there is a lot of data to migrate.
         error:{erpc, noconnection} = Reason
           when is_integer(Timeout) andalso Timeout > SleepBetweenRetries ->
-            ?LOG_ERROR(
+            ?LOG_WARNING(
                "Feature flags: no connection to node `~ts`; "
                "retrying in ~b milliseconds",
                [Node, SleepBetweenRetries],
@@ -1267,14 +1453,17 @@ rpc_calls(Nodes, Module, Function, Args, Timeout) when is_list(Nodes) ->
 %% Feature flag support queries.
 %% --------------------------------------------------------------------
 
--spec is_known(Inventory, FeatureFlag) -> IsKnown when
+-spec is_known(Inventory, FeatureProps) -> IsKnown when
       Inventory :: rabbit_feature_flags:cluster_inventory(),
-      FeatureFlag :: rabbit_feature_flags:feature_props_extended(),
+      FeatureProps ::
+      rabbit_feature_flags:feature_props_extended() |
+      rabbit_deprecated_features:feature_props_extended(),
       IsKnown :: boolean().
+%% @private
 
 is_known(
   #{applications_per_node := ScannedAppsPerNode},
-  #{provided_by := App} = _FeatureFlag) ->
+  #{provided_by := App} = _FeatureProps) ->
     maps:fold(
       fun
           (_Node, ScannedApps, false) -> lists:member(App, ScannedApps);
@@ -1408,14 +1597,34 @@ enable_dependencies1(
 %% Migration function.
 %% --------------------------------------------------------------------
 
--spec run_callback(Nodes, FeatureName, Command, Extra, Timeout) ->
-    Rets when
+-spec run_callback
+(Nodes, FeatureName, Command, Extra, Timeout) -> Rets when
       Nodes :: [node()],
       FeatureName :: rabbit_feature_flags:feature_name(),
-      Command :: rabbit_feature_flags:callback_name(),
+      Command :: enable,
       Extra :: map(),
       Timeout :: timeout(),
-      Rets :: #{node() => term()}.
+      Rets :: #{node() =>
+                rabbit_feature_flags:enable_callback_ret() |
+                run_callback_error()};
+(Nodes, FeatureName, Command, Extra, Timeout) -> Rets when
+      Nodes :: [node()],
+      FeatureName :: rabbit_feature_flags:feature_name(),
+      Command :: post_enable,
+      Extra :: map(),
+      Timeout :: timeout(),
+      Rets :: #{node() =>
+                rabbit_feature_flags:post_enable_callback_ret() |
+                run_callback_error()};
+(Nodes, FeatureName, Command, Extra, Timeout) -> Rets when
+      Nodes :: [node()],
+      FeatureName :: rabbit_feature_flags:feature_name(),
+      Command :: is_feature_used,
+      Extra :: map(),
+      Timeout :: timeout(),
+      Rets :: #{node() =>
+                rabbit_deprecated_features:is_feature_used_callback_ret() |
+                run_callback_error()}.
 
 run_callback(Nodes, FeatureName, Command, Extra, Timeout) ->
     FeatureProps = rabbit_ff_registry_wrapper:get(FeatureName),
@@ -1443,8 +1652,9 @@ run_callback(Nodes, FeatureName, Command, Extra, Timeout) ->
             %% No callbacks defined for this feature flag. Consider it a
             %% success!
             Ret = case Command of
-                      enable      -> ok;
-                      post_enable -> ok
+                      enable          -> ok;
+                      post_enable     -> ok;
+                      is_feature_used -> false
                   end,
             #{node() => Ret}
     end.
@@ -1454,9 +1664,12 @@ run_callback(Nodes, FeatureName, Command, Extra, Timeout) ->
       Nodes :: [node()],
       CallbackMod :: module(),
       CallbackFun :: atom(),
-      Args :: rabbit_feature_flags:callbacks_args(),
+      Args :: rabbit_feature_flags:callbacks_args() |
+              rabbit_deprecated_features:callbacks_args(),
       Timeout :: timeout(),
-      Rets :: #{node() => rabbit_feature_flags:callbacks_rets()}.
+      Rets :: #{node() => rabbit_feature_flags:callbacks_rets() |
+                          rabbit_deprecated_features:callbacks_rets() |
+                          run_callback_error()}.
 
 do_run_callback(Nodes, CallbackMod, CallbackFun, Args, Timeout) ->
     #{feature_name := FeatureName,
