@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2021 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_logger_exchange_h).
@@ -12,6 +12,7 @@
 -include_lib("rabbit_common/include/rabbit.hrl").
 -include_lib("rabbit_common/include/rabbit_framing.hrl").
 -include_lib("rabbit_common/include/logging.hrl").
+-include_lib("rabbit/include/mc.hrl").
 
 %% logger callbacks
 -export([log/2, adding_handler/1, removing_handler/1, changing_config/3,
@@ -51,11 +52,18 @@ log(LogEvent, Config) ->
 
 do_log(LogEvent, #{config := #{exchange := Exchange}} = Config) ->
     RoutingKey = make_routing_key(LogEvent, Config),
-    AmqpMsg = log_event_to_amqp_msg(LogEvent, Config),
+    PBasic = log_event_to_amqp_msg(LogEvent, Config),
     Body = try_format_body(LogEvent, Config),
-    case rabbit_basic:publish(Exchange, RoutingKey, AmqpMsg, Body) of
-        ok                 -> ok;
-        {error, not_found} -> ok
+    Content = rabbit_basic:build_content(PBasic, Body),
+    case mc_amqpl:message(Exchange, RoutingKey, Content) of
+        {ok, Msg} ->
+            case rabbit_queue_type:publish_at_most_once(Exchange, Msg) of
+                ok -> ok;
+                {error, not_found} -> ok
+            end;
+        {error, _Reason} ->
+            %% it would be good to log this error but can we?
+            ok
     end.
 
 removing_handler(Config) ->
@@ -120,11 +128,27 @@ start_setup_proc(#{config := InternalConfig} = Config) ->
     {ok, DefaultVHost} = application:get_env(rabbit, default_vhost),
     Exchange = rabbit_misc:r(DefaultVHost, exchange, ?LOG_EXCH_NAME),
     InternalConfig1 = InternalConfig#{exchange => Exchange},
-
-    Pid = spawn(fun() -> setup_proc(Config#{config => InternalConfig1}) end),
+    Pid = spawn(fun() ->
+                        wait_for_initial_pass(60),
+                        setup_proc(Config#{config => InternalConfig1})
+                end),
     InternalConfig2 = InternalConfig1#{setup_proc => Pid},
     Config#{config => InternalConfig2}.
 
+%% Declaring an exchange requires the metadata store to be ready
+%% which happens on a boot step after the second phase of the prelaunch.
+%% This function waits for the store initialisation.
+wait_for_initial_pass(0) ->
+    ok;
+wait_for_initial_pass(N) ->
+    case rabbit_db:is_init_finished() of
+        false ->
+            timer:sleep(1000),
+            wait_for_initial_pass(N - 1);
+        true ->
+            ok
+    end.
+
 setup_proc(
   #{config := #{exchange := #resource{name = Name,
                                       virtual_host = VHost}}} = Config) ->
