@@ -2,13 +2,20 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2018-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_maintenance).
 
 -include_lib("rabbit_common/include/rabbit.hrl").
 
+%% FIXME: Ra consistent queries are currently fragile in the sense that the
+%% query function may run on a remote node and the function reference or MFA
+%% may not be valid on that node. That's why consistent queries in this module
+%% are in fact local queries when Khepri is enabled.
+%%
+%% See `rabbit_db_maintenance:get_consistent_in_khepri/1'.
+
 -export([
     is_enabled/0,
     drain/0,
@@ -28,7 +35,7 @@
     random_primary_replica_transfer_candidate_node/2,
     transfer_leadership_of_quorum_queues/1,
     transfer_leadership_of_classic_mirrored_queues/1,
-    boot/0
+    table_definitions/0
 ]).
 
 -define(DEFAULT_STATUS,  regular).
@@ -44,13 +51,8 @@
 %% Boot
 %%
 
--rabbit_boot_step({rabbit_maintenance_mode_state,
-    [{description, "initializes maintenance mode state"},
-        {mfa,         {?MODULE, boot, []}},
-        {requires,    networking}]}).
-
-boot() ->
-    rabbit_db_maintenance:setup_schema().
+table_definitions() ->
+    rabbit_db_maintenance:table_definitions().
 
 %%
 %% API
@@ -85,6 +87,8 @@ drain() ->
         _Pid -> transfer_leadership_of_stream_coordinator(TransferCandidates)
     end,
 
+    transfer_leadership_of_metadata_store(TransferCandidates),
+
     %% allow plugins to react
     rabbit_event:notify(maintenance_draining, #{
         reason => <<"node is being put into maintenance">>
@@ -209,8 +213,20 @@ transfer_leadership_of_quorum_queues(_TransferCandidates) ->
      end || Q <- Queues],
     rabbit_log:info("Leadership transfer for quorum queues hosted on this node has been initiated").
 
+transfer_leadership_of_metadata_store(TransferCandidates) ->
+    rabbit_log:info("Will transfer leadership of metadata store with current leader on this node",
+                    []),
+    case rabbit_khepri:transfer_leadership(TransferCandidates) of
+        {ok, Node} when Node == node(); Node == undefined ->
+            rabbit_log:info("Skipping leadership transfer of metadata store: current leader is not on this node");
+        {ok, Node} ->
+            rabbit_log:info("Leadership transfer for metadata store on this node has been done. The new leader is ~p", [Node]);
+        Error ->
+            rabbit_log:warning("Skipping leadership transfer of metadata store: ~p", [Error])
+    end.
+
 -spec transfer_leadership_of_classic_mirrored_queues([node()]) -> ok.
-%% This function is no longer used by maintanence mode. We retain it in case
+%% This function is no longer used by maintenance mode. We retain it in case
 %% classic mirrored queue leadership transfer would be reconsidered.
 %%
 %% With a lot of CMQs in a cluster, the transfer procedure can take prohibitively long
