@@ -2,13 +2,12 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%%  Copyright (c) 2015-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%%  Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_priority_queue).
 
 -include_lib("rabbit_common/include/rabbit.hrl").
--include_lib("rabbit_common/include/rabbit_framing.hrl").
 -include("amqqueue.hrl").
 
 -behaviour(rabbit_backing_queue).
@@ -414,6 +413,8 @@ info(backing_queue_status, #state{bq = BQ, bqss = BQSs}) ->
           end, nothing, BQSs);
 info(head_message_timestamp, #state{bq = BQ, bqss = BQSs}) ->
     find_head_message_timestamp(BQ, BQSs, '');
+info(oldest_message_received_timestamp, #state{bq = BQ, bqss = BQSs}) ->
+    find_oldest_message_received_timestamp(BQ, BQSs);
 info(online, _) ->
     '';
 info(Item, #state{bq = BQ, bqss = BQSs}) ->
@@ -628,11 +629,8 @@ priority(Priority, MaxP) when is_integer(Priority), Priority =< MaxP ->
     Priority;
 priority(Priority, MaxP) when is_integer(Priority), Priority > MaxP ->
     MaxP;
-priority(#basic_message{content = Content}, MaxP) ->
-    priority(rabbit_binary_parser:ensure_content_decoded(Content), MaxP);
-priority(#content{properties = Props}, MaxP) ->
-    #'P_basic'{priority = Priority0} = Props,
-    priority(Priority0, MaxP).
+priority(Msg, MaxP) ->
+    priority(mc:priority(Msg), MaxP).
 
 add_maybe_infinity(infinity, _) -> infinity;
 add_maybe_infinity(_, infinity) -> infinity;
@@ -656,7 +654,13 @@ priority_on_acktags(P, AckTags) ->
 combine_status(P, New, nothing) ->
     [{priority_lengths, [{P, proplists:get_value(len, New)}]} | New];
 combine_status(P, New, Old) ->
-    Combined = [{K, cse(V, proplists:get_value(K, Old))} || {K, V} <- New],
+    Combined = [case K of
+                    version ->
+                        {K, proplists:get_value(version, Old)};
+                    _ ->
+                        {K, cse(V, proplists:get_value(K, Old))}
+                end
+                || {K, V} <- New],
     Lens = [{P, proplists:get_value(len, New)} |
             proplists:get_value(priority_lengths, Old)],
     [{priority_lengths, Lens} | Combined].
@@ -687,8 +691,31 @@ find_head_message_timestamp(BQ, [{_, BQSN} | Rest], Timestamp) ->
 find_head_message_timestamp(_, [], Timestamp) ->
     Timestamp.
 
+find_oldest_message_received_timestamp(BQ, BQs) ->
+    %% Oldest message timestamp among all priority queues
+    Timestamps =
+        lists:foldl(
+          fun({_, BQSN}, Acc) ->
+                  case oldest_message_received_timestamp(BQ, BQSN) of
+                      '' -> Acc;
+                      Ts -> [Ts | Acc]
+                  end
+          end, [], BQs),
+    case Timestamps of
+        [] -> '';
+        _ -> lists:min(Timestamps)
+    end.
+
+oldest_message_received_timestamp(BQ, BQSN) ->
+    MsgCount = BQ:len(BQSN) + BQ:info(messages_unacknowledged_ram, BQSN),
+    if
+        MsgCount =/= 0 -> BQ:info(oldest_message_received_timestamp, BQSN);
+        true -> ''
+    end.
+
 zip_msgs_and_acks(Pubs, AckTags) ->
     lists:zipwith(
-      fun ({#basic_message{ id = Id }, _Props}, AckTag) ->
-                  {Id, AckTag}
+      fun ({Msg, _Props}, AckTag) ->
+              Id = mc:get_annotation(id, Msg),
+              {Id, AckTag}
       end, Pubs, AckTags).
