@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_queue_index).
@@ -428,8 +428,9 @@ maybe_needs_confirming(MsgProps, MsgOrId,
         State = #qistate{unconfirmed     = UC,
                          unconfirmed_msg = UCM}) ->
     MsgId = case MsgOrId of
-                #basic_message{id = Id} -> Id;
-                Id when is_binary(Id)   -> Id
+                Id when is_binary(Id)   -> Id;
+                Msg ->
+                    mc:get_annotation(id, Msg)
             end,
     ?MSG_ID_BYTES = size(MsgId),
     case {MsgProps#message_properties.needs_confirming, MsgOrId} of
@@ -874,7 +875,8 @@ create_pub_record_body(MsgOrId, #message_properties { expiry = Expiry,
     case MsgOrId of
         MsgId when is_binary(MsgId) ->
             {<<MsgId/binary, ExpiryBin/binary, Size:?SIZE_BITS>>, <<>>};
-        #basic_message{id = MsgId} ->
+        Msg ->
+            MsgId = mc:get_annotation(id, Msg),
             MsgBin = term_to_binary(MsgOrId),
             {<<MsgId/binary, ExpiryBin/binary, Size:?SIZE_BITS>>, MsgBin}
     end.
@@ -894,8 +896,11 @@ parse_pub_record_body(<<MsgIdNum:?MSG_ID_BITS, Expiry:?EXPIRY_BITS,
                                 size   = Size},
     case MsgBin of
         <<>> -> {MsgId, Props};
-        _    -> Msg = #basic_message{id = MsgId} = binary_to_term(MsgBin),
-                {Msg, Props}
+        _  ->
+            Msg = binary_to_term(MsgBin),
+            %% assertion
+            MsgId = mc:get_annotation(id, Msg),
+            {Msg, Props}
     end.
 
 %%----------------------------------------------------------------------------
