@@ -2,12 +2,15 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_queue_type).
+
+-behaviour(rabbit_registry_class).
+
 -include("amqqueue.hrl").
--include_lib("rabbit_common/include/resource.hrl").
+-include_lib("rabbit_common/include/rabbit.hrl").
 
 -export([
          init/0,
@@ -24,6 +27,7 @@
          purge/1,
          policy_changed/1,
          stat/1,
+         format/2,
          remove/2,
          info/2,
          state_info/1,
@@ -37,7 +41,7 @@
          handle_down/4,
          handle_event/3,
          module/2,
-         deliver/3,
+         deliver/4,
          settle/5,
          credit/5,
          dequeue/5,
@@ -46,9 +50,17 @@
          is_server_named_allowed/1,
          arguments/1,
          arguments/2,
-         notify_decorators/1
+         notify_decorators/1,
+         publish_at_most_once/2
          ]).
 
+-export([
+         added_to_rabbit_registry/2,
+         removed_from_rabbit_registry/1,
+         known_queue_type_names/0,
+         known_queue_type_modules/0
+        ]).
+
 -type queue_name() :: rabbit_amqqueue:name().
 -type queue_state() :: term().
 -type msg_tag() :: term().
@@ -64,7 +76,8 @@
 -define(DOWN_KEYS, [name, durable, auto_delete, arguments, pid, recoverable_slaves, type, state]).
 
 %% TODO resolve all registered queue types from registry
--define(QUEUE_TYPES, [rabbit_classic_queue, rabbit_quorum_queue, rabbit_stream_queue]).
+-define(QUEUE_MODULES, [rabbit_classic_queue, rabbit_quorum_queue, rabbit_stream_queue]).
+-define(KNOWN_QUEUE_TYPES, [<<"classic">>, <<"quorum">>, <<"stream">>]).
 
 %% anything that the host process needs to do on behalf of the queue type session
 -type action() ::
@@ -108,12 +121,14 @@
                           ok_msg := term(),
                           acting_user :=  rabbit_types:username()}.
 
-
+-type delivery_options() :: #{correlation => term(), %% sequence no typically
+                              atom() => term()}.
 
 -type settle_op() :: 'complete' | 'requeue' | 'discard'.
 
 -export_type([state/0,
               consume_spec/0,
+              delivery_options/0,
               action/0,
               actions/0,
               settle_op/0]).
@@ -183,7 +198,8 @@
     {protocol_error, Type :: atom(), Reason :: string(), Args :: term()}.
 
 -callback deliver([{amqqueue:amqqueue(), queue_state()}],
-                  Delivery :: term()) ->
+                  Message :: mc:state(),
+                  Options :: delivery_options()) ->
     {[{amqqueue:amqqueue(), queue_state()}], actions()}.
 
 -callback settle(queue_name(), settle_op(), rabbit_types:ctag(),
@@ -213,6 +229,9 @@
 -callback stat(amqqueue:amqqueue()) ->
     {'ok', non_neg_integer(), non_neg_integer()}.
 
+-callback format(amqqueue:amqqueue(), Context :: map()) ->
+    [{atom(), term()}].
+
 -callback capabilities() ->
     #{atom() := term()}.
 
@@ -225,7 +244,11 @@ discover(<<"quorum">>) ->
 discover(<<"classic">>) ->
     rabbit_classic_queue;
 discover(<<"stream">>) ->
-    rabbit_stream_queue.
+    rabbit_stream_queue;
+discover(Other) when is_binary(Other) ->
+    T = rabbit_registry:binary_to_type(Other),
+    {ok, Mod} = rabbit_registry:lookup_module(queue, T),
+    Mod.
 
 feature_flag_name(<<"quorum">>) ->
     quorum_queue;
@@ -285,6 +308,12 @@ stat(Q) ->
     Mod = amqqueue:get_type(Q),
     Mod:stat(Q).
 
+-spec format(amqqueue:amqqueue(), map()) ->
+    [{atom(), term()}].
+format(Q, Context) ->
+    Mod = amqqueue:get_type(Q),
+    Mod:format(Q, Context).
+
 -spec remove(queue_name(), state()) -> state().
 remove(QRef, #?STATE{ctxs = Ctxs0} = State) ->
     case maps:take(QRef, Ctxs0) of
@@ -354,7 +383,7 @@ is_server_named_allowed(Type) ->
 arguments(ArgumentType) ->
     Args0 = lists:map(fun(T) ->
                               maps:get(ArgumentType, T:capabilities(), [])
-                      end, ?QUEUE_TYPES),
+                      end, known_queue_type_modules()),
     Args = lists:flatten(Args0),
     lists:usort(Args).
 
@@ -423,7 +452,7 @@ is_recoverable(Q) ->
     {Recovered :: [amqqueue:amqqueue()],
      Failed :: [amqqueue:amqqueue()]}.
 recover(VHost, Qs) ->
-    ByType0 = maps:from_keys(?QUEUE_TYPES, []),
+    ByType0 = maps:from_keys(known_queue_type_modules(), []),
     ByType = lists:foldl(
                fun (Q, Acc) ->
                        T = amqqueue:get_type(Q),
@@ -482,52 +511,81 @@ module(QRef, State) ->
             {error, not_found}
     end.
 
--spec deliver([amqqueue:amqqueue()], Delivery :: term(),
+%% convenience function for throwaway publishes
+-spec publish_at_most_once(rabbit_types:exchange() |
+                           rabbit_exchange:name(),
+                           mc:state()) ->
+    ok | {error, not_found}.
+publish_at_most_once(#resource{} = XName, Msg) ->
+    case rabbit_exchange:lookup(XName) of
+        {ok, X} ->
+            publish_at_most_once(X, Msg);
+        Err ->
+            Err
+    end;
+publish_at_most_once(X, Msg)
+    when element(1, X) == exchange -> % hacky but good enough
+    QNames = rabbit_exchange:route(X, Msg, #{return_binding_keys => true}),
+    Qs = rabbit_amqqueue:lookup_many(QNames),
+    _ = deliver(Qs, Msg, #{}, stateless),
+    ok.
+
+-spec deliver([amqqueue:amqqueue() |
+               {amqqueue:amqqueue(), rabbit_exchange:route_infos()}],
+              Message :: mc:state(),
+              delivery_options(),
               stateless | state()) ->
     {ok, state(), actions()} | {error, Reason :: term()}.
-deliver(Qs, Delivery, State) ->
+deliver(Qs, Message, Options, State) ->
     try
-        deliver0(Qs, Delivery, State)
+        deliver0(Qs, Message, Options, State)
     catch
         exit:Reason ->
             {error, Reason}
     end.
 
-deliver0(Qs, Delivery, stateless) ->
-    ByType = lists:foldl(fun(Q, Acc) ->
-                                 Mod = amqqueue:get_type(Q),
-                                 maps:update_with(
-                                   Mod, fun(A) ->
+deliver0(Qs, Message0, Options, stateless) ->
+    ByTypeAndBindingKeys =
+    lists:foldl(fun(Elem, Acc) ->
+                        {Q, BKeys} = queue_binding_keys(Elem),
+                        Mod = amqqueue:get_type(Q),
+                        maps:update_with(
+                          {Mod, BKeys}, fun(A) ->
                                                 [{Q, stateless} | A]
                                         end, [{Q, stateless}], Acc)
-                         end, #{}, Qs),
-    maps:foreach(fun(Mod, QSs) ->
-                         _ = Mod:deliver(QSs, Delivery)
-                 end, ByType),
+                end, #{}, Qs),
+    maps:foreach(fun({Mod, BKeys}, QSs) ->
+                         Message = add_binding_keys(Message0, BKeys),
+                         _ = Mod:deliver(QSs, Message, Options)
+                 end, ByTypeAndBindingKeys),
     {ok, stateless, []};
-deliver0(Qs, Delivery, #?STATE{} = State0) ->
+deliver0(Qs, Message0, Options, #?STATE{} = State0) ->
     %% TODO: optimise single queue case?
     %% sort by queue type - then dispatch each group
-    ByType = lists:foldl(
-               fun (Q, Acc) ->
-                       Mod = amqqueue:get_type(Q),
-                       QState = case Mod:is_stateful() of
-                                    true ->
-                                        #ctx{state = S} = get_ctx(Q, State0),
-                                        S;
-                                    false ->
-                                        stateless
-                                end,
-                       maps:update_with(
-                         Mod, fun (A) ->
+    ByTypeAndBindingKeys =
+    lists:foldl(
+      fun (Elem, Acc) ->
+              {Q, BKeys} = queue_binding_keys(Elem),
+              Mod = amqqueue:get_type(Q),
+              QState = case Mod:is_stateful() of
+                           true ->
+                               #ctx{state = S} = get_ctx(Q, State0),
+                               S;
+                           false ->
+                               stateless
+                       end,
+              maps:update_with(
+                {Mod, BKeys}, fun (A) ->
                                       [{Q, QState} | A]
                               end, [{Q, QState}], Acc)
-               end, #{}, Qs),
+      end, #{}, Qs),
     %%% dispatch each group to queue type interface?
-    {Xs, Actions} = maps:fold(fun(Mod, QSs, {X0, A0}) ->
-                                      {X, A} = Mod:deliver(QSs, Delivery),
-                                      {X0 ++ X, A0 ++ A}
-                              end, {[], []}, ByType),
+    {Xs, Actions} = maps:fold(
+                      fun({Mod, BKeys}, QSs, {X0, A0}) ->
+                              Message = add_binding_keys(Message0, BKeys),
+                              {X, A} = Mod:deliver(QSs, Message, Options),
+                              {X0 ++ X, A0 ++ A}
+                      end, {[], []}, ByTypeAndBindingKeys),
     State = lists:foldl(
               fun({Q, S}, Acc) ->
                       Ctx = get_ctx_with(Q, Acc, S),
@@ -535,6 +593,22 @@ deliver0(Qs, Delivery, #?STATE{} = State0) ->
               end, State0, Xs),
     {ok, State, Actions}.
 
+queue_binding_keys(Q)
+  when ?is_amqqueue(Q) ->
+    {Q, #{}};
+queue_binding_keys({Q, #{binding_keys := BindingKeys}})
+  when ?is_amqqueue(Q) andalso is_map(BindingKeys) ->
+    {Q, BindingKeys};
+queue_binding_keys({Q, _RouteInfos})
+  when ?is_amqqueue(Q) ->
+    {Q, #{}}.
+
+add_binding_keys(Message, BindingKeys)
+  when map_size(BindingKeys) =:= 0 ->
+    Message;
+add_binding_keys(Message, BindingKeys) ->
+    mc:set_annotation(binding_keys, maps:keys(BindingKeys), Message).
+
 -spec settle(queue_name(), settle_op(), rabbit_types:ctag(),
              [non_neg_integer()], state()) ->
           {ok, state(), actions()} |
@@ -587,6 +661,13 @@ dequeue(Q, NoAck, LimiterPid, CTag, Ctxs) ->
             Err
     end.
 
+
+-spec added_to_rabbit_registry(atom(), atom()) -> ok.
+added_to_rabbit_registry(_Type, _ModuleName) -> ok.
+
+-spec removed_from_rabbit_registry(atom()) -> ok.
+removed_from_rabbit_registry(_Type) -> ok.
+
 get_ctx(QOrQref, State) ->
     get_ctx_with(QOrQref, State, undefined).
 
@@ -634,3 +715,16 @@ qref(#resource{kind = queue} = QName) ->
     QName;
 qref(Q) when ?is_amqqueue(Q) ->
     amqqueue:get_name(Q).
+
+-spec known_queue_type_modules() -> [module()].
+known_queue_type_modules() ->
+    Registered = rabbit_registry:lookup_all(queue),
+    {_, Modules} = lists:unzip(Registered),
+    ?QUEUE_MODULES ++ Modules.
+
+-spec known_queue_type_names() -> [binary()].
+known_queue_type_names() ->
+    Registered = rabbit_registry:lookup_all(queue),
+    {QueueTypes, _} = lists:unzip(Registered),
+    QTypeBins = lists:map(fun(X) -> atom_to_binary(X) end, QueueTypes),
+    ?KNOWN_QUEUE_TYPES ++ QTypeBins.
