@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_pbe).
@@ -10,8 +10,11 @@
 -export([supported_ciphers/0, supported_hashes/0, default_cipher/0, default_hash/0, default_iterations/0]).
 -export([encrypt_term/5, decrypt_term/5]).
 -export([encrypt/5, decrypt/5]).
+-export([encrypt/2, decrypt/2]).
 
--export_type([encryption_result/0]).
+-type encryptable_input() :: iodata() | '$pending-secret'.
+
+-export_type([encryptable_input/0, encryption_result/0]).
 
 supported_ciphers() ->
     credentials_obfuscation_pbe:supported_ciphers().
@@ -46,7 +49,7 @@ decrypt_term(Cipher, Hash, Iterations, PassPhrase, {encrypted, _Base64Binary}=En
 -type crypto_hash_algorithm() :: atom().
 
 -spec encrypt(crypto_cipher(), crypto_hash_algorithm(),
-    pos_integer(), iodata() | '$pending-secret', binary()) -> encryption_result().
+    pos_integer(), encryptable_input(), binary()) -> encryption_result().
 encrypt(Cipher, Hash, Iterations, PassPhrase, ClearText) ->
     credentials_obfuscation_pbe:encrypt(Cipher, Hash, Iterations, PassPhrase, ClearText).
 
@@ -57,3 +60,15 @@ decrypt(_Cipher, _Hash, _Iterations, _PassPhrase, {plaintext, Term}) ->
     Term;
 decrypt(Cipher, Hash, Iterations, PassPhrase, {encrypted, _Base64Binary}=Encrypted) ->
     credentials_obfuscation_pbe:decrypt(Cipher, Hash, Iterations, PassPhrase, Encrypted).
+
+
+-spec encrypt(encryptable_input(), binary()) -> encryption_result().
+encrypt(PassPhrase, ClearText) ->
+    credentials_obfuscation_pbe:encrypt(default_cipher(), default_hash(), default_iterations(), PassPhrase, ClearText).
+
+
+-spec decrypt(iodata(), encryption_result()) -> any().
+decrypt(_PassPhrase, {plaintext, Term}) ->
+    Term;
+decrypt(PassPhrase, {encrypted, _Base64Binary}=Encrypted) ->
+    credentials_obfuscation_pbe:decrypt(default_cipher(), default_hash(), default_iterations(), PassPhrase, Encrypted).
