@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_registry).
@@ -46,9 +46,11 @@ unregister(Class, TypeName) ->
 %% can throw a badarg, indicating that the type cannot have been
 %% registered.
 binary_to_type(TypeBin) when is_binary(TypeBin) ->
-    case catch list_to_existing_atom(binary_to_list(TypeBin)) of
-        {'EXIT', {badarg, _}} -> {error, not_found};
-        TypeAtom              -> TypeAtom
+    case catch binary_to_existing_atom(TypeBin) of
+        {'EXIT', {badarg, _}} ->
+            {error, not_found};
+        TypeAtom ->
+            TypeAtom
     end.
 
 lookup_module(Class, T) when is_atom(T) ->
@@ -65,7 +67,7 @@ lookup_all(Class) ->
 %%---------------------------------------------------------------------------
 
 internal_binary_to_type(TypeBin) when is_binary(TypeBin) ->
-    list_to_atom(binary_to_list(TypeBin)).
+    binary_to_atom(TypeBin).
 
 internal_register(Class, TypeName, ModuleName)
   when is_atom(Class), is_binary(TypeName), is_atom(ModuleName) ->
@@ -124,6 +126,7 @@ sanity_check_module(ClassModule, Module) ->
 % rabbit_registry_class behaviour itself: export added_to_rabbit_registry/2
 % and removed_from_rabbit_registry/1 functions.
 class_module(exchange)            -> rabbit_exchange_type;
+class_module(queue)               -> rabbit_queue_type;
 class_module(auth_mechanism)      -> rabbit_auth_mechanism;
 class_module(runtime_parameter)   -> rabbit_runtime_parameter;
 class_module(exchange_decorator)  -> rabbit_exchange_decorator;
