@@ -2,14 +2,18 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_types).
 
 -include("rabbit.hrl").
 
--export_type([maybe/1, info/0, infos/0, info_key/0, info_keys/0,
+-export_type([
+              %% deprecated
+              maybe/1,
+              option/1,
+              info/0, infos/0, info_key/0, info_keys/0,
               message/0, msg_id/0, basic_message/0,
               delivery/0, content/0, decoded_content/0, undecoded_content/0,
               unencoded_content/0, encoded_content/0, message_properties/0,
@@ -26,8 +30,11 @@
               node_type/0, topic_access_context/0,
               authz_data/0, authz_context/0,
               permission_atom/0, rabbit_amqqueue_name/0, binding_key/0, channel_number/0,
-              exchange_name/0, exchange_type/0, guid/0, routing_key/0]).
+              exchange_name/0, exchange_type/0, guid/0, routing_key/0,
+              sup_ref/0, child/0, child_id/0]).
 
+-type(option(T) :: T | 'none' | 'undefined').
+%% Deprecated, 'maybe' is a keyword in modern Erlang
 -type(maybe(T) :: T | 'none').
 -type(timestamp() :: {non_neg_integer(), non_neg_integer(), non_neg_integer()}).
 
@@ -49,11 +56,11 @@
 -type(decoded_content() ::
         #content{class_id              :: rabbit_framing:amqp_class_id(),
                  properties            :: rabbit_framing:amqp_property_record(),
-                 properties_bin        :: maybe(binary()),
+                 properties_bin        :: option(binary()),
                  payload_fragments_rev :: [binary()]}).
 -type(encoded_content() ::
         #content{class_id       :: rabbit_framing:amqp_class_id(),
-                 properties     :: maybe(rabbit_framing:amqp_property_record()),
+                 properties     :: option(rabbit_framing:amqp_property_record()),
                  properties_bin        :: binary(),
                  payload_fragments_rev :: [binary()]}).
 -type(content() :: undecoded_content() | decoded_content()).
@@ -210,3 +217,13 @@
 -type(authz_context() :: map()).
 
 -type(permission_atom() :: 'configure' | 'write' | 'read').
+
+%% not exported by OTP supervisor
+-type sup_ref()   :: (Name :: atom())
+                   | {Name :: atom(), Node :: node()}
+                   | {'global', Name :: term()}
+                   | {'via', Module :: module(), Name :: any()}
+                   | pid().
+
+-type child() :: 'undefined' | pid().
+-type child_id() :: term().
