@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_auth_backend_http).
@@ -15,7 +15,7 @@
 -export([description/0, p/1, q/1, join_tags/1]).
 -export([user_login_authentication/2, user_login_authorization/2,
          check_vhost_access/3, check_resource_access/4, check_topic_access/4,
-         state_can_expire/0]).
+         expiry_timestamp/1]).
 
 %% If keepalive connection is closed, retry N times before failing.
 -define(RETRY_ON_KEEPALIVE_CLOSED, 3).
@@ -33,17 +33,57 @@ description() ->
 %%--------------------------------------------------------------------
 
 user_login_authentication(Username, AuthProps) ->
-    case http_req(p(user_path), q([{username, Username}|AuthProps])) of
+    case http_req(p(user_path), q([{username, Username}] ++ extract_other_credentials(AuthProps))) of
         {error, _} = E  -> E;
         "deny"          -> {refused, "Denied by the backing HTTP service", []};
         "allow" ++ Rest -> Tags = [rabbit_data_coercion:to_atom(T) ||
-                                      T <- string:tokens(Rest, " ")],
+                                   T <- string:tokens(Rest, " ")],
+
                            {ok, #auth_user{username = Username,
                                            tags     = Tags,
-                                           impl     = fun() -> none end}};
+                                           impl     = fun() -> proplists:delete(username, AuthProps) end}};
         Other           -> {error, {bad_response, Other}}
     end.
 
+%% When a protocol plugin uses an internal AMQP 0-9-1 client to interact with RabbitMQ core,
+%% what happens that the plugin authenticates the entire authentication context (e.g. all of: password, client_id, vhost, etc)
+%% and the internal AMQP 0-9-1 client also performs further authentication.
+%%
+%% In the latter case, the complete set of credentials are persisted behind a function call
+%% that returns an AuthProps.
+%% If the user was first authenticated by rabbit_auth_backend_http, there will be one property called
+%% `rabbit_auth_backend_http` whose value is a function that returns a proplist with all the credentials used
+%% on the first successful login.
+%%
+%% When rabbit_auth_backend_cache is involved,
+%% the property `rabbit_auth_backend_cache` is a function which returns a proplist with all the credentials used
+%% on the first successful login.
+resolve_using_persisted_credentials(AuthProps) ->
+  case proplists:get_value(rabbit_auth_backend_http, AuthProps, undefined) of
+    undefined ->
+      case proplists:get_value(rabbit_auth_backend_cache, AuthProps, undefined) of
+          undefined -> AuthProps;
+          CacheAuthPropsFun -> AuthProps ++ CacheAuthPropsFun()
+      end;
+    HttpAuthPropsFun -> AuthProps ++ HttpAuthPropsFun()
+  end.
+
+
+%% Some protocols may add additional credentials into the AuthProps that should be propagated to
+%% the external authentication backends
+%% This function excludes any attribute that starts with rabbit_auth_backend_
+is_internal_property(rabbit_auth_backend_http) -> true;
+is_internal_property(rabbit_auth_backend_cache) -> true;
+is_internal_property(_Other) -> false.
+
+extract_other_credentials(AuthProps) ->
+  PublicAuthProps = [{K,V} || {K,V} <-AuthProps, not is_internal_property(K)],
+  case PublicAuthProps of
+    [] -> resolve_using_persisted_credentials(AuthProps);
+    _ -> PublicAuthProps
+  end.
+
+
 user_login_authorization(Username, AuthProps) ->
     case user_login_authentication(Username, AuthProps) of
         {ok, #auth_user{impl = Impl}} -> {ok, Impl};
@@ -89,7 +129,7 @@ check_topic_access(#auth_user{username = Username, tags = Tags},
         {permission, Permission},
         {tags, join_tags(Tags)}] ++ OptionsParameters).
 
-state_can_expire() -> false.
+expiry_timestamp(_) -> never.
 
 %%--------------------------------------------------------------------
 
@@ -179,11 +219,11 @@ p(PathName) ->
     Path.
 
 q(Args) ->
-    string:join([escape(K, V) || {K, V} <- Args], "&").
+    string:join([escape(K, V) || {K, V} <- Args, not is_function(V)], "&").
 
 escape(K, Map) when is_map(Map) ->
     string:join([escape(rabbit_data_coercion:to_list(K) ++ "." ++ rabbit_data_coercion:to_list(Key), Value)
-        || {Key, Value} <- maps:to_list(Map)], "&");
+        || {Key, Value} <- maps:to_list(Map), not is_function(Value)], "&");
 escape(K, V) ->
     rabbit_data_coercion:to_list(K) ++ "=" ++ rabbit_http_util:quote_plus(V).
 
