@@ -2,109 +2,103 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 -module(uaa_jwt).
 
 -export([add_signing_key/3,
-         remove_signing_key/1,
          decode_and_verify/1,
-         get_jwk/1,
-         verify_signing_key/2,
-         signing_keys/0]).
+         get_jwk/2,
+         verify_signing_key/2]).
 
 -export([client_id/1, sub/1, client_id/2, sub/2]).
 
 -include_lib("jose/include/jose_jwk.hrl").
+-include_lib("oauth2_client/include/oauth2_client.hrl").
 
 -define(APP, rabbitmq_auth_backend_oauth2).
 
 -type key_type() :: json | pem | map.
 
 -spec add_signing_key(binary(), key_type(), binary() | map()) -> {ok, map()} | {error, term()}.
-
 add_signing_key(KeyId, Type, Value) ->
     case verify_signing_key(Type, Value) of
         ok ->
-            SigningKeys0 = signing_keys(),
-            SigningKeys1 = maps:put(KeyId, {Type, Value}, SigningKeys0),
-            ok = update_uaa_jwt_signing_keys(SigningKeys1),
-            {ok, SigningKeys1};
+            {ok, rabbit_oauth2_config:add_signing_key(KeyId, {Type, Value})};
         {error, _} = Err ->
             Err
     end.
 
-remove_signing_key(KeyId) ->
-    UaaEnv = application:get_env(?APP, key_config, []),
-    Keys0 = proplists:get_value(signing_keys, UaaEnv),
-    Keys1 = maps:remove(KeyId, Keys0),
-    update_uaa_jwt_signing_keys(UaaEnv, Keys1).
-
--spec update_uaa_jwt_signing_keys(map()) -> ok.
-update_uaa_jwt_signing_keys(SigningKeys) ->
-    UaaEnv0 = application:get_env(?APP, key_config, []),
-    update_uaa_jwt_signing_keys(UaaEnv0, SigningKeys).
-
--spec update_uaa_jwt_signing_keys([term()], map()) -> ok.
-update_uaa_jwt_signing_keys(UaaEnv0, SigningKeys) ->
-    UaaEnv1 = proplists:delete(signing_keys, UaaEnv0),
-    UaaEnv2 = [{signing_keys, SigningKeys} | UaaEnv1],
-    application:set_env(?APP, key_config, UaaEnv2).
-
--spec update_jwks_signing_keys() -> ok | {error, term()}.
-update_jwks_signing_keys() ->
-    UaaEnv = application:get_env(?APP, key_config, []),
-    case proplists:get_value(jwks_url, UaaEnv) of
-        undefined ->
-            {error, no_jwks_url};
-        JwksUrl ->
-            rabbit_log:debug("Retrieving signing keys from ~ts", [JwksUrl]),
-            case uaa_jwks:get(JwksUrl) of
+-spec update_jwks_signing_keys(term()) -> ok | {error, term()}.
+update_jwks_signing_keys(ResourceServerId) ->
+    case rabbit_oauth2_config:get_oauth_provider_for_resource_server_id(ResourceServerId, [jwks_uri]) of
+        {error, _} = Error ->
+            rabbit_log:error("Failed to obtain a JWKS URL for resource_server_id '~tp'", [ResourceServerId]),
+            Error;
+        {ok, #oauth_provider{jwks_uri = JwksUrl, ssl_options = SslOptions}} ->
+            rabbit_log:debug("OAuth 2 JWT: downloading keys from ~tp (TLS options: ~p)", [JwksUrl, SslOptions]),
+            case uaa_jwks:get(JwksUrl, SslOptions) of
                 {ok, {_, _, JwksBody}} ->
                     KeyList = maps:get(<<"keys">>, jose:decode(erlang:iolist_to_binary(JwksBody)), []),
                     Keys = maps:from_list(lists:map(fun(Key) -> {maps:get(<<"kid">>, Key, undefined), {json, Key}} end, KeyList)),
-                    update_uaa_jwt_signing_keys(UaaEnv, Keys);
+                    rabbit_log:debug("OAuth 2 JWT: downloaded keys ~tp", [Keys]),
+                    case rabbit_oauth2_config:replace_signing_keys(ResourceServerId, Keys) of
+                      {error, _} = Err -> Err;
+                      _ -> ok
+                    end;
                 {error, _} = Err ->
+                    rabbit_log:error("OAuth 2 JWT: failed to download keys: ~tp", [Err]),
                     Err
             end
     end.
 
--spec decode_and_verify(binary()) -> {boolean(), map()} | {error, term()}.
+-spec decode_and_verify(binary()) -> {boolean(), binary(), map()} | {error, term()}.
 decode_and_verify(Token) ->
-    case uaa_jwt_jwt:get_key_id(Token) of
-        {ok, KeyId} ->
-            case get_jwk(KeyId) of
-                {ok, JWK} ->
-                    uaa_jwt_jwt:decode_and_verify(JWK, Token);
-                {error, _} = Err ->
-                    Err
-            end;
+    case uaa_jwt_jwt:resolve_resource_server_id(Token) of
         {error, _} = Err ->
-            Err
+            Err;
+        ResourceServerId ->
+            rabbit_log:debug("OAuth 2 JWT: resolved resource_server_id: '~tp'", [ResourceServerId]),
+            case uaa_jwt_jwt:get_key_id(ResourceServerId, Token) of
+                {ok, KeyId} ->
+                    rabbit_log:debug("OAuth 2 JWT: signing_key_id : '~tp'", [KeyId]),
+                    case get_jwk(KeyId, ResourceServerId) of
+                        {ok, JWK} ->
+                            case uaa_jwt_jwt:decode_and_verify(ResourceServerId, JWK, Token) of
+                                {true, Payload} -> {true, ResourceServerId, Payload};
+                                {false, Payload} -> {false, ResourceServerId, Payload}
+                            end;
+                        {error, _} = Err ->
+                            Err
+                    end;
+                {error, _} = Err -> Err
+          end
     end.
 
--spec get_jwk(binary()) -> {ok, map()} | {error, term()}.
-get_jwk(KeyId) ->
-    get_jwk(KeyId, true).
+-spec get_jwk(binary(), binary()) -> {ok, map()} | {error, term()}.
+get_jwk(KeyId, ResourceServerId) ->
+    get_jwk(KeyId, ResourceServerId, true).
 
-get_jwk(KeyId, AllowUpdateJwks) ->
-    Keys = signing_keys(),
-    case maps:get(KeyId, Keys, undefined) of
+get_jwk(KeyId, ResourceServerId, AllowUpdateJwks) ->
+    case rabbit_oauth2_config:get_signing_key(KeyId, ResourceServerId) of
         undefined ->
             if
                 AllowUpdateJwks ->
-                    case update_jwks_signing_keys() of
+                    rabbit_log:debug("OAuth 2 JWT: signing key '~tp' not found. Downloading it... ", [KeyId]),
+                    case update_jwks_signing_keys(ResourceServerId) of
                         ok ->
-                            get_jwk(KeyId, false);
+                            get_jwk(KeyId, ResourceServerId, false);
                         {error, no_jwks_url} ->
                             {error, key_not_found};
                         {error, _} = Err ->
                             Err
                     end;
                 true            ->
+                    rabbit_log:debug("OAuth 2 JWT: signing key '~tp' not found. Downloading is not allowed", [KeyId]),
                     {error, key_not_found}
             end;
         {Type, Value} ->
+            rabbit_log:debug("OAuth 2 JWT: signing key found: '~tp', '~tp'", [Type, Value]),
             case Type of
                 json     -> uaa_jwt_jwk:make_jwk(Value);
                 pem      -> uaa_jwt_jwk:from_pem(Value);
@@ -131,9 +125,6 @@ verify_signing_key(Type, Value) ->
         Err -> Err
     end.
 
-signing_keys() ->
-    UaaEnv = application:get_env(?APP, key_config, []),
-    proplists:get_value(signing_keys, UaaEnv, #{}).
 
 -spec client_id(map()) -> binary() | undefined.
 client_id(DecodedToken) ->
