@@ -2,7 +2,7 @@
 ## License, v. 2.0. If a copy of the MPL was not distributed with this
 ## file, You can obtain one at https://mozilla.org/MPL/2.0/.
 ##
-## Copyright (c) 2016-2023 VMware, Inc. or its affiliates.  All rights reserved.
+## Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 
 defmodule RabbitMQ.CLI.Core.DocGuide.Macros do
   @moduledoc """
@@ -18,7 +18,7 @@ defmodule RabbitMQ.CLI.Core.DocGuide.Macros do
 
     quote do
       def unquote(fn_name)() do
-        unquote("https://#{domain}/#{path_segment}.html")
+        unquote("https://#{domain}/docs/#{path_segment}/")
       end
     end
   end
@@ -54,6 +54,7 @@ defmodule RabbitMQ.CLI.Core.DocGuide do
   Macros.defguide("monitoring")
   Macros.defguide("networking")
   Macros.defguide("parameters")
+  Macros.defguide("passwords")
   Macros.defguide("plugins")
   Macros.defguide("prometheus")
   Macros.defguide("publishers")
