@@ -2,7 +2,7 @@
 ## License, v. 2.0. If a copy of the MPL was not distributed with this
 ## file, You can obtain one at https://mozilla.org/MPL/2.0/.
 ##
-## Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+## Copyright (c) 2007-2023 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries.  All rights reserved.
 
 defmodule RabbitMQ.CLI.Ctl.Commands.ClusterStatusCommand do
   alias RabbitMQ.CLI.Core.DocGuide
@@ -32,8 +32,20 @@ defmodule RabbitMQ.CLI.Ctl.Commands.ClusterStatusCommand do
   use RabbitMQ.CLI.Core.AcceptsNoPositionalArguments
   use RabbitMQ.CLI.Core.RequiresRabbitAppRunning
 
-  def run([], %{node: node_name, timeout: timeout}) do
-    case :rabbit_misc.rpc_call(node_name, :rabbit_mnesia, :status, []) do
+  def run([], %{node: node_name, timeout: timeout} = opts) do
+    status =
+      case :rabbit_misc.rpc_call(node_name, :rabbit_db_cluster, :cli_cluster_status, []) do
+        {:badrpc, {:EXIT, {:undef, _}}} ->
+          :rabbit_misc.rpc_call(node_name, :rabbit_mnesia, :status, [])
+
+        {:badrpc, _} = err ->
+          err
+
+        status ->
+          status
+      end
+
+    case status do
       {:badrpc, _} = err ->
         err
 
@@ -60,7 +72,7 @@ defmodule RabbitMQ.CLI.Ctl.Commands.ClusterStatusCommand do
             maintenance_status_by_node =
               Enum.map(
                 nodes,
-                fn n -> maintenance_status_by_node(n, per_node_timeout(timeout, count)) end
+                fn n -> maintenance_status_by_node(n, per_node_timeout(timeout, count), opts) end
               )
 
             cpu_cores_by_node =
@@ -273,23 +285,21 @@ defmodule RabbitMQ.CLI.Ctl.Commands.ClusterStatusCommand do
   end
 
   defp listeners_of(node, timeout) do
-    # This may seem inefficient since this call returns all known listeners
-    # in the cluster, so why do we run it on every node? See the badrpc clause,
-    # some nodes may be inavailable or partitioned from other nodes. This way we
-    # gather as complete a picture as possible. MK.
+    node = to_atom(node)
+
     listeners =
       case :rabbit_misc.rpc_call(
-             to_atom(node),
+             node,
              :rabbit_networking,
-             :active_listeners,
-             [],
+             :node_listeners,
+             [node],
              timeout
            ) do
         {:badrpc, _} -> []
         xs -> xs
       end
 
-    {node, listeners_on(listeners, node)}
+    {node, listeners}
   end
 
   defp versions_by_node(node, timeout) do
@@ -330,26 +340,26 @@ defmodule RabbitMQ.CLI.Ctl.Commands.ClusterStatusCommand do
      }}
   end
 
-  defp maintenance_status_by_node(node, timeout) do
+  defp maintenance_status_by_node(node, timeout, opts) do
     target = to_atom(node)
+    formatter = Map.get(opts, :formatter)
+
+    rpc_result =
+      :rabbit_misc.rpc_call(target, :rabbit_maintenance, :status_local_read, [target], timeout)
 
     result =
-      case :rabbit_misc.rpc_call(
-             target,
-             :rabbit_maintenance,
-             :status_local_read,
-             [target],
-             timeout
-           ) do
-        {:badrpc, _} -> "unknown"
-        :regular -> "not under maintenance"
-        :draining -> magenta("marked for maintenance")
+      case {rpc_result, formatter} do
+        {{:badrpc, _}, _} -> "unknown"
+        {:regular, _} -> "not under maintenance"
+        {:draining, "json"} -> "marked for maintenance"
+        {:draining, _} -> magenta("marked for maintenance")
         # forward compatibility: should we figure out a way to know when
         # draining completes (it involves inherently asynchronous cluster
         # operations such as quorum queue leader re-election), we'd introduce
         # a new state
-        :drained -> magenta("marked for maintenance")
-        value -> to_string(value)
+        {:drained, "json"} -> "marked for maintenance"
+        {:drained, _} -> magenta("marked for maintenance")
+        {value, _} -> to_string(value)
       end
 
     {node, result}
