@@ -2,7 +2,7 @@
 ## License, v. 2.0. If a copy of the MPL was not distributed with this
 ## file, You can obtain one at https://mozilla.org/MPL/2.0/.
 ##
-## Copyright (c) 2016-2023 VMware, Inc. or its affiliates.  All rights reserved.
+## Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 
 defmodule RabbitMQ.CLI.Ctl.Commands.JoinClusterCommand do
   alias RabbitMQ.CLI.Core.{Config, DocGuide, Helpers}
@@ -28,8 +28,6 @@ defmodule RabbitMQ.CLI.Ctl.Commands.JoinClusterCommand do
   def validate([_], _), do: :ok
   def validate(_, _), do: {:validation_failure, :too_many_args}
 
-  use RabbitMQ.CLI.Core.RequiresRabbitAppStopped
-
   def run([target_node], %{node: node_name, ram: ram, disc: disc} = opts) do
     node_type =
       case {ram, disc} do
@@ -42,12 +40,26 @@ defmodule RabbitMQ.CLI.Ctl.Commands.JoinClusterCommand do
     long_or_short_names = Config.get_option(:longnames, opts)
     target_node_normalised = Helpers.normalise_node(target_node, long_or_short_names)
 
-    :rabbit_misc.rpc_call(
-      node_name,
-      :rabbit_mnesia,
-      :join_cluster,
-      [target_node_normalised, node_type]
-    )
+    ret =
+      :rabbit_misc.rpc_call(
+        node_name,
+        :rabbit_db_cluster,
+        :join,
+        [target_node_normalised, node_type]
+      )
+
+    case ret do
+      {:badrpc, {:EXIT, {:undef, _}}} ->
+        :rabbit_misc.rpc_call(
+          node_name,
+          :rabbit_mnesia,
+          :join_cluster,
+          [target_node_normalised, node_type]
+        )
+
+      _ ->
+        ret
+    end
   end
 
   def output({:ok, :already_member}, _) do
@@ -64,6 +76,21 @@ defmodule RabbitMQ.CLI.Ctl.Commands.JoinClusterCommand do
      "Error: cannot cluster node with itself: #{node_name}"}
   end
 
+  def output({:error, {:node_type_unsupported, db, node_type}}, %{node: _node_name}) do
+    {:error, RabbitMQ.CLI.Core.ExitCodes.exit_software(),
+     "Error: `#{node_type}` node type is unsupported by the #{db} by database engine"}
+  end
+
+  def output(
+        {:error,
+         {:khepri_mnesia_migration_ex, :all_mnesia_nodes_must_run,
+          %{all_nodes: nodes, running_nodes: running}}},
+        _opts
+      ) do
+    {:error, RabbitMQ.CLI.Core.ExitCodes.exit_software(),
+     "Error: all mnesia nodes must run to join the cluster, mnesia nodes: #{inspect(nodes)}, running nodes: #{inspect(running)}"}
+  end
+
   use RabbitMQ.CLI.DefaultOutput
 
   def banner([target_node], %{node: node_name}) do
