@@ -2,7 +2,7 @@
 ## License, v. 2.0. If a copy of the MPL was not distributed with this
 ## file, You can obtain one at https://mozilla.org/MPL/2.0/.
 ##
-## Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+## Copyright (c) 2007-2023 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries.  All rights reserved.
 
 defmodule RabbitMQ.CLI.Ctl.Commands.ResetCommand do
   alias RabbitMQ.CLI.Core.DocGuide
@@ -14,7 +14,13 @@ defmodule RabbitMQ.CLI.Ctl.Commands.ResetCommand do
   use RabbitMQ.CLI.Core.RequiresRabbitAppStopped
 
   def run([], %{node: node_name}) do
-    :rabbit_misc.rpc_call(node_name, :rabbit_mnesia, :reset, [])
+    case :rabbit_misc.rpc_call(node_name, :rabbit_db, :reset, []) do
+      {:badrpc, {:EXIT, {:undef, _}}} ->
+        :rabbit_misc.rpc_call(node_name, :rabbit_mnesia, :reset, [])
+
+      ret ->
+        ret
+    end
   end
 
   def output({:error, :mnesia_unexpectedly_running}, %{node: node_name}) do
