@@ -2,18 +2,17 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_exchange_type_consistent_hash).
 -include_lib("rabbit_common/include/rabbit.hrl").
--include_lib("rabbit_common/include/rabbit_framing.hrl").
 
 -include("rabbitmq_consistent_hash_exchange.hrl").
 
 -behaviour(rabbit_exchange_type).
 
--export([description/0, serialise_events/0, route/2]).
+-export([description/0, serialise_events/0, route/3]).
 -export([validate/1, validate_binding/2,
          create/2, delete/2, policy_changed/2,
          add_binding/3, remove_bindings/3, assert_args_equivalence/2]).
@@ -59,9 +58,9 @@ description() ->
 
 serialise_events() -> false.
 
-route(#exchange {name      = Name,
-                 arguments = Args},
-      #delivery {message = Msg}) ->
+route(#exchange{name = Name,
+                arguments = Args},
+      Msg, _Options) ->
     case rabbit_db_ch_exchange:get(Name) of
         undefined ->
             [];
@@ -69,13 +68,14 @@ route(#exchange {name      = Name,
             case maps:size(BM) of
                 0 -> [];
                 N ->
-                    K              = value_to_hash(hash_on(Args), Msg),
+                    K = value_to_hash(hash_on(Args), Msg),
                     SelectedBucket = jump_consistent_hash(K, N),
                     case maps:get(SelectedBucket, BM, undefined) of
                         undefined ->
                             rabbit_log:warning("Bucket ~tp not found", [SelectedBucket]),
                             [];
-                        Queue     -> [Queue]
+                        Queue ->
+                            [Queue]
                     end
             end
     end.
@@ -259,26 +259,22 @@ jump_consistent_hash_value(_B0, J0, NumberOfBuckets, SeedState0) ->
     J = trunc((B + 1) / R),
     jump_consistent_hash_value(B, J, NumberOfBuckets, SeedState).
 
-value_to_hash(undefined, #basic_message { routing_keys = Routes }) ->
-    Routes;
-value_to_hash({header, Header}, #basic_message { content = Content }) ->
-    Headers = rabbit_basic:extract_headers(Content),
-    case Headers of
-        undefined -> undefined;
-        _         -> rabbit_misc:table_lookup(Headers, Header)
-    end;
-value_to_hash({property, Property}, #basic_message { content = Content }) ->
-    #content{properties = #'P_basic'{ correlation_id = CorrId,
-                                      message_id     = MsgId,
-                                      timestamp      = Timestamp }} =
-        rabbit_binary_parser:ensure_content_decoded(Content),
+value_to_hash(undefined, Msg) ->
+    mc:routing_keys(Msg);
+value_to_hash({header, Header}, Msg0) ->
+    maps:get(Header, mc:routing_headers(Msg0, [x_headers]));
+value_to_hash({property, Property}, Msg) ->
     case Property of
-        <<"correlation_id">> -> CorrId;
-        <<"message_id">>     -> MsgId;
-        <<"timestamp">>      ->
-            case Timestamp of
-                undefined -> undefined;
-                _         -> integer_to_binary(Timestamp)
+        <<"correlation_id">> ->
+            unwrap(mc:correlation_id(Msg));
+        <<"message_id">> ->
+            unwrap(mc:message_id(Msg));
+        <<"timestamp">> ->
+            case mc:timestamp(Msg) of
+                undefined ->
+                    undefined;
+                Timestamp ->
+                    integer_to_binary(Timestamp div 1000)
             end
     end.
 
@@ -298,8 +294,8 @@ hash_args(Args) ->
 hash_on(Args) ->
     case hash_args(Args) of
         {undefined, undefined} -> undefined;
-        {Header, undefined}    -> Header;
-        {undefined, Property}  -> Property
+        {Header, undefined} -> Header;
+        {undefined, Property} -> Property
     end.
 
 -spec map_has_value(#{bucket() => rabbit_types:binding_destination()},
@@ -320,3 +316,8 @@ map_has_value0({_Bucket, SameVal, _I}, SameVal) ->
     true;
 map_has_value0({_Bucket, _OtherVal, I}, Val) ->
     map_has_value0(maps:next(I), Val).
+
+unwrap(undefined) ->
+    undefined;
+unwrap({_, V}) ->
+    V.
