@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_exchange_type_event).
@@ -51,7 +51,7 @@ exchange() ->
 
 exchange(VHost) ->
     _ = ensure_vhost_exists(VHost),
-    rabbit_misc:r(VHost, exchange, ?EXCH_NAME).
+   rabbit_misc:r(VHost, exchange, ?EXCH_NAME).
 
 %%----------------------------------------------------------------------------
 
@@ -74,21 +74,22 @@ handle_event(#event{type      = Type,
                     timestamp = TS,
                     reference = none}, #state{vhost = VHost} = State) ->
     _ = case key(Type) of
-        ignore -> ok;
-        Key    ->
-                  Props2 = [{<<"timestamp_in_ms">>, TS} | Props],
-                  PBasic = #'P_basic'{delivery_mode = 2,
-                                      headers = fmt_proplist(Props2),
-                                      %% 0-9-1 says the timestamp is a
-                                      %% "64 bit POSIX
-                                      %% timestamp". That's second
-                                      %% resolution, not millisecond.
-                                      timestamp = erlang:convert_time_unit(
-                                                    TS, milli_seconds, seconds)},
-            Msg = rabbit_basic:message(exchange(VHost), Key, PBasic, <<>>),
-                  rabbit_basic:publish(
-                    rabbit_basic:delivery(false, false, Msg, undefined))
-    end,
+            ignore -> ok;
+            Key    ->
+                Props2 = [{<<"timestamp_in_ms">>, TS} | Props],
+                PBasic = #'P_basic'{delivery_mode = 2,
+                                    headers = fmt_proplist(Props2),
+                                    %% 0-9-1 says the timestamp is a
+                                    %% "64 bit POSIX
+                                    %% timestamp". That's second
+                                    %% resolution, not millisecond.
+                                    timestamp = erlang:convert_time_unit(
+                                                  TS, milli_seconds, seconds)},
+                Content = rabbit_basic:build_content(PBasic, <<>>),
+                XName = exchange(VHost),
+                {ok, Msg} = mc_amqpl:message(XName, Key, Content),
+                rabbit_queue_type:publish_at_most_once(XName, Msg)
+        end,
     {ok, State};
 handle_event(_Event, State) ->
     {ok, State}.
@@ -106,10 +107,7 @@ code_change(_OldVsn, State, _Extra) -> {ok, State}.
 %%----------------------------------------------------------------------------
 
 ensure_vhost_exists(VHost) ->
-    case rabbit_vhost:exists(VHost) of
-        false -> rabbit_vhost:add(VHost, ?INTERNAL_USER);
-        _     -> ok
-    end.
+    rabbit_vhost:add(VHost, ?INTERNAL_USER).
 
 %% pattern matching is way more efficient that the string operations,
 %% let's use all the keys we're aware of to speed up the handler.
