@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_mgmt_app).
@@ -26,11 +26,18 @@
                    [{description, "Imports definition file at management.load_definitions"},
                     {mfa,         {rabbit_mgmt_load_definitions, boot, []}}]}).
 
+-rabbit_feature_flag(
+   {detailed_queues_endpoint,
+    #{desc          => "Add a detailed queues HTTP API endpoint. Reduce number of metrics in the default endpoint.",
+      stability     => stable,
+      depends_on    => [feature_flags_v2]
+     }}).
+
 start(_Type, _StartArgs) ->
-    case application:get_env(rabbitmq_management_agent, disable_metrics_collector, false) of
-        false ->
-            start();
+    case rabbit_mgmt_agent_config:is_metrics_collector_enabled() of
         true ->
+            start();
+        false ->
             rabbit_log:warning("Metrics collection disabled in management agent, "
                                "management only interface started", []),
             start()
