@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_mgmt_dispatcher).
@@ -134,6 +134,7 @@ dispatcher() ->
      {"/exchanges/:vhost/:exchange/bindings/source",           rabbit_mgmt_wm_bindings, [exchange_source]},
      {"/exchanges/:vhost/:exchange/bindings/destination",      rabbit_mgmt_wm_bindings, [exchange_destination]},
      {"/queues",                                               rabbit_mgmt_wm_queues, []},
+     {"/queues/detailed",                                      rabbit_mgmt_wm_queues, [detailed]},
      {"/queues/:vhost",                                        rabbit_mgmt_wm_queues, []},
      {"/queues/:vhost/:queue",                                 rabbit_mgmt_wm_queue, []},
      {"/queues/:vhost/:destination/bindings",                  rabbit_mgmt_wm_bindings, [queue]},
@@ -168,6 +169,8 @@ dispatcher() ->
      {"/user-limits/:user",                                    rabbit_mgmt_wm_user_limits, []},
      {"/feature-flags",                                        rabbit_mgmt_wm_feature_flags, []},
      {"/feature-flags/:name/enable",                           rabbit_mgmt_wm_feature_flag_enable, []},
+     {"/deprecated-features",                                  rabbit_mgmt_wm_deprecated_features, [all]},
+     {"/deprecated-features/used",                             rabbit_mgmt_wm_deprecated_features, [used]},
      {"/whoami",                                               rabbit_mgmt_wm_whoami, []},
      {"/permissions",                                          rabbit_mgmt_wm_permissions, []},
      {"/permissions/:vhost/:user",                             rabbit_mgmt_wm_permission, []},
