@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_mgmt_util).
@@ -74,67 +74,70 @@
 %%--------------------------------------------------------------------
 
 is_authorized(ReqData, Context) ->
-    is_authorized(ReqData, Context, '', fun(_) -> true end).
+    rabbit_web_dispatch_access_control:is_authorized(ReqData,
+                                           Context,
+                                           auth_config()).
 
 is_authorized_admin(ReqData, Context) ->
-    is_authorized(ReqData, Context,
-                  <<"Not administrator user">>,
-                  fun(#user{tags = Tags}) -> is_admin(Tags) end).
+    rabbit_web_dispatch_access_control:is_authorized_admin(ReqData,
+                                                 Context,
+                                                 auth_config()).
 
 is_authorized_admin(ReqData, Context, Token) ->
-    is_authorized(ReqData, Context,
-                  rabbit_data_coercion:to_binary(
-                    application:get_env(rabbitmq_management, oauth_client_id, "")),
-                  Token, <<"Not administrator user">>,
-                  fun(#user{tags = Tags}) -> is_admin(Tags) end).
+    AuthConfig = auth_config(),
+    rabbit_web_dispatch_access_control:is_authorized(
+      ReqData, Context,
+      AuthConfig#auth_settings.oauth_client_id,
+      Token, <<"Not administrator user">>,
+      fun(#user{tags = Tags}) ->
+              rabbit_web_dispatch_access_control:is_admin(Tags)
+      end, AuthConfig).
 
 is_authorized_admin(ReqData, Context, Username, Password) ->
-    case is_basic_auth_disabled() of
-        true ->
-            Msg = "HTTP access denied: basic auth disabled",
-            rabbit_log:warning(Msg),
-            not_authorised(Msg, ReqData, Context);
-        false ->
-            is_authorized(ReqData, Context, Username, Password,
-                          <<"Not administrator user">>,
-                          fun(#user{tags = Tags}) -> is_admin(Tags) end)
-    end.
+    rabbit_web_dispatch_access_control:is_authorized_admin(ReqData,
+                                                 Context,
+                                                 Username,
+                                                 Password,
+                                                 auth_config()).
 
 is_authorized_monitor(ReqData, Context) ->
-    is_authorized(ReqData, Context,
-                  <<"Not monitor user">>,
-                  fun(#user{tags = Tags}) -> is_monitor(Tags) end).
+    rabbit_web_dispatch_access_control:is_authorized_monitor(ReqData,
+                                                   Context,
+                                                   auth_config()).
 
 is_authorized_vhost(ReqData, Context) ->
-    is_authorized(ReqData, Context,
-                  <<"User not authorised to access virtual host">>,
-                  fun(#user{tags = Tags} = User) ->
-                          is_admin(Tags) orelse user_matches_vhost(ReqData, User)
-                  end).
+    rabbit_web_dispatch_access_control:is_authorized_vhost(ReqData,
+                                                 Context,
+                                                 auth_config()).
 
 is_authorized_vhost_visible(ReqData, Context) ->
-    is_authorized(ReqData, Context,
-                  <<"User not authorised to access virtual host">>,
-                  fun(#user{tags = Tags} = User) ->
-                          is_admin(Tags) orelse user_matches_vhost_visible(ReqData, User)
-                  end).
+    rabbit_web_dispatch_access_control:is_authorized_vhost_visible(ReqData,
+                                                         Context,
+                                                         auth_config()).
 
 is_authorized_vhost_visible_for_monitoring(ReqData, Context) ->
-  is_authorized(ReqData, Context,
-    <<"User not authorised to access virtual host">>,
-    fun(#user{tags = Tags} = User) ->
-      is_admin(Tags)
-          orelse is_monitor(Tags)
-          orelse user_matches_vhost_visible(ReqData, User)
-    end).
+    rabbit_web_dispatch_access_control:is_authorized_vhost_visible_for_monitoring(ReqData,
+                                                                        Context,
+                                                                        auth_config()).
+
+auth_config() ->
+    BasicAuthEnabled = not get_bool_env(rabbitmq_management, disable_basic_auth, false),
+    OauthEnabled = get_bool_env(rabbitmq_management, oauth_enabled, false),
+    OauthClientId = rabbit_data_coercion:to_binary(
+                        application:get_env(rabbitmq_management, oauth_client_id, "")),
+    #auth_settings{auth_realm = ?AUTH_REALM,
+                   basic_auth_enabled = BasicAuthEnabled,
+                   oauth2_enabled = OauthEnabled,
+                   oauth_client_id = OauthClientId}.
 
 disable_stats(ReqData) ->
     MgmtOnly = case qs_val(<<"disable_stats">>, ReqData) of
                    <<"true">> -> true;
                    _ -> false
                end,
-    MgmtOnly orelse get_bool_env(rabbitmq_management, disable_management_stats, false)
-        orelse get_bool_env(rabbitmq_management_agent, disable_metrics_collector, false).
+    MgmtOnly orelse
+    not rabbit_mgmt_agent_config:is_metrics_collector_enabled() orelse
+    not rabbit_mgmt_features:are_stats_enabled().
 
 enable_queue_totals(ReqData) ->
     EnableTotals = case qs_val(<<"enable_queue_totals">>, ReqData) of
@@ -153,24 +156,6 @@ get_bool_env(Application, Par, Default) ->
             Default
     end.
 
-user_matches_vhost(ReqData, User) ->
-    case vhost(ReqData) of
-        not_found -> true;
-        none      -> true;
-        V         ->
-            AuthzData = get_authz_data(ReqData),
-            lists:member(V, list_login_vhosts_names(User, AuthzData))
-    end.
-
-user_matches_vhost_visible(ReqData, User) ->
-    case vhost(ReqData) of
-        not_found -> true;
-        none      -> true;
-        V         ->
-            AuthzData = get_authz_data(ReqData),
-            lists:member(V, list_visible_vhosts_names(User, AuthzData))
-    end.
-
 get_authz_data(ReqData) ->
     {PeerAddress, _PeerPort} = cowboy_req:peer(ReqData),
     {ip, PeerAddress}.
@@ -179,158 +164,51 @@ get_authz_data(ReqData) ->
 %% their own stuff. Monitors can see other users' and delete their
 %% own. Admins can do it all.
 is_authorized_user(ReqData, Context, Item) ->
-    is_authorized(ReqData, Context,
-                  <<"User not authorised to access object">>,
-                  fun(#user{username = Username, tags = Tags}) ->
-                          case cowboy_req:method(ReqData) of
-                              <<"DELETE">> -> is_admin(Tags);
-                              _            -> is_monitor(Tags)
-                          end orelse Username == pget(user, Item)
-                  end).
+    rabbit_web_dispatch_access_control:is_authorized_user(ReqData,
+                                                Context,
+                                                Item,
+                                                auth_config()).
 
 %% For policies / parameters. Like is_authorized_vhost but you have to
 %% be a policymaker.
 is_authorized_policies(ReqData, Context) ->
-    is_authorized(ReqData, Context,
-                  <<"User not authorised to access object">>,
-                  fun(User = #user{tags = Tags}) ->
-                          is_admin(Tags) orelse
-                                           (is_policymaker(Tags) andalso
-                                            user_matches_vhost(ReqData, User))
-                  end).
+    rabbit_web_dispatch_access_control:is_authorized_policies(ReqData,
+                                                    Context,
+                                                    auth_config()).
 
 %% For global parameters. Must be policymaker.
 is_authorized_global_parameters(ReqData, Context) ->
-    is_authorized(ReqData, Context,
-                  <<"User not authorised to access object">>,
-                  fun(#user{tags = Tags}) ->
-                           is_policymaker(Tags)
-                  end).
+    rabbit_web_dispatch_access_control:is_authorized_global_parameters(ReqData,
+                                                             Context,
+                                                             auth_config()).
 
-is_basic_auth_disabled() ->
-    get_bool_env(rabbitmq_management, disable_basic_auth, false).
+%% is_basic_auth_disabled() ->
+%%     get_bool_env(rabbitmq_management, disable_basic_auth, false).
 
-is_oauth2_enabled() ->
-    get_bool_env(rabbitmq_management, oauth_enabled, false).
-
-is_authorized(ReqData, Context, ErrorMsg, Fun) ->
-    case cowboy_req:method(ReqData) of
-        <<"OPTIONS">> -> {true, ReqData, Context};
-        _             -> is_authorized1(ReqData, Context, ErrorMsg, Fun)
-    end.
-
-is_authorized1(ReqData, Context, ErrorMsg, Fun) ->
-    case cowboy_req:parse_header(<<"authorization">>, ReqData) of
-        {basic, Username, Password} ->
-            case is_basic_auth_disabled() of
-                true ->
-                    Msg = "HTTP access denied: basic auth disabled",
-                    rabbit_log:warning(Msg),
-                    not_authorised(Msg, ReqData, Context);
-                false ->
-                    is_authorized(ReqData, Context,
-                                  Username, Password,
-                                  ErrorMsg, Fun)
-            end;
-        {bearer, Token} ->
-            % Username is only used in case is_authorized is not able to parse the token
-            % and extact the username from it
-            Username = rabbit_data_coercion:to_binary(
-                         application:get_env(rabbitmq_management, oauth_client_id, "")),
-            is_authorized(ReqData, Context, Username, Token, ErrorMsg, Fun);
-        _ ->
-            case is_basic_auth_disabled() of
-                true ->
-                    Msg = "HTTP access denied: basic auth disabled",
-                    rabbit_log:warning(Msg),
-                    not_authorised(Msg, ReqData, Context);
-                false ->
-                    {{false, ?AUTH_REALM}, ReqData, Context}
-            end
-    end.
+%% is_oauth2_enabled() ->
+%%     get_bool_env(rabbitmq_management, oauth_enabled, false).
 
 is_authorized_user(ReqData, Context, Username, Password) ->
-    Msg = <<"User not authorized">>,
-    Fun = fun(_) -> true end,
-    is_authorized(ReqData, Context, Username, Password, Msg, Fun).
+
+    rabbit_web_dispatch_access_control:is_authorized_user(ReqData,
+                                                Context,
+                                                Username,
+                                                Password,
+                                                auth_config()).
 
 is_authorized_user(ReqData, Context, Username, Password, ReplyWhenFailed) ->
-    Msg = <<"User not authorized">>,
-    Fun = fun(_) -> true end,
-    is_authorized(ReqData, Context, Username, Password, Msg, Fun, ReplyWhenFailed).
-
-is_authorized(ReqData, Context, Username, Password, ErrorMsg, Fun) ->
-    is_authorized(ReqData, Context, Username, Password, ErrorMsg, Fun, true).
-
-is_authorized(ReqData, Context, Username, Password, ErrorMsg, Fun, ReplyWhenFailed) ->
-    ErrFun = fun (ResolvedUserName, Msg) ->
-                     rabbit_log:warning("HTTP access denied: user '~ts' - ~ts",
-                                        [ResolvedUserName, Msg]),
-                     case ReplyWhenFailed of
-                       true -> not_authorised(Msg, ReqData, Context);
-                       false -> {false, ReqData, "Not_Authorized"}
-                     end
-             end,
-    AuthProps = [{password, Password}] ++ case vhost(ReqData) of
-        VHost when is_binary(VHost) -> [{vhost, VHost}];
-        _                           -> []
-    end,
-    {IP, _} = cowboy_req:peer(ReqData),
-    case rabbit_access_control:check_user_login(Username, AuthProps) of
-        {ok, User = #user{username = ResolvedUsername, tags = Tags}} ->
-            case rabbit_access_control:check_user_loopback(ResolvedUsername, IP) of
-                ok ->
-                    case is_mgmt_user(Tags) of
-                        true ->
-                            case Fun(User) of
-                                true  ->
-                                    rabbit_core_metrics:auth_attempt_succeeded(IP, ResolvedUsername, http),
-                                    {true, ReqData,
-                                     Context#context{user     = User,
-                                                     password = Password}};
-                                false ->
-                                    rabbit_core_metrics:auth_attempt_failed(IP, ResolvedUsername, http),
-                                    ErrFun(ResolvedUsername, ErrorMsg)
-                            end;
-                        false ->
-                            rabbit_core_metrics:auth_attempt_failed(IP, ResolvedUsername, http),
-                            ErrFun(ResolvedUsername, <<"Not management user">>)
-                    end;
-                not_allowed ->
-                    rabbit_core_metrics:auth_attempt_failed(IP, ResolvedUsername, http),
-                    ErrFun(ResolvedUsername, <<"User can only log in via localhost">>)
-            end;
-        {refused, _Username, Msg, Args} ->
-            rabbit_core_metrics:auth_attempt_failed(IP, Username, http),
-            rabbit_log:warning("HTTP access denied: ~ts",
-                               [rabbit_misc:format(Msg, Args)]),
-            case ReplyWhenFailed of
-              true -> not_authenticated(<<"Not_Authorized">>, ReqData, Context);
-              false -> {false, ReqData, "Not_Authorized"}
-            end
-    end.
+    rabbit_web_dispatch_access_control:is_authorized_user(ReqData,
+                                                Context,
+                                                Username,
+                                                Password,
+                                                ReplyWhenFailed,
+                                                auth_config()).
 
 vhost_from_headers(ReqData) ->
-    case cowboy_req:header(<<"x-vhost">>, ReqData) of
-        undefined -> none;
-        %% blank x-vhost means "All hosts" is selected in the UI
-        <<>>      -> none;
-        VHost     -> VHost
-    end.
+    rabbit_web_dispatch_access_control:vhost_from_headers(ReqData).
 
 vhost(ReqData) ->
-    Value = case id(vhost, ReqData) of
-      none  -> vhost_from_headers(ReqData);
-      VHost -> VHost
-    end,
-    case Value of
-      none -> none;
-      Name ->
-        case rabbit_vhost:exists(Name) of
-          true  -> Name;
-          false -> not_found
-        end
-    end.
+    rabbit_web_dispatch_access_control:vhost(ReqData).
 
 user(ReqData) ->
   case id(user, ReqData) of
@@ -394,13 +272,23 @@ get_value_param(Name, ReqData) ->
         Bin       -> binary_to_list(Bin)
     end.
 
+get_sorts_param(ReqData, Def) ->
+    case get_value_param(<<"sort">>, ReqData) of
+        undefined ->
+            Def;
+        [] ->
+            Def;
+        S ->
+            [S]
+    end.
+
 reply_list(Facts, DefaultSorts, ReqData, Context, Pagination) ->
     SortList =
-    sort_list_and_paginate(
-          extract_columns_list(Facts, ReqData),
-          DefaultSorts,
-          get_value_param(<<"sort">>, ReqData),
-          get_sort_reverse(ReqData), Pagination),
+        sort_list_and_paginate(
+              extract_columns_list(Facts, ReqData),
+              DefaultSorts,
+              get_sorts_param(ReqData, undefined),
+              get_sort_reverse(ReqData), Pagination),
 
     reply(SortList, ReqData, Context).
 
@@ -442,7 +330,9 @@ reply_list_or_paginate(Facts, ReqData, Context) ->
 merge_sorts(DefaultSorts, Extra) ->
     case Extra of
         undefined -> DefaultSorts;
-        Extra     -> [Extra | DefaultSorts]
+        Extra ->
+            %% it is possible that the extra sorts have an overlap with default
+            lists:uniq(Extra ++ DefaultSorts)
     end.
 
 %% Resource augmentation. Works out the most optimal configuration of the operations:
@@ -474,9 +364,9 @@ augment_resources0(Resources, DefaultSort, BasicColumns, Pagination, ReqData,
     SortFun = fun (AugCtx) -> sort(DefaultSort, AugCtx) end,
     AugFun = fun (AugCtx) -> augment(AugmentFun, AugCtx) end,
     PageFun = fun page/1,
-    Pagination = pagination_params(ReqData),
-    Sort = def(get_value_param(<<"sort">>, ReqData), DefaultSort),
-    Columns = def(columns(ReqData), all),
+    %% Sort needs to be a list of, erm, strings which are lists
+    Sort = get_sorts_param(ReqData, DefaultSort),
+    Columns = columns(ReqData),
     ColumnsAsStrings = columns_as_strings(Columns),
     Pipeline =
         case {Pagination =/= undefined,
@@ -490,7 +380,10 @@ augment_resources0(Resources, DefaultSort, BasicColumns, Pagination, ReqData,
             {true, basic, basic} ->
                 [SortFun, PageFun];
             {true, extended, _} ->
-                % pagination with extended sort columns - SLOW
+                Path = cowboy_req:path(ReqData),
+                rabbit_log:debug("HTTP API: ~s slow query mode requested - extended sort on ~0p",
+                                 [Path, Sort]),
+                % pagination with extended sort columns - SLOW!
                 [AugFun, SortFun, PageFun];
             {true, basic, extended} ->
                 % pagination with extended columns and sorting on basic
@@ -646,7 +539,7 @@ pagination_params(ReqData) ->
                                   [PageNum, PageSize])})
     end.
 
--spec maybe_reverse([any()], string() | true | false) -> [any()].
+-spec maybe_reverse([any()], string() | boolean()) -> [any()].
 maybe_reverse([], _) ->
     [];
 maybe_reverse(RangeList, true) when is_list(RangeList) ->
@@ -771,18 +664,9 @@ bad_request(Reason, ReqData, Context) ->
 service_unavailable(Reason, ReqData, Context) ->
     halt_response(503, service_unavailable, Reason, ReqData, Context).
 
-not_authenticated(Reason, ReqData, Context) ->
-    case is_oauth2_enabled() of
-      false ->
-          ReqData1 = cowboy_req:set_resp_header(<<"www-authenticate">>, ?AUTH_REALM, ReqData),
-          halt_response(401, not_authorized, Reason, ReqData1, Context);
-      true ->
-          halt_response(401, not_authorized, Reason, ReqData, Context)
-    end.
 
 not_authorised(Reason, ReqData, Context) ->
-    %% TODO: consider changing to 403 in 4.0
-    halt_response(401, not_authorised, Reason, ReqData, Context).
+    rabbit_web_dispatch_access_control:not_authorised(Reason, ReqData, Context).
 
 not_found(Reason, ReqData, Context) ->
     halt_response(404, not_found, Reason, ReqData, Context).
@@ -806,57 +690,38 @@ redirect_to_home(ReqData, Reason, Context) ->
     {stop, ReqData1, Context}.
 
 halt_response(Code, Type, Reason, ReqData, Context) ->
-    ReasonFormatted = format_reason(Reason),
-    Json = #{<<"error">>  => Type,
-             <<"reason">> => ReasonFormatted},
-    ReqData1 = cowboy_req:reply(Code,
-        #{<<"content-type">> => <<"application/json">>},
-        rabbit_json:encode(Json), ReqData),
-    {stop, ReqData1, Context}.
+    rabbit_web_dispatch_access_control:halt_response(Code,
+                                           Type,
+                                           Reason,
+                                           ReqData,
+                                           Context).
 
-format_reason(Tuple) when is_tuple(Tuple) ->
-    rabbit_mgmt_format:tuple(Tuple);
-format_reason(Binary) when is_binary(Binary) ->
-    Binary;
-format_reason(Other) ->
-    case is_string(Other) of
-        true ->  rabbit_mgmt_format:print("~ts", [Other]);
-        false -> rabbit_mgmt_format:print("~tp", [Other])
-    end.
-
-is_string(List) when is_list(List) ->
-    lists:all(
-        fun(El) -> is_integer(El) andalso El > 0 andalso El < 16#10ffff end,
-        List);
-is_string(_) -> false.
-
-id(Key, ReqData) when Key =:= exchange;
-                      Key =:= source;
-                      Key =:= destination ->
-    case id0(Key, ReqData) of
-        <<"amq.default">> -> <<"">>;
-        Name              -> Name
-    end;
 id(Key, ReqData) ->
-    id0(Key, ReqData).
-
-id0(Key, ReqData) ->
-    case cowboy_req:binding(Key, ReqData) of
-        undefined -> none;
-        Id        -> Id
-    end.
+    rabbit_web_dispatch_access_control:id(Key, ReqData).
 
 read_complete_body(Req) ->
     read_complete_body(Req, <<"">>).
-read_complete_body(Req0, Acc) ->
-    case cowboy_req:read_body(Req0) of
-        {ok, Data, Req}   -> {ok, <<Acc/binary, Data/binary>>, Req};
-        {more, Data, Req} -> read_complete_body(Req, <<Acc/binary, Data/binary>>)
+read_complete_body(Req, Acc) ->
+    BodySizeLimit = application:get_env(rabbitmq_management, max_http_body_size, ?MANAGEMENT_DEFAULT_HTTP_MAX_BODY_SIZE),
+    read_complete_body(Req, Acc, BodySizeLimit).
+read_complete_body(Req0, Acc, BodySizeLimit) ->
+    case bit_size(Acc) > BodySizeLimit of
+        true ->
+            {error, "Exceeded HTTP request body size limit"};
+        false ->
+            case cowboy_req:read_body(Req0) of
+                {ok, Data, Req}   -> {ok, <<Acc/binary, Data/binary>>, Req};
+                {more, Data, Req} -> read_complete_body(Req, <<Acc/binary, Data/binary>>)
+            end
     end.
 
 with_decode(Keys, ReqData, Context, Fun) ->
-    {ok, Body, ReqData1} = read_complete_body(ReqData),
-    with_decode(Keys, Body, ReqData1, Context, Fun).
+    case read_complete_body(ReqData) of
+        {error, Reason} ->
+            bad_request(Reason, ReqData, Context);
+        {ok, Body, ReqData1} ->
+            with_decode(Keys, Body, ReqData1, Context, Fun)
+    end.
 
 with_decode(Keys, Body, ReqData, Context, Fun) ->
     case decode(Keys, Body) of
@@ -946,7 +811,8 @@ direct_request(MethodName, Transformers, Extra, ErrorMsg, ReqData,
                       rabbit_log:warning(ErrorMsg, [Explanation]),
                       bad_request(list_to_binary(Explanation), ReqData1, Context);
                   {badrpc, Reason} ->
-                      rabbit_log:warning(ErrorMsg, [Reason]),
+                      Msg = io_lib:format("~tp", [Reason]),
+                      rabbit_log:warning(ErrorMsg, [Msg]),
                       bad_request(
                         list_to_binary(
                           io_lib:format("Request to node ~ts failed with ~tp",
@@ -1068,7 +934,7 @@ all_or_one_vhost(ReqData, Fun) ->
 
 filter_vhost(List, ReqData, Context) ->
     User = #user{tags = Tags} = Context#context.user,
-    Fn   = case is_admin(Tags) of
+    Fn   = case rabbit_web_dispatch_access_control:is_admin(Tags) of
                true  -> fun list_visible_vhosts_names/2;
                false -> fun list_login_vhosts_names/2
            end,
@@ -1134,13 +1000,7 @@ post_respond({stop, ReqData, Context}) ->
 post_respond({JSON, ReqData, Context}) ->
     {true, cowboy_req:set_resp_body(JSON, ReqData), Context}.
 
-is_admin(T)       -> intersects(T, [administrator]).
-is_policymaker(T) -> intersects(T, [administrator, policymaker]).
-is_monitor(T)     -> intersects(T, [administrator, monitoring]).
-is_mgmt_user(T)   -> intersects(T, [administrator, monitoring, policymaker,
-                                    management]).
-
-intersects(A, B) -> lists:any(fun(I) -> lists:member(I, B) end, A).
+is_monitor(T) -> rabbit_web_dispatch_access_control:is_monitor(T).
 
 %% The distinction between list_visible_vhosts and list_login_vhosts
 %% is there to ensure that monitors can always learn of the
@@ -1264,9 +1124,6 @@ int(Name, ReqData) ->
                      end
     end.
 
-def(undefined, Def) -> Def;
-def(V, _) -> V.
-
 -spec qs_val(binary(), cowboy_req:req()) -> any() | undefined.
 qs_val(Name, ReqData) ->
     Qs = cowboy_req:parse_qs(ReqData),
