@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 -module(mqtt_machine).
 -behaviour(ra_machine).
@@ -23,8 +23,8 @@
 
 -type reply() :: {ok, term()} | {error, term()}.
 
--type command() :: {register, client_id(), pid()} |
-                   {unregister, client_id(), pid()} |
+-type command() :: {register, client_id_ra(), pid()} |
+                   {unregister, client_id_ra(), pid()} |
                    list.
 version() -> 1.
 
