@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_mqtt).
@@ -10,6 +10,8 @@
 -behaviour(application).
 
 -include("rabbit_mqtt.hrl").
+-include("rabbit_mqtt_packet.hrl").
+-include_lib("rabbit/include/rabbit_global_counters.hrl").
 -include_lib("stdlib/include/assert.hrl").
 
 -export([start/2, stop/1]).
@@ -60,7 +62,7 @@ emit_connection_info_all(Nodes, Items, Ref, AggregatorPid) ->
 
 -spec emit_connection_info_local(rabbit_types:info_keys(), reference(), pid()) -> ok.
 emit_connection_info_local(Items, Ref, AggregatorPid) ->
-    LocalPids = local_connection_pids(),
+    LocalPids = list_local_mqtt_connections(),
     emit_connection_info(Items, Ref, AggregatorPid, LocalPids).
 
 emit_connection_info(Items, Ref, AggregatorPid, Pids) ->
@@ -70,7 +72,7 @@ emit_connection_info(Items, Ref, AggregatorPid, Pids) ->
               rabbit_mqtt_reader:info(Pid, Items)
       end, Pids).
 
--spec close_local_client_connections(string() | binary()) -> {'ok', non_neg_integer()}.
+-spec close_local_client_connections(atom()) -> {'ok', non_neg_integer()}.
 close_local_client_connections(Reason) ->
     Pids = local_connection_pids(),
     lists:foreach(fun(Pid) ->
@@ -91,20 +93,53 @@ local_connection_pids() ->
                           end, pg:which_groups(PgScope))
     end.
 
+%% This function excludes Web MQTT connections
+list_local_mqtt_connections() ->
+    PlainPids = rabbit_networking:list_local_connections_of_protocol(?MQTT_TCP_PROTOCOL),
+    TLSPids   = rabbit_networking:list_local_connections_of_protocol(?MQTT_TLS_PROTOCOL),
+    PlainPids ++ TLSPids.
+
+
 init_global_counters() ->
-    init_global_counters(?MQTT_PROTO_V3),
-    init_global_counters(?MQTT_PROTO_V4).
+    lists:foreach(fun init_global_counters/1, [?MQTT_PROTO_V3,
+                                               ?MQTT_PROTO_V4,
+                                               ?MQTT_PROTO_V5]),
+    rabbit_global_counters:init([{queue_type, ?QUEUE_TYPE_QOS_0}, {dead_letter_strategy, disabled}],
+                                [?MESSAGES_DEAD_LETTERED_MAXLEN_COUNTER]).
 
 init_global_counters(ProtoVer) ->
     Proto = {protocol, ProtoVer},
     rabbit_global_counters:init([Proto]),
-    rabbit_global_counters:init([Proto, {queue_type, ?QUEUE_TYPE_QOS_0}]),
     rabbit_global_counters:init([Proto, {queue_type, rabbit_classic_queue}]),
-    rabbit_global_counters:init([Proto, {queue_type, rabbit_quorum_queue}]).
+    rabbit_global_counters:init([Proto, {queue_type, rabbit_quorum_queue}]),
+    rabbit_global_counters:init([Proto, {queue_type, ?QUEUE_TYPE_QOS_0}]).
 
 persist_static_configuration() ->
     rabbit_mqtt_util:init_sparkplug(),
 
+    {ok, Exchange} = application:get_env(?APP_NAME, exchange),
+    ?assert(is_binary(Exchange)),
+    ok = persistent_term:put(?PERSISTENT_TERM_EXCHANGE, Exchange),
+
     {ok, MailboxSoftLimit} = application:get_env(?APP_NAME, mailbox_soft_limit),
     ?assert(is_integer(MailboxSoftLimit)),
-    ok = persistent_term:put(?PERSISTENT_TERM_MAILBOX_SOFT_LIMIT, MailboxSoftLimit).
+    ok = persistent_term:put(?PERSISTENT_TERM_MAILBOX_SOFT_LIMIT, MailboxSoftLimit),
+
+    {ok, TopicAliasMax} = application:get_env(?APP_NAME, topic_alias_maximum),
+    ?assert(is_integer(TopicAliasMax) andalso
+            TopicAliasMax >= 0 andalso
+            TopicAliasMax =< ?TWO_BYTE_INTEGER_MAX),
+    ok = persistent_term:put(?PERSISTENT_TERM_TOPIC_ALIAS_MAXIMUM, TopicAliasMax),
+
+    {ok, MaxSizeUnauth} = application:get_env(?APP_NAME, max_packet_size_unauthenticated),
+    assert_valid_max_packet_size(MaxSizeUnauth),
+    ok = persistent_term:put(?PERSISTENT_TERM_MAX_PACKET_SIZE_UNAUTHENTICATED, MaxSizeUnauth),
+
+    {ok, MaxSizeAuth} = application:get_env(?APP_NAME, max_packet_size_authenticated),
+    assert_valid_max_packet_size(MaxSizeAuth),
+    ok = persistent_term:put(?PERSISTENT_TERM_MAX_PACKET_SIZE_AUTHENTICATED, MaxSizeAuth).
+
+assert_valid_max_packet_size(Val) ->
+    ?assert(is_integer(Val) andalso
+            Val > 0 andalso
+            Val =< ?MAX_PACKET_SIZE).
