@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2018-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 %% This module is a pseudo queue type.
@@ -27,7 +27,7 @@
          is_stateful/0,
          declare/2,
          delete/4,
-         deliver/2,
+         deliver/3,
          is_enabled/0,
          is_compatible/3,
          is_recoverable/1,
@@ -36,6 +36,7 @@
          policy_changed/1,
          info/2,
          stat/1,
+         format/2,
          capabilities/0,
          notify_decorators/1
         ]).
@@ -70,19 +71,26 @@ is_stateful() ->
     {'new' | 'existing' | 'owner_died', amqqueue:amqqueue()} |
     {'absent', amqqueue:amqqueue(), rabbit_amqqueue:absent_reason()}.
 declare(Q0, _Node) ->
+    Q1 = case amqqueue:get_pid(Q0) of
+             none ->
+                 %% declaring process becomes the queue
+                 amqqueue:set_pid(Q0, self());
+             Pid when is_pid(Pid) ->
+                 Q0
+         end,
     %% The queue gets persisted such that routing to this
     %% queue (via the topic exchange) works as usual.
-    case rabbit_amqqueue:internal_declare(Q0, false) of
+    case rabbit_amqqueue:internal_declare(Q1, false) of
         {created, Q} ->
             Opts = amqqueue:get_options(Q),
             ActingUser = maps:get(user, Opts, ?UNKNOWN_USER),
             rabbit_event:notify(queue_created,
-                                [{name, amqqueue:get_name(Q0)},
+                                [{name, amqqueue:get_name(Q)},
                                  {durable, true},
                                  {auto_delete, false},
                                  {exclusive, true},
-                                 {type, amqqueue:get_type(Q0)},
-                                 {arguments, amqqueue:get_arguments(Q0)},
+                                 {type, amqqueue:get_type(Q)},
+                                 {arguments, amqqueue:get_arguments(Q)},
                                  {user_who_performed_action, ActingUser}]),
             {new, Q};
         Other ->
@@ -100,38 +108,39 @@ delete(Q, _IfUnused, _IfEmpty, ActingUser) ->
     ok = rabbit_amqqueue:internal_delete(Q, ActingUser),
     {ok, 0}.
 
--spec deliver([{amqqueue:amqqueue(), stateless}], Delivery :: term()) ->
+-spec deliver([{amqqueue:amqqueue(), stateless}],
+              Msg :: mc:state(),
+              rabbit_queue_type:delivery_options()) ->
     {[], rabbit_queue_type:actions()}.
-deliver(Qs, #delivery{message = BasicMessage,
-                      confirm = Confirm,
-                      msg_seq_no = SeqNo}) ->
-    Msg = {queue_event, ?MODULE,
-           {?MODULE, _QPid = none, _QMsgId = none, _Redelivered = false, BasicMessage}},
+deliver(Qs, Msg, Options) ->
+    Evt = {queue_event, ?MODULE,
+           {?MODULE, _QPid = none, _QMsgId = none, _Redelivered = false, Msg}},
     {Pids, Actions} =
-    case Confirm of
-        false ->
-            Pids0 = lists:map(fun({Q, stateless}) -> amqqueue:get_pid(Q) end, Qs),
-            {Pids0, []};
-        true ->
-            %% We confirm the message directly here in the queue client.
-            %% Alternatively, we could have the target MQTT connection process confirm the message.
-            %% However, given that this message might be lost anyway between target MQTT connection
-            %% process and MQTT subscriber, and we know that the MQTT subscriber wants to receive
-            %% this message at most once, we confirm here directly.
-            %% Benefits:
-            %% 1. We do not block sending the confirmation back to the publishing client just because a single
-            %% (at-most-once) target queue out of potentially many (e.g. million) queues might be unavailable.
-            %% 2. Memory usage in this (publishing) process is kept lower because the target queue name can be
-            %% directly removed from rabbit_mqtt_confirms and rabbit_confirms.
-            %% 3. Reduced network traffic across RabbitMQ nodes.
-            %% 4. Lower latency of sending publisher confirmation back to the publishing client.
-            SeqNos = [SeqNo],
-            lists:mapfoldl(fun({Q, stateless}, Actions) ->
-                                   {amqqueue:get_pid(Q),
-                                    [{settled, amqqueue:get_name(Q), SeqNos} | Actions]}
-                           end, [], Qs)
-    end,
-    delegate:invoke_no_result(Pids, {gen_server, cast, [Msg]}),
+        case maps:get(correlation, Options, undefined) of
+            undefined ->
+                Pids0 = lists:map(fun({Q, stateless}) -> amqqueue:get_pid(Q) end, Qs),
+                {Pids0, []};
+            Corr ->
+                %% We confirm the message directly here in the queue client.
+                %% Alternatively, we could have the target MQTT connection process confirm the message.
+                %% However, given that this message might be lost anyway between target MQTT connection
+                %% process and MQTT subscriber, and we know that the MQTT subscriber wants to receive
+                %% this message at most once, we confirm here directly.
+                %% Benefits:
+                %% 1. We do not block sending the confirmation back to the publishing client just because a single
+                %% (at-most-once) target queue out of potentially many (e.g. million) queues might be unavailable.
+                %% 2. Memory usage in this (publishing) process is kept lower because the target queue name can be
+                %% directly removed from rabbit_mqtt_confirms and rabbit_confirms.
+                %% 3. Reduced network traffic across RabbitMQ nodes.
+                %% 4. Lower latency of sending publisher confirmation back to the publishing client.
+                Corrs = [Corr],
+                lists:mapfoldl(fun({Q, stateless}, Actions) ->
+                                       {amqqueue:get_pid(Q),
+                                        [{settled, amqqueue:get_name(Q), Corrs}
+                                         | Actions]}
+                               end, [], Qs)
+        end,
+    delegate:invoke_no_result(Pids, {gen_server, cast, [Evt]}),
     {[], Actions}.
 
 -spec is_enabled() -> boolean().
@@ -194,6 +203,12 @@ notify_decorators(_) ->
 stat(_Q) ->
     {ok, 0, 0}.
 
+-spec format(amqqueue:amqqueue(), map()) ->
+    [{atom(), term()}].
+format(Q, _Ctx) ->
+    [{type, ?MODULE},
+     {state, amqqueue:get_state(Q)}].
+
 -spec capabilities() ->
     #{atom() := term()}.
 capabilities() ->
