@@ -3,7 +3,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2019-2022 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_boot_state).
@@ -35,7 +35,7 @@ get() ->
 
 -spec set(boot_state()) -> ok.
 set(BootState) ->
-    ?LOG_DEBUG("Change boot state to `~s`", [BootState],
+    ?LOG_DEBUG("Change boot state to `~ts`", [BootState],
                #{domain => ?RMQLOG_DOMAIN_PRELAUNCH}),
     ?assert(is_valid(BootState)),
     case BootState of
