@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_exchange_type_random).
@@ -19,14 +19,14 @@
 -export([
   add_binding/3, 
   assert_args_equivalence/2,
-  create/2, 
-  delete/2, 
+  create/2,
+  delete/2,
   policy_changed/2,
   description/0, 
   recover/2, 
   remove_bindings/3,
   validate_binding/2,
-  route/2,
+  route/3,
   serialise_events/0,
   validate/1,
   info/1,
@@ -36,7 +36,7 @@
 description() ->
     [{name, <<"x-random">>}, {description, <<"Randomly picks a binding (queue) to route via (to).">>}].
 
-route(_X=#exchange{name = Name}, _Delivery) ->
+route(#exchange{name = Name}, _Message, _Options) ->
     Matches = rabbit_router:match_routing_key(Name, ['_']),
     case length(Matches) of
       Len when Len < 2 -> Matches;
