@@ -2,13 +2,14 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_shovel_worker_sup).
 -behaviour(mirrored_supervisor).
 
 -export([start_link/2, init/1]).
+-export([id_to_khepri_path/1]).
 
 -include("rabbit_shovel.hrl").
 -include_lib("rabbit_common/include/rabbit.hrl").
@@ -18,7 +19,7 @@ start_link(ShovelName, ShovelConfig) ->
                                    ?MODULE, [ShovelName, ShovelConfig]).
 
 init([Name, Config]) ->
-    ChildSpecs = [{Name,
+    ChildSpecs = [{id(Name),
                    {rabbit_shovel_worker, start_link, [static, Name, Config]},
                    case Config of
                        #{reconnect_delay := N}
@@ -29,3 +30,9 @@ init([Name, Config]) ->
                    worker,
                    [rabbit_shovel_worker]}],
     {ok, {{one_for_one, 1, ?MAX_WAIT}, ChildSpecs}}.
+
+id(Name) when is_atom(Name) ->
+    Name.
+
+id_to_khepri_path(Name) when is_atom(Name) ->
+    [Name].
