@@ -2,7 +2,7 @@
 %% License, v. 2.0. If a copy of the MPL was not distributed with this
 %% file, You can obtain one at https://mozilla.org/MPL/2.0/.
 %%
-%% Copyright (c) 2007-2023 VMware, Inc. or its affiliates.  All rights reserved.
+%% Copyright (c) 2007-2024 Broadcom. All Rights Reserved. The term “Broadcom” refers to Broadcom Inc. and/or its subsidiaries. All rights reserved.
 %%
 
 -module(rabbit_web_mqtt_handler).
@@ -12,6 +12,7 @@
 -include_lib("kernel/include/logger.hrl").
 -include_lib("rabbit_common/include/logging.hrl").
 -include_lib("rabbitmq_mqtt/include/rabbit_mqtt.hrl").
+-include_lib("rabbitmq_mqtt/include/rabbit_mqtt_packet.hrl").
 
 -export([
     init/2,
@@ -22,13 +23,14 @@
 ]).
 
 -export([conserve_resources/3]).
+-export([info/2]).
 
 %% cowboy_sub_protocol
 -export([upgrade/4,
          upgrade/5,
          takeover/7]).
 
--type option(T) :: undefined | T.
+-define(APP, rabbitmq_web_mqtt).
 
 -record(state, {
           socket :: {rabbit_proxy_socket, any(), any()} | rabbit_net:socket(),
@@ -39,13 +41,15 @@
           conserve = false :: boolean(),
           stats_timer :: option(rabbit_event:state()),
           keepalive = rabbit_mqtt_keepalive:init() :: rabbit_mqtt_keepalive:state(),
-          conn_name :: option(binary())
+          conn_name :: option(binary()),
+          should_use_fhc :: rabbit_types:option(boolean())
         }).
 
 -type state() :: #state{}.
 
 %% Close frame status codes as defined in https://www.rfc-editor.org/rfc/rfc6455#section-7.4.1
 -define(CLOSE_NORMAL, 1000).
+-define(CLOSE_SERVER_GOING_DOWN, 1001).
 -define(CLOSE_PROTOCOL_ERROR, 1002).
 -define(CLOSE_UNACCEPTABLE_DATA_TYPE, 1003).
 
@@ -72,25 +76,50 @@ init(Req, Opts) ->
         undefined ->
             no_supported_sub_protocol(undefined, Req);
         Protocol ->
-            WsOpts0 = proplists:get_value(ws_opts, Opts, #{}),
-            WsOpts  = maps:merge(#{compress => true}, WsOpts0),
             case lists:member(<<"mqtt">>, Protocol) of
                 false ->
                     no_supported_sub_protocol(Protocol, Req);
                 true ->
+                    WsOpts0 = proplists:get_value(ws_opts, Opts, #{}),
+                    WsOpts  = maps:merge(#{compress => true}, WsOpts0),
+                    ShouldUseFHC = application:get_env(?APP, use_file_handle_cache, true),
+                    case ShouldUseFHC of
+                      true  -> ?LOG_INFO("Web MQTT: file handle cache use is enabled");
+                      false -> ?LOG_INFO("Web MQTT: file handle cache use is disabled")
+                    end,
+
                     {?MODULE,
                      cowboy_req:set_resp_header(<<"sec-websocket-protocol">>, <<"mqtt">>, Req),
-                     #state{socket = maps:get(proxy_header, Req, undefined)},
+                     #state{socket = maps:get(proxy_header, Req, undefined), should_use_fhc = ShouldUseFHC},
                      WsOpts}
             end
     end.
 
+%% We cannot use a gen_server call, because the handler process is a
+%% special cowboy_websocket process (not a gen_server) which assumes
+%% all gen_server calls are supervisor calls, and does not pass on the
+%% request to this callback module. (see cowboy_websocket:loop/3 and
+%% cowboy_children:handle_supervisor_call/4) However using a generic
+%% gen:call with a special label ?MODULE works fine.
+-spec info(pid(), rabbit_types:info_keys()) ->
+    rabbit_types:infos().
+info(Pid, all) ->
+    info(Pid, ?INFO_ITEMS);
+info(Pid, Items) ->
+    {ok, Res} = gen:call(Pid, ?MODULE, {info, Items}),
+    Res.
 -spec websocket_init(state()) ->
     {cowboy_websocket:commands(), state()} |
     {cowboy_websocket:commands(), state(), hibernate}.
-websocket_init(State0 = #state{socket = Sock}) ->
+websocket_init(State0 = #state{socket = Sock, should_use_fhc = ShouldUseFHC}) ->
     logger:set_process_metadata(#{domain => ?RMQLOG_DOMAIN_CONN ++ [web_mqtt]}),
-    ok = file_handle_cache:obtain(),
+    case ShouldUseFHC of
+      true  ->
+        ok = file_handle_cache:obtain();
+      false -> ok;
+      undefined ->
+        ok = file_handle_cache:obtain()
+    end,
     case rabbit_net:connection_string(Sock, inbound) of
         {ok, ConnStr} ->
             ConnName = rabbit_data_coercion:to_binary(ConnStr),
@@ -138,8 +167,8 @@ websocket_info({bump_credit, Msg}, State) ->
     handle_credits(State);
 websocket_info({reply, Data}, State) ->
     {[{binary, Data}], State, hibernate};
-websocket_info({stop, CloseCode, Error}, State) ->
-    stop(State, CloseCode, Error);
+websocket_info({stop, CloseCode, Error, SendWill}, State) ->
+    stop({SendWill, State}, CloseCode, Error);
 websocket_info({'EXIT', _, _}, State) ->
     stop(State);
 websocket_info({'$gen_cast', QueueEvent = {queue_event, _, _}},
@@ -152,16 +181,30 @@ websocket_info({'$gen_cast', QueueEvent = {queue_event, _, _}},
                        [State#state.conn_name, Reason]),
             stop(State#state{proc_state = PState})
     end;
-websocket_info({'$gen_cast', duplicate_id}, State = #state{ proc_state = ProcState,
-                                                            conn_name = ConnName }) ->
+websocket_info({'$gen_cast', duplicate_id}, State) ->
+    %% Delete this backward compatibility clause when feature flag
+    %% delete_ra_cluster_mqtt_node becomes required.
+    websocket_info({'$gen_cast', {duplicate_id, true}}, State);
+websocket_info({'$gen_cast', {duplicate_id, SendWill}},
+               State = #state{proc_state = ProcState,
+                              conn_name = ConnName}) ->
     ?LOG_WARNING("Web MQTT disconnecting a client with duplicate ID '~s' (~p)",
                  [rabbit_mqtt_processor:info(client_id, ProcState), ConnName]),
-    stop(State);
-websocket_info({'$gen_cast', {close_connection, Reason}}, State = #state{ proc_state = ProcState,
-                                                                          conn_name = ConnName }) ->
+    rabbit_mqtt_processor:send_disconnect(?RC_SESSION_TAKEN_OVER, ProcState),
+    defer_close(?CLOSE_NORMAL, SendWill),
+    {[], State};
+websocket_info({'$gen_cast', {close_connection, Reason}}, State = #state{proc_state = ProcState,
+                                                                         conn_name = ConnName}) ->
     ?LOG_WARNING("Web MQTT disconnecting client with ID '~s' (~p), reason: ~s",
                  [rabbit_mqtt_processor:info(client_id, ProcState), ConnName, Reason]),
-    stop(State);
+    case Reason of
+        maintenance ->
+            rabbit_mqtt_processor:send_disconnect(?RC_SERVER_SHUTTING_DOWN, ProcState),
+            defer_close(?CLOSE_SERVER_GOING_DOWN),
+            {[], State};
+        _ ->
+            stop(State)
+    end;
 websocket_info({'$gen_cast', {force_event_refresh, Ref}}, State0) ->
     Infos = infos(?EVENT_KEYS, State0),
     rabbit_event:notify(connection_created, Infos, Ref),
@@ -173,14 +216,17 @@ websocket_info({'$gen_cast', refresh_config},
     PState = rabbit_mqtt_processor:update_trace(ConnName, PState0),
     State = State0#state{proc_state = PState},
     {[], State, hibernate};
-websocket_info({keepalive, Req}, State = #state{keepalive = KState0,
+websocket_info({keepalive, Req}, State = #state{proc_state = ProcState,
+                                                keepalive = KState0,
                                                 conn_name = ConnName}) ->
     case rabbit_mqtt_keepalive:handle(Req, KState0) of
         {ok, KState} ->
             {[], State#state{keepalive = KState}, hibernate};
         {error, timeout} ->
             ?LOG_ERROR("keepalive timeout in Web MQTT connection ~p", [ConnName]),
-            stop(State, ?CLOSE_NORMAL, <<"MQTT keepalive timeout">>);
+            rabbit_mqtt_processor:send_disconnect(?RC_KEEP_ALIVE_TIMEOUT, ProcState),
+            defer_close(?CLOSE_NORMAL),
+            {[], State};
         {error, Reason} ->
             ?LOG_ERROR("keepalive error in Web MQTT connection ~p: ~p",
                        [ConnName, Reason]),
@@ -212,6 +258,10 @@ websocket_info(connection_created, State) ->
     rabbit_core_metrics:connection_created(self(), Infos),
     rabbit_event:notify(connection_created, Infos),
     {[], State, hibernate};
+websocket_info({?MODULE, From, {info, Items}}, State) ->
+    Infos = infos(Items, State),
+    gen:reply(From, Infos),
+    {[], State, hibernate};
 websocket_info(Msg, State) ->
     ?LOG_WARNING("Web MQTT: unexpected message ~tp", [Msg]),
     {[], State, hibernate}.
@@ -221,17 +271,24 @@ terminate(Reason, Request, #state{} = State) ->
 terminate(_Reason, _Request,
           {SendWill, #state{conn_name = ConnName,
                             proc_state = PState,
-                            keepalive = KState} = State}) ->
+                            keepalive = KState,
+                            should_use_fhc = ShouldUseFHC} = State}) ->
     ?LOG_INFO("Web MQTT closing connection ~ts", [ConnName]),
     maybe_emit_stats(State),
     _ = rabbit_mqtt_keepalive:cancel_timer(KState),
-    ok = file_handle_cache:release(),
+    case ShouldUseFHC of
+      true  ->
+        ok = file_handle_cache:release();
+      false -> ok;
+      undefined ->
+        ok = file_handle_cache:release()
+    end,
     case PState of
         connect_packet_unprocessed ->
             ok;
         _ ->
             Infos = infos(?EVENT_KEYS, State),
-            rabbit_mqtt_processor:terminate(SendWill, ConnName, Infos, PState)
+            rabbit_mqtt_processor:terminate(SendWill, Infos, PState)
     end.
 
 %% Internal.
@@ -239,7 +296,12 @@ terminate(_Reason, _Request,
 no_supported_sub_protocol(Protocol, Req) ->
     %% The client MUST include “mqtt” in the list of WebSocket Sub Protocols it offers [MQTT-6.0.0-3].
     ?LOG_ERROR("Web MQTT: 'mqtt' not included in client offered subprotocols: ~tp", [Protocol]),
-    {ok, cowboy_req:reply(400, #{<<"connection">> => <<"close">>}, Req), #state{}}.
+    %% Set should_use_fhc to false, because at this early stage of init no fhc
+    %% obtain was called, so terminate/3 should not call fhc release
+    %% (even if use_file_handle_cache is true)
+    {ok,
+     cowboy_req:reply(400, #{<<"connection">> => <<"close">>}, Req),
+     #state{should_use_fhc = false}}.
 
 handle_data(Data, State0 = #state{}) ->
     case handle_data1(Data, State0) of
@@ -255,12 +317,12 @@ handle_data1(Data, State = #state{socket = Socket,
                                   parse_state = ParseState,
                                   proc_state = ProcState,
                                   conn_name = ConnName}) ->
-    case parse(Data, ParseState) of
+    try rabbit_mqtt_packet:parse(Data, ParseState) of
         {more, ParseState1} ->
             {ok, ensure_stats_timer(
                    control_throttle(
                      State#state{parse_state = ParseState1})), hibernate};
-        {ok, Packet, Rest} ->
+        {ok, Packet, Rest, ParseState1} ->
             case ProcState of
                 connect_packet_unprocessed ->
                     case rabbit_mqtt_processor:init(Packet, rabbit_net:unwrap_socket(Socket),
@@ -269,40 +331,50 @@ handle_data1(Data, State = #state{socket = Socket,
                             ?LOG_INFO("Accepted Web MQTT connection ~ts for client ID ~ts",
                                       [ConnName, rabbit_mqtt_processor:info(client_id, ProcState1)]),
                             handle_data1(
-                              Rest, State#state{parse_state = rabbit_mqtt_packet:reset_state(),
+                              Rest, State#state{parse_state = ParseState1,
                                                 proc_state = ProcState1});
                         {error, Reason} ->
                             ?LOG_ERROR("Rejected Web MQTT connection ~ts: ~p", [ConnName, Reason]),
-                            self() ! {stop, ?CLOSE_PROTOCOL_ERROR, connect_packet_rejected},
-                            {[], {_SendWill = false, State}}
+                            self() ! {stop, ?CLOSE_PROTOCOL_ERROR, connect_packet_rejected,
+                                      _SendWill = false},
+                            {[], State}
                     end;
                 _ ->
                     case rabbit_mqtt_processor:process_packet(Packet, ProcState) of
                         {ok, ProcState1} ->
                             handle_data1(
                               Rest,
-                              State#state{parse_state = rabbit_mqtt_packet:reset_state(),
+                              State#state{parse_state = ParseState1,
                                           proc_state = ProcState1});
                         {error, Reason, _} ->
                             stop_mqtt_protocol_error(State, Reason, ConnName);
-                        {stop, disconnect, ProcState1} ->
-                            stop({_SendWill = false, State#state{proc_state = ProcState1}})
+                        {stop, {disconnect, server_initiated}, _} ->
+                            defer_close(?CLOSE_PROTOCOL_ERROR),
+                            {[], State};
+                        {stop, {disconnect, {client_initiated, SendWill}}, ProcState1} ->
+                            stop({SendWill, State#state{proc_state = ProcState1}})
                     end
             end;
+        {error, {disconnect_reason_code, ReasonCode}} ->
+            rabbit_mqtt_processor:send_disconnect(ReasonCode, ProcState),
+            defer_close(?CLOSE_PROTOCOL_ERROR),
+            {[], State};
         {error, Reason} ->
             stop_mqtt_protocol_error(State, Reason, ConnName)
+    catch _:Reason:Stacktrace ->
+              ?LOG_DEBUG("Web MQTT cannot parse a packet, reason: ~tp, "
+                         "stacktrace: ~tp, payload (first 100 bytes): ~tp",
+                         [Reason, Stacktrace, rabbit_mqtt_util:truncate_binary(Data, 100)]),
+              stop_mqtt_protocol_error(State, cannot_parse, ConnName)
     end.
 
-parse(Data, ParseState) ->
-    try
-        rabbit_mqtt_packet:parse(Data, ParseState)
-    catch
-        _:Reason:Stacktrace ->
-            ?LOG_DEBUG("Web MQTT cannot parse a packet, reason: ~tp, stacktrace: ~tp, "
-                       "payload (first 100 bytes): ~tp",
-                       [Reason, Stacktrace, rabbit_mqtt_util:truncate_binary(Data, 100)]),
-            {error, cannot_parse}
-    end.
+%% Allow DISCONNECT packet to be sent to client before closing the connection.
+defer_close(CloseStatusCode) ->
+    defer_close(CloseStatusCode, true).
+
+defer_close(CloseStatusCode, SendWill) ->
+    self() ! {stop, CloseStatusCode, server_initiated_disconnect, SendWill},
+    ok.
 
 stop_mqtt_protocol_error(State, Reason, ConnName) ->
     ?LOG_WARNING("Web MQTT protocol error ~tp for connection ~tp", [Reason, ConnName]),
